package ru.crp.lvms.managers;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.events.NodeStatusUpdatedEvent;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.NodesDao;
import ru.crp.lvms.model.dao.SpaceNodesDao;
import ru.crp.lvms.model.dao.TaskDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.nodes.ExtendedNode;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.Set;

@Service
public class NodesManager extends BaseManagerSuggest<NodeEntity> {
    private static final Logger LOGGER = LoggerFactory.getLogger(NodesManager.class);

    @Autowired
    private NodesDao nodesDao;

    @Autowired
    private SpaceNodesDao spaceNodesDao;

    @Autowired
    private TaskDao taskDao;

    // expected values = 'local', 'stage', 'production'
    @Value("${environment}")
    private String environment;

    @Autowired
    private ClustersManager clustersManager;

    @Autowired
    private ApplicationEventPublisher applicationEventPublisher;

    @PostConstruct
    public void init() {
        // we need to modify local solver url while deploying on stage
        if ("stage".equals(environment)) {
            List<NodeEntity> localSolvers = nodesDao.find("WHERE url LIKE ?", "%localhost%");
            localSolvers.forEach(localSolver -> {
                localSolver.setUrl("http://localhost:9090/cml-solver/");
                nodesDao.save(localSolver);
            });
        }
    }


    // override to suggest only space nodes not slurm
    @Override
    public List<NodeEntity> suggest(SuggestRequest suggestRequest, UserEntity currentUser) {
        return spaceNodesDao.suggest(suggestRequest, currentUser.getId());
    }

    public Optional<NodeEntity> getMostSuitableNodeForTask(TaskEntity task) {
        NodeEntity freeNode = nodesDao.findOneFreeWithRestrictions(task.getMemory(),
                task.getNumOfCores(), task.getSolverName());
        return Optional.ofNullable(freeNode);
    }

    @Nullable
    public ExtendedNode extendNode(@Nullable ExtendedNode node) {
        if (node == null) {
            return null;
        }
        TaskEntity runningOnNode = taskDao.findRunningOnNode(node.getId());
        if (runningOnNode != null) {
            node.setHasSolvingTask(true);
            node.setStatus(taskToNodeStatus(runningOnNode.getStatusId()));
        } else {
            node.setStatus("Free");
            node.setHasSolvingTask(false);
        }
        return node;
    }

    public List<ExtendedNode> extendNodes(List<ExtendedNode> extendedNodes) {
        extendedNodes.forEach(this::extendNode);
        return extendedNodes;
    }

    private String taskToNodeStatus(int taskStatusId) {
        CalculationStatus taskStatus = CalculationStatus.valueOf(taskStatusId);
        switch (taskStatus) {
            case Sending:
            case Starting:
                return "Copying data";
            case Solving:
                return "Solving";
            case CalculationFinished:
            case ImportingResults:
                return "Importing results";
            case Terminating:
                return "Terminating";
            default:
                LOGGER.error("Unexpected task status is used " + taskStatus);
                return "Internal error";
        }
    }

    @Nonnull
    @Override
    public NodeEntity createAndRead(NodeEntity newNodeIn, @Nullable UserEntity currentUser) {
        if (currentUser != null) {
            newNodeIn.setOwnerId(currentUser.getId());
        }
        newNodeIn.setCreationDate(new Date());
        newNodeIn.setModificationDate(new Date());
        NodeEntity newNode = super.createAndRead(newNodeIn, currentUser);
        applicationEventPublisher.publishEvent(new NodeStatusUpdatedEvent(null, newNode));
        return newNode;
    }

    @Override
    public void delete(NodeEntity entity, @Nullable UserEntity currentUser) {
        super.delete(entity, currentUser);
        applicationEventPublisher.publishEvent(new NodeStatusUpdatedEvent(entity, null));
    }

    @Override
    protected BaseSpaceDao<NodeEntity> aliveDao() {
        return nodesDao;
    }

    public void updateNodeSolvers(NodeEntity node, Set<String> solverNames) {
        nodesDao.updateNodeSolvers(node.getId(), solverNames);
    }

    public List<NodeEntity> findByUrl(String url) {
        return nodesDao.findByUrl(url);
    }

    public List<String> findNodeSolvers(int nodeId) {
        return nodesDao.findNodeSolvers(nodeId);
    }

    @Transactional(readOnly = true)
    // returns alive status from db which can be outdated
    public boolean getNodeAliveStatus(int nodeId) {
        return nodesDao.getNodeAliveStatus(nodeId);
    }

    @Nullable
    public NodeStatus getNodeStatus(int nodeId) {
        NodeEntity node = nodesDao.findOne(nodeId);
        NodeStatus nodeStatus = getNodeStatus(node);
        if (nodeStatus != null) {
            setNodeAliveStatus(node, true);
        }
        return nodeStatus;
    }

    @Nullable
    public NodeStatus getNodeStatus(NodeEntity node) {
        return clustersManager.getNodeStatus(node);
    }

    @Transactional
    public void setNodeAliveStatus(NodeEntity node, boolean alive) {
        node.setIsAlive(alive);
        nodesDao.update(node);
    }

    @Transactional(readOnly = true)
    public List<NodeEntity> getAllForCluster(int clusterId) {
        return nodesDao.getAllForCluster(clusterId);
    }

    // returns most recent alive status
    public boolean checkIsNodeAlive(NodeEntity node) {
        return getNodeStatus(node) != null;
    }
}
