package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.dao.JdbcCrudDao;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.model.dao.CompositeMaterialDao;
import ru.crp.lvms.model.dao.MaterialDao;
import ru.crp.lvms.model.dao.ReferenceGraphDao;
import ru.crp.lvms.model.dao.construction.elements.ConstructionElementDao;
import ru.crp.lvms.model.entities.ReferenceGraphEntry;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;

import javax.annotation.PostConstruct;
import java.util.*;

import static java.util.Collections.emptyList;

@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
@Service
public class ReferenceGraphManager {
    public static final String MATERIAL = "material";
    public static final String COMPOSITE_MATERIAL = "compositeMaterial";
    public static final String CONSTRUCTION_ELEMENT = "constructionElement";
    public static final String OPTIMIZATION = "optimization";


    @Autowired
    private ReferenceGraphDao referenceGraphDao;

    @Autowired
    private MaterialDao materialDao;

    @Autowired
    private CompositeMaterialDao compositeMaterialDao;

    @Autowired
    private ConstructionElementDao constructionElementDao;

    @Autowired
    private OptimizationDao optimizationDao;

    private Map<String, JdbcCrudDao<?>> tableNames;


    @PostConstruct
    private void init() {
        tableNames = new HashMap<>();
        tableNames.put(MATERIAL, materialDao);
        tableNames.put(COMPOSITE_MATERIAL, compositeMaterialDao);
        tableNames.put(CONSTRUCTION_ELEMENT, constructionElementDao);
        tableNames.put(OPTIMIZATION, optimizationDao);
    }


    public boolean hasRelationGraph(String type) {
        return tableNames.containsKey(type);
    }


    public Set<String> getAllReferenceTypes() {
        return tableNames.keySet();
    }


    private List<ReferenceGraphEntry> getReferenceGraphFlatten(int entranceId, String tableName) {
        // entrance is first element
        List<ReferenceGraphEntry> ascendingBranch =
                referenceGraphDao.getAscendentBranchAndMe(entranceId, tableName);
        // then no node found
        if (ascendingBranch.isEmpty()) {
            return emptyList();
        }
        ascendingBranch.forEach(graphEntry -> graphEntry.setIsMainPath(true));
        ascendingBranch.get(0).setIsEntrance(true);
        // entrance is last element
        List<ReferenceGraphEntry> descendingBranch =
                referenceGraphDao.getDescendantBranch(entranceId, tableName);

        descendingBranch.remove(descendingBranch.size() - 1); // to prevent duplication of entrance element
        descendingBranch.addAll(ascendingBranch);
        return descendingBranch;
    }


    public ReferenceGraphEntry getReferenceGraph(int entranceId, String type)
            throws NotFoundException {
        if (!tableNames.containsKey(type)) {
            throw new IllegalArgumentException("Unexpected item type parameter: " + type);
        }

        JdbcCrudDao<?> daoForObjectType = tableNames.get(type);
        Entity entranceEntity = daoForObjectType.findOne(entranceId);
        if (entranceEntity == null) {
            throw new NotFoundException("No object with id " + entranceId + " found");
        }

        List<ReferenceGraphEntry> referenceGraphFlatten =
                getReferenceGraphFlatten(entranceId, daoForObjectType.getTableName());
        Map<Integer, List<ReferenceGraphEntry>> parentChildren = new HashMap<>();
        ReferenceGraphEntry root = null;
        for (final ReferenceGraphEntry current : referenceGraphFlatten) {
            if (current.isRoot()) {
                current.setChildren(parentChildren.getOrDefault(current.getId(), null));
                root = current;
                break;
            }
            Integer parentId = current.getReferenceId();
            parentChildren.putIfAbsent(parentId, new ArrayList<>());
            parentChildren.computeIfPresent(parentId, (pid, children) -> {
                children.add(current);
                return children;
            });
            parentChildren.computeIfPresent(current.getId(), (id, children) -> {
                current.setChildren(children);
                return children;
            });
        }
        return root;
    }
}
