package ru.crp.lvms.managers;

import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.*;
import ru.crp.lvms.model.dao.effective.characteristics.ECDao;
import ru.crp.lvms.model.dao.effective.characteristics.ECParamDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.Layer;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.model.entities.technology.TechnologySimulationsEntity;
import ru.crp.lvms.model.entities.technology.TechnologySimulationsParamEntity;
import ru.crp.lvms.model.entities.technology.TechnologySimulationsRibsEntity;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.tasks.fileGenerators.TSInputFileGenerator;
import ru.crp.lvms.tasks.fileGenerators.TSParameters;
import ru.crp.lvms.tasks.slurm.callbacks.SlurmCheckTaskCommandCallback;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.net.URLEncoder;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.function.Function;
import java.util.logging.Logger;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;
import static ru.crp.lvms.utils.MessageUtils.TECHNOLOGY_USED_IN_MATERIAL;

@Service
public class TechnologySimulationsManager extends BaseManagerSuggest<TechnologySimulationsEntity> {
    private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(SlurmCheckTaskCommandCallback.class);

    @Autowired
    private TechnologySimulationsDao technologySimulationsDao;

    @Autowired
    private TechnologySimulationsRibsDao technologySimulationsRibsDao;

    @Autowired
    private TechnologySimulationsParamDao technologySimulationsParamDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private TSInputFileGenerator tsInputFileGenerator;

    @Nonnull
//    @Override
//    public BaseSpaceDao<TechnologySimulationsEntity> aliveDao() {
//        return technologySimulationsDao;
//    }

    @Override
    public TechnologySimulationsDao aliveDao() {
        return technologySimulationsDao;
    }

    public TechnologySimulationsEntity createTechnologySimulations(TechnologySimulationsEntity newTechnologySimulations,
                                             @Nullable UserEntity currentUser) {
        newTechnologySimulations.setOwnerId(currentUser.getId());
        TechnologySimulationsEntity createdTS = createAndRead(newTechnologySimulations, currentUser);
        technologySimulationsParamDao.getAllParamIds().stream()
                .map(paramId -> new TechnologySimulationsParamEntity(createdTS.getId(), paramId))
                .forEach(technologySimulationsParamDao::insert);

        return createdTS;
    }

    public TechnologySimulationsEntity extend(TechnologySimulationsEntity entity, UserEntity currentUser) {
        entity.setRibs(technologySimulationsRibsDao.findByTechnologySimulationsId(entity.getId()));
        entity.setParams(technologySimulationsParamDao.getByTechnologyId(entity.getId()));
        return entity;
    }


    public List<TechnologySimulationsParamEntity> getTSParameters(Integer technologySimualtionsId) {
        return technologySimulationsParamDao.getByTechnologyId(technologySimualtionsId);
    }
    public TechnologySimulationsParamEntity createTSParam(TechnologySimulationsParamEntity newParamDTO) {
        TechnologySimulationsParamEntity existedParam = technologySimulationsParamDao.getByTechnologyIdAndParamId(newParamDTO.getTechnologySimulationId(), newParamDTO.getParameterId());
        if (existedParam != null) {
            throw new BadRequestException("Параметр '" + existedParam.getParameterName() + "' уже добавлен к эффективной характеристике");
        }
        return technologySimulationsParamDao.insertAndRead(newParamDTO);
    }

    public TechnologySimulationsParamEntity getTSParameter(Integer technologySimualtionsId, Integer id, PortalUser currentUser) {
        return technologySimulationsParamDao.findOne(id, technologySimualtionsId);
    }

    public TechnologySimulationsParamEntity updateTSParam(TechnologySimulationsParamEntity existedParam, TechnologySimulationsParamEntity updateDTO) {
        existedParam.setParameterValue(updateDTO.getParameterValue());
        return technologySimulationsParamDao.updateAndRead(existedParam);
    }

    public void deleteTSParam(TechnologySimulationsParamEntity param) {
        technologySimulationsParamDao.delete(param.getTechnologySimulationId());
    }


    public void startTechnologySimulations(HttpServletResponse response, int technologySimulationsId, UserEntity currentUser) throws IOException, ResourcePreparationException {
        TechnologySimulationsEntity technologySimulations = technologySimulationsDao.findOne(technologySimulationsId);
        Path tsSimulationFiles = tsInputFileGenerator.getInputFilesForTechnologySimulations(technologySimulations,getRibs(technologySimulations.getId(),true, currentUser));

        response.setHeader("Content-Type", "application/octet-stream");


        // see more http://stackoverflow.com/questions/93551/how-to-encode-the-filename-parameter-of-content-disposition-header-in-http
        try {
            String contentDisposition = "attachment; filename=\"" + technologySimulations.getName()+".gcode" + "\"; filename*=UTF-8''" + URLEncoder.encode(technologySimulations.getName()+".gcode", "UTF-8").replace("+", "%20");
            response.setHeader("Content-Disposition", contentDisposition);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("UTF-8 is not supported");
        }

        Path tempGcodeFile = Files.createTempFile(Paths.get("/home/space/Gcodes_work"),"generated",".gcode");

        convertJsontoGcode(tempGcodeFile, tsSimulationFiles);

        Files.copy(tempGcodeFile, response.getOutputStream());


        return;
    }


    public void convertJsontoGcode (Path tempGcodeFile, Path pathToJsonFile ) {
        try {
            LOGGER.warn("I want to start converting"+ " -i " + pathToJsonFile.toString() + " -o " + tempGcodeFile.toString());
            String carbonnetPath = "/home/space/Gcodes3/bin/release/carbonnet_main.exe";
            String convertJsonPath = " -i " + pathToJsonFile.toString() + " -o " + tempGcodeFile.toString();
            Process p = Runtime.getRuntime().exec(carbonnetPath + convertJsonPath);
            p.waitFor();
            LOGGER.warn("Converting " +convertJsonPath.toString());
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        }
    }



    public List<TechnologySimulationsRibsEntity> getRibs(int technologySimulationsId, boolean checkRights, UserEntity currentUser) {
        if (checkRights) {
            checkEntityExistsAndAccesible(technologySimulationsId, AccessRights.update, currentUser);
        }
        return technologySimulationsRibsDao.findByTechnologySimulationsId(technologySimulationsId);
    }

    public void addRibs(int technologySimulationsId, TechnologySimulationsRibsEntity newRibs, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologySimulationsId, AccessRights.update, currentUser);
        validateRibStep(newRibs,currentUser);
        if (newRibs.getPosition() < 0) {
            newRibs.setPosition(0);
        }
        List<TechnologySimulationsRibsEntity> existedRibs = getRibs(technologySimulationsId, false, currentUser);
        if (existedRibs.size() < newRibs.getPosition()) {
            existedRibs.add(newRibs);
        } else {
            existedRibs.add(newRibs.getPosition(), newRibs);
        }
        setRibsForTechnologySimulations(technologySimulationsId, existedRibs, currentUser);
    }

    public void validateRibStep(TechnologySimulationsRibsEntity newRib, UserEntity currentUser) {
        Arrays.stream(newRib.getSteps().split(","))
                .forEach(s->{
                    try {
                    Double.parseDouble(s);
                } catch (NumberFormatException e) {
                        throw new BadRequestException("Шаг должен иметь численный формат");
                }
                });
    }

    public void deleteRibs(int technologySimulationsId, int ribsId, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologySimulationsId, AccessRights.update, currentUser);
        List<TechnologySimulationsRibsEntity> existedRibs = getRibs(technologySimulationsId, false, currentUser);
        existedRibs.removeIf(ribs -> ribs.getId() == ribsId);
        setRibsForTechnologySimulations(technologySimulationsId, existedRibs, currentUser);
    }

    public void setRibsForTechnologySimulations(int technologySimulationsId, List<TechnologySimulationsRibsEntity> newRibs, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologySimulationsId, AccessRights.update, currentUser);
        newRibs.sort(Comparator.comparing(TechnologySimulationsRibsEntity::getPosition));
        technologySimulationsRibsDao.deleteTechnologySimulationsRibs(technologySimulationsId);
        if (newRibs.isEmpty()) {
            return;
        }
        for (int i = 0; i < newRibs.size(); ++i) {
            newRibs.get(i).setPosition(i);
            newRibs.get(i).setTechnologySimulationId(technologySimulationsId);
        }
        technologySimulationsRibsDao.batchInsertTechnologySimulationsRibs(newRibs);
    }

    public void updateRibsOrder(int technologySimulationsId, List<Integer> newOrder, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologySimulationsId, AccessRights.update, currentUser);
        Map<Integer, TechnologySimulationsRibsEntity> existedRibs = getRibs(technologySimulationsId, false, currentUser).stream()
                .collect(toMap(TechnologySimulationsRibsEntity::getId, Function.identity()));
        if (existedRibs.size() != newOrder.size()) {
            throw new BadRequestException("Неверное количество ребер");
        }
        List<TechnologySimulationsRibsEntity> orderedRibs = new ArrayList<>(existedRibs.size());
        newOrder.forEach(id -> {
            TechnologySimulationsRibsEntity technologySimulationsRibsEntity = existedRibs.get(id);
            if (technologySimulationsRibsEntity == null) {
                throw new BadRequestException("Несуществующий идентификатор ребра");
            }
            orderedRibs.add(technologySimulationsRibsEntity);
        });
        setRibsForTechnologySimulations(technologySimulationsId, orderedRibs, currentUser);
    }

    public TSParameters getTechnologySimulationsParameters(Integer technologySimulationsId) {
        return technologySimulationsDao.getTechnologySimulationsParameters(technologySimulationsId);
    }

    public TechnologySimulationsEntity copyTechnologySimulation (Integer referenceTechnologySimulationId, String name, UserEntity user) {
        TechnologySimulationsEntity existed = super.findOne(referenceTechnologySimulationId, user);
        existed.setName(name);
        TechnologySimulationsEntity newTS = technologySimulationsDao.insertAndRead(existed);
        List<TechnologySimulationsRibsEntity> ribs = getRibs(referenceTechnologySimulationId, false, user);
        ribs.forEach(rib->{
            rib.setTechnologySimulationId(newTS.getId());
            addRibs(newTS.getId(),rib,user);
        });
        List<TechnologySimulationsParamEntity> parameters = getTSParameters(referenceTechnologySimulationId);
        parameters.forEach(par->{
            par.setTechnologySimulationId(newTS.getId());
            technologySimulationsParamDao.insert(par);
        });
        return newTS;
    }



    //   @Override
    public FileManager fileManager() {
        return fileManager;
    }


}

