package ru.crp.lvms.managers.misc;

import com.fasterxml.classmate.TypeResolver;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.entities.UserEntity;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.List;
import java.util.Map;

import static java.util.stream.Collectors.toList;

@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public abstract class BaseManager<E extends Entity> {
    protected Class<E> entityClass;

    protected abstract BaseSpaceDao<E> aliveDao();


    @PostConstruct
    public void init() {
        entityClass = (Class<E>) new TypeResolver()
                .resolve(getClass()).typeParametersFor(BaseManager.class).get(0).getErasedType();
        Assert.notNull(entityClass);
    }


    public void delete(int id, @Nullable UserEntity currentUser) {
        E entity = findOne(id, currentUser);
        if (entity == null) {
            throw new NotFoundException(id, entityClass);
        }
        delete(entity, currentUser);
    }


    public void delete(E entity, @Nullable UserEntity currentUser) {
        checkDelete(entity, currentUser);
        aliveDao().delete(entity.getId());
    }


    @Nullable
    public E findOne(int id, @Nullable UserEntity currentUser) {
        E one = aliveDao().findOne(id);
        checkView(one, currentUser);
        return one;
    }


    @Nonnull
    public List<E> findAll(@Nullable UserEntity currentUser) {
        return aliveDao().findAll().stream()
                .filter(e -> filterEntitiesList(e, currentUser))
                .collect(toList());
    }

    private boolean filterEntitiesList(E listEntry, UserEntity currentUser) {
        return true;
    }


    // doesn't check access rights. Because this can breake pagination
    @Nonnull
    public PageImpl<E> find(Map<String, Object> filters, Pageable pageable, @Nullable UserEntity currentUser) {
        return aliveDao().find(filters, pageable);
    }


    public void update(E entity, @Nullable UserEntity currentUser) {
        checkUpdate(entity, currentUser);
        aliveDao().update(entity);
    }


    @Nonnull
    public E updateAndRead(E entity, @Nullable UserEntity currentUser) {
        update(entity, currentUser);
        return aliveDao().findOne(entity.getId());
    }


    public int create(E entity, @Nullable UserEntity currentUser) {
        checkCreate(entity, currentUser);
        return aliveDao().insert(entity);
    }


    @Nonnull
    public E createAndRead(E entity, @Nullable UserEntity currentUser) {
        int id = create(entity, currentUser);
        return aliveDao().findOne(id);
    }


    public int count() {
        return aliveDao().count();
    }

    public void checkView(@Nullable E one, @Nullable UserEntity currentUser) {

    }

    public void checkDelete(E entity, @Nullable UserEntity currentUser) {

    }

    public void checkUpdate(E entity, @Nullable UserEntity currentUser) {

    }

    public void checkCreate(E entity, @Nullable UserEntity currentUser) {

    }

    public E checkEntityExistsAndAccesible(int entityId, AccessRights requestedRights, @Nullable UserEntity currentUser) {
        E entity = findOne(entityId, currentUser);
        if (entity == null) {
            throw new NotFoundException(entityId, entityClass);
        }
        if (requestedRights == AccessRights.view) {
            checkView(entity, currentUser);
        }
        if (requestedRights == AccessRights.update) {
            checkUpdate(entity, currentUser);
        }
        if (requestedRights == AccessRights.delete) {
            checkDelete(entity, currentUser);
        }
        return entity;
    }
}
