package ru.crp.lvms.managers.misc;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.model.dao.ColumnsPreferencesDao;
import ru.crp.lvms.model.entities.*;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.compositeMaterials.TechnologyCompositeMaterial;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.model.entities.technology.TechnologyParametersEntity;
import ru.crp.lvms.model.entities.technology.TechnologySimulationsEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationDependency;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationReportEntity;
import ru.crp.lvms.utils.GridUtils;
import ru.crp.lvms.web.dto.TechnologyMaterial;
import ru.crp.lvms.web.dto.columns.FieldMapping;
import ru.crp.lvms.web.dto.columns.GridColumnType;
import ru.crp.lvms.web.dto.user.UserOut;

import javax.annotation.PostConstruct;
import java.util.*;

import static ru.crp.lvms.utils.MessageUtils.PREFERENCES_COLUMN_NOT_FOUND;
import static ru.crp.lvms.utils.MessageUtils.PREFERENCES_NOT_FOUND;

@Component
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class PreferencesManager {
    private Map<Class<?>, String> tableMapping;

    @Autowired
    private ColumnsPreferencesDao columnsPreferencesDao;

    @Autowired
    private MessageSource msgSrc;

    @PostConstruct
    public void init() {
        tableMapping = new HashMap<>();
        tableMapping.put(FileEntity.class, "files");
        tableMapping.put(MaterialEntity.class, "materials");
        tableMapping.put(CompositeMaterialEntity.class, "composite_materials");
        tableMapping.put(ParametersEntityTable.class, "parameters");
        tableMapping.put(UserOut.class, "users");
        tableMapping.put(TechnologyEntity.class, "technologies");
        tableMapping.put(SimulationEntity.class, "simulations");
        tableMapping.put(ConstructionElementEntity.class, "construction_elements");
        tableMapping.put(PossibleDefectsEntity.class, "possible_defects");
        tableMapping.put(EquipmentEntity.class, "equipments");
        tableMapping.put(TechnologyParametersEntity.class, "equipments");
        tableMapping.put(TechnologyMaterial.class, "technology_materials");
        tableMapping.put(TechnologySimulationsEntity.class, "technology_simulations");
        tableMapping.put(OptimizationEntity.class, "optimizations");
        tableMapping.put(ResultEntity.class, "results");
        tableMapping.put(OptimizationReportEntity.class, "optimization_reports");
        tableMapping.put(ECEntity.class, "effective_characteristics");
        tableMapping.put(ECDefectEntity.class, "ec_defects");
        tableMapping.put(ECParamEntity.class, "ec_params");
        tableMapping.put(TaskEntity.class, "tasks");
        tableMapping.put(OptimizationDependency.class, "dependencies");
        tableMapping.put(SupportLetterEntity.class, "support");
        tableMapping.put(TechnologyCompositeMaterial.class, "technology_composite_material");
    }

    private String checkAndGetTableName(Class<?> uiModel) {
        String tableName = tableMapping.get(uiModel);
        if (tableName == null) {
            throw new NotFoundException(msgSrc.getMessage(PREFERENCES_NOT_FOUND,
                    new Object[] {uiModel.getSimpleName()}, null, null));
        }
        return tableName;
    }

    private List<FieldMapping> getCustomOrDefaultPrefs(Class<?> uiModel, List<ColumnPrefsEntity> customPrefs) {
        if (customPrefs.isEmpty()) {
            return GridUtils.getColumnsByClass(uiModel);
        } else {
            return fromEntities(customPrefs);
        }
    }

    @Transactional(readOnly = true)
    // returns preferences included pref for excluded fields
    public List<FieldMapping> getAllPrefsForTable(UserEntity user, Class<?> uiModel) {
        String tableName = checkAndGetTableName(uiModel);
        List<ColumnPrefsEntity> allColumns = columnsPreferencesDao.findAllColumns(user.getId(), tableName);
        List<FieldMapping> customOrDefaultPrefs = getCustomOrDefaultPrefs(uiModel, allColumns);
        if (user.isGuest()) {
            customOrDefaultPrefs.stream()
                    .filter(field -> "owner".equals(field.getName()))
                    .forEach(field -> field.setCell(GridColumnType.string));
        }
        return customOrDefaultPrefs;
    }


    @Transactional
    public void changeColumnPrefs(Integer userId, Class<?> uiModel, FieldMapping prefsToChange) {
        String tableName = checkAndGetTableName(uiModel);
        createCustomPrefsIfNeed(userId, uiModel);
        ColumnPrefsEntity entityToChange = columnsPreferencesDao.getPreferencesForColumn(userId, tableName,
                prefsToChange.getName());
        if (entityToChange == null) {
            throw new NotFoundException(msgSrc.getMessage(PREFERENCES_COLUMN_NOT_FOUND,
                    new Object[] {prefsToChange.getName()}, null, null));
        }
        changeColumnPrefsEntity(entityToChange, prefsToChange);

        columnsPreferencesDao.update(entityToChange);
    }

    private void changeColumnPrefsEntity(ColumnPrefsEntity entityToChange, FieldMapping changes) {
        entityToChange.setWidth(changes.getWidth());
        if (changes.getLabel() != null) {
            entityToChange.setLabel(changes.getLabel());
        }
        entityToChange.setIsExcluded(!changes.getRenderable());
    }

    // reset isExcluded flag from all columns for table that match <uiModel> param
    @Transactional
    public void restoreAllColumns(Integer userId, Class<?> uiModel) {
        String tableName = checkAndGetTableName(uiModel);
        createCustomPrefsIfNeed(userId, uiModel);
        columnsPreferencesDao.restoreColumns(userId, tableName);
    }

    private void createCustomPrefsIfNeed(Integer userId, Class<?> uiModel) {
        if (!hasCustomPrefs(userId, uiModel)) {
            String tableName = checkAndGetTableName(uiModel);
            List<FieldMapping> defaultPrefs = GridUtils.getColumnsByClass(uiModel);
            columnsPreferencesDao.createPreferences(toEntities(defaultPrefs, tableName, userId));
        }
    }

    public boolean hasCustomPrefs(Integer userId, Class<?> uiModel) {
        String tableName = checkAndGetTableName(uiModel);
        return columnsPreferencesDao.getPreferencesCount(userId, tableName) > 0;
    }

    // todo change to use Mapper
    private List<FieldMapping> fromEntities(List<ColumnPrefsEntity> entities) {
        List<FieldMapping> result = new ArrayList<>(entities.size());
        for (ColumnPrefsEntity entity : entities) {
            FieldMapping uiModel = new FieldMapping(entity);
            result.add(uiModel);
        }
        return result;
    }

    private List<ColumnPrefsEntity> toEntities(List<FieldMapping> uiModels, String tableName, int userId) {
        List<ColumnPrefsEntity> result = new ArrayList<>(uiModels.size());
        for (FieldMapping uiModel : uiModels) {
            ColumnPrefsEntity entity = new ColumnPrefsEntity();
            entity.setName(uiModel.getName());
            entity.setCell(uiModel.getCell());
            entity.setIsExcluded(!uiModel.getRenderable());
            entity.setLabel(uiModel.getLabel());
            entity.setOrder(uiModel.getOrder());
            entity.setWidth(uiModel.getWidth());
            entity.setTableName(tableName);
            entity.setUserId(userId);
            result.add(entity);
        }
        return result;
    }

    @Transactional
    public void changeColumnsOrder(Integer userId, Class<?> uiModel, List<String> columnNames) {
        createCustomPrefsIfNeed(userId, uiModel);

        String tableName = checkAndGetTableName(uiModel);
        List<ColumnPrefsEntity> allPrefs = columnsPreferencesDao.findAllColumns(userId, tableName);

        Set<ColumnPrefsEntity> orderedPrefs = new LinkedHashSet<>(allPrefs.size());
        int order = 0;
        for (String colName : columnNames) {
            ColumnPrefsEntity prefs = getPrefsByColumnName(allPrefs, colName);
            prefs.setOrder(order++);
            orderedPrefs.add(prefs);
        }
        for (ColumnPrefsEntity en : allPrefs) {
            if (!orderedPrefs.contains(en)) {
                en.setOrder(order++);
                en.setIsExcluded(true);
                orderedPrefs.add(en);
            }
        }
        columnsPreferencesDao.deletePreferences(userId, tableName);
        columnsPreferencesDao.createPreferences(orderedPrefs);
    }

    private ColumnPrefsEntity getPrefsByColumnName(List<ColumnPrefsEntity> allPrefs, String colName) {
        for (ColumnPrefsEntity entity : allPrefs) {
            if (entity.getName().equals(colName)) {
                return entity;
            }
        }
        throw new NotFoundException(msgSrc.getMessage(PREFERENCES_COLUMN_NOT_FOUND, new Object[] {colName}, null, null));
    }
}
