package ru.crp.lvms.managers.tasks;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.dao.TaskDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.SlurmTaskInfo;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tasks.TaskIn;
import ru.crp.lvms.model.entities.tasks.TaskParent;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationType;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.StartTasksJob;
import ru.crp.lvms.tasks.slurm.SlurmTaskInfoDao;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.validators.task.CompositeTaskParamsValidator;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

import static ru.crp.cmls.api.TaskParentType.optimization;

@Service
public class TaskManager extends BaseManagerSuggest<TaskEntity> {
    private static final Logger LOGGER = LoggerFactory.getLogger(TaskManager.class);

    @Autowired
    private TaskDao taskDao;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private ECManager ecManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private CompositeTaskParamsValidator taskParamsValidator;

    @Autowired
    private SlurmTaskInfoDao slurmTaskInfoDao;

    @Autowired
    private TaskResultsExtractor taskResultsExtractor;

    @Autowired
    private ClustersManager clustersManager;

    @Autowired
    private ECTaskManager ecTaskManager;

    @Autowired
    private SimulationTaskManager simulationTaskManager;

    @Autowired
    private OptimizationTaskManager optimizationTaskManager;

    @Autowired
    private TaskStatusUpdater taskStatusUpdater;

    @Autowired
    private TaskLogsExtractor taskLogsExtractor;


    private Lock refreshResultsLock = new ReentrantLock();

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private StartTasksJob startTasksJob;

    @Autowired
    private TaskCurrentResultUpdater taskCurrentResultUpdater;

    @Override
    protected TaskDao aliveDao() {
        return taskDao;
    }

    @Nonnull
    @Override
    public TaskEntity createAndRead(TaskEntity entity, @Nullable UserEntity currentUser) {
        throw new IllegalStateException("Use method with slurm task info parameter");
    }

    @Nonnull
    public TaskEntity createAndRead(TaskEntity taskToCreate, @Nullable SlurmTaskInfo slurmTaskInfo,
                                    @Nullable UserEntity currentUser) {
        Errors errors = new BeanPropertyBindingResult(taskToCreate, taskToCreate.getClass().getName());
        taskParamsValidator.validate(taskToCreate, errors);
        if (errors.hasErrors()) {
            throw new TaskStartingException("Не удалось добавить задачу в очередь:", errors);
        }
        TaskEntity newTask = super.createAndRead(taskToCreate, currentUser);
        if (slurmTaskInfo != null) {
            slurmTaskInfo.setTaskId(newTask.getId());
            slurmTaskInfoDao.createSlurmTaskInfo(slurmTaskInfo);
        }
        return newTask;
    }

    public List<TaskEntity> findRunning() {
        return aliveDao().findRunning();
    }

    public TaskEntity findRunning(ResultsParent resultsParent) {
        return aliveDao().findRunning(resultsParent);
    }

    public PageImpl<TaskEntity> findRunning(Map<String, Object> filters, Pageable pageable) {
        return aliveDao().findRunning(pageable);
    }

    public PageImpl<TaskEntity> getFinishedTasksPage(Map<String, Object> filters, Pageable pageable) {
        return taskDao.getFinishedTasksPage(filters, pageable);
    }

    public List<TaskEntity> getOnlyEditableTasksFromQueue(UserEntity currentUser) {
        if (currentUser.isAdmin()) {
            return taskDao.getWholeTaskQueue();
        } else {
            return taskDao.getUserTaskQueue(currentUser.getId());
        }
    }

    public PageImpl<TaskEntity> getTaskQueue(Map<String, Object> filters, Pageable pageable) {
        return taskDao.getTaskQueue(filters, pageable);
    }

    public void swapTasks(TaskEntity firstTask, TaskEntity secondTask, UserEntity currentUser) {
        if (firstTask.getQueuePriority() != secondTask.getQueuePriority()) {
            throw new BadRequestException("Невозможно поменять местами задачи с различными типами");
        }
        if (!isWaiting(firstTask) || !isWaiting(secondTask)) {
            throw new BadRequestException("Можно менять местами только ожидающие задачи");
        }
        if (!canChangeTasksPriority(firstTask, secondTask, currentUser)) {
            throw new AccessDeniedException("Недостаточно прав для совершения данного действия");
        }
        taskDao.swapTasks(firstTask, secondTask);
    }

    public boolean isWaiting(TaskEntity task) {
        return task.getStatusId() == CalculationStatus.Waiting.getId();
    }

    private boolean canChangeTasksPriority(TaskEntity firstTask, TaskEntity secondTask, UserEntity currentUser) {
        return currentUser.isAdmin()
                || (firstTask.getOwnerId() == currentUser.getId() && secondTask.getOwnerId() == currentUser.getId());
    }


    public TaskEntity findRunningOnNode(int nodeId) {
        return taskDao.findRunningOnNode(nodeId);
    }

    public boolean isNodeFree(int nodeId) {
        return taskDao.isNodeFree(nodeId);
    }

    private List<TaskEntity> getLastSuccessfulChainOfTasks(ResultsParent resultsParent) {
        return taskDao.getLastSuccessfulChainOfTasks(resultsParent);
    }

    public void refreshResultsHandling(ResultsParent resultsParent, UserEntity user) {

        TaskParent parent;
        switch (resultsParent.getType()) {
            case effectiveCharacteristic:
                parent = ecManager.findOne(resultsParent.getId(), user);
                break;
            case simulation:
                parent = simulationManager.findOne(resultsParent.getId(), user);
                break;
            case optimization:
                parent = optimizationManager.findOne(resultsParent.getId(), user);
                break;
            default:
                throw new IllegalArgumentException("Unexpected parent type: " + resultsParent.getType());
        }
        if (parent.getStatus() != SolvingStatus.Completed) {
            throw new BadRequestException("Успешно завершенные задачи не найдены или задача запущена на рассчет");
        }

        if (refreshResultsLock.tryLock()) {
            try {
                resultsManager.deleteResults(resultsParent, false, CreatedBy.Auto);
                List<TaskEntity> lastSuccessfulTasksChain = getLastSuccessfulChainOfTasks(resultsParent);
                if (lastSuccessfulTasksChain.isEmpty()) {
                    throw new NotFoundException("Успешно завершенные задачи не найдены");
                }
                lastSuccessfulTasksChain.forEach(task -> taskResultsExtractor.extractResults(task));
            } finally {
                refreshResultsLock.unlock();
            }
        } else {
            throw new BadRequestException("Результаты уже обновляются");
        }
    }

    public void downloadCurrentTaskResults(ResultsParent resultsParent, UserEntity user) throws Exception {

        TaskParent parent;
        switch (resultsParent.getType()) {
            case effectiveCharacteristic:
                parent = ecManager.findOne(resultsParent.getId(), user);
                break;
            case simulation:
                parent = simulationManager.findOne(resultsParent.getId(), user);
                break;
            case optimization:
                parent = optimizationManager.findOne(resultsParent.getId(), user);
                break;
            default:
                throw new IllegalArgumentException("Unexpected parent type: " + resultsParent.getType());
        }
        if (parent.getStatus() != SolvingStatus.Processing) {
            throw new BadRequestException("Выполняющиеся расчеты не найдены");
        }

        if (refreshResultsLock.tryLock()) {
            try {
                TaskEntity runningTask = findRunning(resultsParent);
                if (runningTask == null) {
                    throw new NotFoundException("Выполняющиеся расчеты не найдены");
                }
                if (runningTask.getType() == Type.optistructGenerateFEM ||
                        runningTask.getType() == Type.toscaSmoothing ||
                        runningTask.getType() == Type.toscaAnimation) {
                    throw new NotFoundException("Оптимизационный расчет не выполняется");
                }
                taskCurrentResultUpdater.updateResults(runningTask);
            } finally {
                refreshResultsLock.unlock();
            }
        } else {
            throw new BadRequestException("Результаты уже обновляются");
        }
    }

    public void extractLogs(ResultsParent resultsParent, UserEntity user) throws Exception {

        TaskParent parent;
        switch (resultsParent.getType()) {
            case effectiveCharacteristic:
                parent = ecManager.findOne(resultsParent.getId(), user);
                break;
            case simulation:
                parent = simulationManager.findOne(resultsParent.getId(), user);
                break;
            case optimization:
                parent = optimizationManager.findOne(resultsParent.getId(), user);
                break;
            default:
                throw new IllegalArgumentException("Unexpected parent type: " + resultsParent.getType());
        }
        if (parent.getStatus() == SolvingStatus.Created) {
            throw new BadRequestException("Для данной задачи расчеты ещё не проводились");
        }

        if (refreshResultsLock.tryLock()) {
            try {
                TaskEntity task = taskDao.getLastPerformedTask(resultsParent);
                if (task == null) {
                    throw new NotFoundException("Для данной задачи расчеты ещё не проводились");
                }
                taskLogsExtractor.extractLogs(task);
            } finally {
                refreshResultsLock.unlock();
            }
        } else {
            throw new BadRequestException("Результаты уже обновляются");
        }
    }

    @Transactional//(propagation = REQUIRES_NEW)
    public TaskEntity createChainOfTasks(TaskIn taskIn, UserEntity user) {
        TaskEntity newTask = TaskEntity.fromDTO(taskIn, user);
        TaskParentType parentType = newTask.getParentType();
        List<TaskEntity> chain;
        switch (parentType) {
            case effectiveCharacteristic:
                chain = ecTaskManager.createChainOfTasksForEC(newTask);
                break;
            case simulation:
                chain = simulationTaskManager.createChainOfTasksForSimulation(newTask, user);
                break;
            case optimization:
                chain = optimizationTaskManager.createChainOfTasksForOptimization(newTask, user);
                break;
            default:
                throw new IllegalArgumentException("Unknown parent type: " + parentType.name());
        }

        List<TaskEntity> savedTasks = chain.stream().map(task -> {
            SlurmTaskInfo slurmTaskInfo = null;
            ClusterEntity cluster = clustersManager.getClusterForTask(task);
            ClusterType clusterType = ClusterType.valueOf(cluster.getTypeId());
            checkClusterIsAvailableForTask(cluster, task, user);
            if (clusterType == ClusterType.Slurm) {
                slurmTaskInfo = Mapper.map(taskIn, SlurmTaskInfo.class);
            }
            TaskEntity savedTask = createAndRead(task, slurmTaskInfo, user);
            taskStatusUpdater.changeTaskStatus(savedTask, true);
            return savedTask;
        }).collect(Collectors.toList());

        if (savedTasks.size() > 1) {
            setTasksOrdering(savedTasks, user);
        }

        startTasksJob.checkTaskQueueManually();
        return savedTasks.get(0);
    }

    private void setTasksOrdering(List<TaskEntity> savedTasks, UserEntity user) {
        Integer prevId = null;
        for (TaskEntity cur : savedTasks) {
            cur.setPrevTaskId(prevId);
            TaskEntity read = updateAndRead(cur, user);
            prevId = read.getId();
        }
    }

    private void checkClusterIsAvailableForTask(ClusterEntity cluster, TaskEntity newTask, UserEntity currentUser) {
        ClusterType clusterType = ClusterType.valueOf(cluster.getTypeId());
        // for now only optimization with type = 'engineering calculation' and cae = ANSYS\LS-DYNA can be solved only on slurm cluster
        // all other tasks can be solved only on Space clusters
        if (newTask.getParentType() == optimization) {
            OptimizationEntity parentOptimization = optimizationManager.findOne(newTask.getParentId(), currentUser);
            if (parentOptimization.getOptimizationTypeId() == OptimizationType.ENGINEERING_CALCULATION_TYPE_ID) {
//                if (clusterType != ClusterType.Slurm) {
//                    throw new BadRequestException(String.format("Инженерные расчеты не могут быть запущены на кластере '%s'", cluster.getName()));
//                }
            } else {
                if (clusterType == ClusterType.Slurm) {
                    throw new BadRequestException(String.format("Только инженерные расчеты могут быть запущены на кластере '%s'", cluster.getName()));
                }
            }
        } else {
            if (clusterType == ClusterType.Slurm) {
                throw new BadRequestException(String.format("Только инженерные расчеты могут быть запущены на кластере '%s'", cluster.getName()));
            }
        }
    }
}
