package ru.crp.lvms.managers.tasks;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.events.TaskStateChangedEvent;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.model.dao.TaskDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;

import javax.annotation.Nullable;
import java.util.Date;
import java.util.Optional;

@Service
@Transactional
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class TaskStatusUpdater {
    @Autowired
    private TaskDao taskDao;

    @Autowired
    private ECManager ecManager;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ApplicationEventPublisher applicationEventPublisher;

    public TaskEntity changeTaskStatus(TaskEntity task, CalculationStatus newStatus) {
        return changeTaskStatus(task, newStatus, null);
    }

    public TaskEntity changeTaskStatus(TaskEntity task, CalculationStatus newStatus, @Nullable UserEntity currentUser) {
        task.setStatusId(newStatus.getId());
        task.setStatus(newStatus.getDisplayName());
        return changeTaskStatus(task, false);
    }

    public TaskEntity changeTaskStatus(TaskEntity task, boolean onlyParent) {
        // changing task status
        TaskEntity oldState = task;
        TaskEntity newState = task;
        if (!onlyParent) {
            // status and progress may left the same but we need to change modification date to know
            // that node is still alive
            task.setModificationDate(new Date());
            newState = taskDao.updateAndRead(task);
        }

        // changing parent status
        CalculationStatus taskStatus = CalculationStatus.valueOf(task.getStatusId());
        if (task.getParentType() == TaskParentType.effectiveCharacteristic) {
            ECEntity parent = ecManager.findOne(task.getParentId(), null);
            parent.setStatusId(SolvingStatus.fromCalculationStatus(taskStatus));
            ecManager.update(parent, null);
        } else if (task.getParentType() == TaskParentType.simulation) {
            SimulationEntity parent = simulationManager.findOne(task.getParentId(), null);
            parent.setStatusId(SolvingStatus.fromCalculationStatus(taskStatus));
            simulationManager.update(parent, null);
        } else if (task.getParentType() == TaskParentType.optimization) {
            OptimizationEntity parent = optimizationManager.findOne(task.getParentId(), null);
            parent.setSolvingStatusId(SolvingStatus.fromCalculationStatus(taskStatus));
            optimizationManager.update(parent, null);
        } else {
            throw new IllegalArgumentException("Недопустимый параметр: " + task.getParentType());
        }
        if (!onlyParent) {
            applicationEventPublisher.publishEvent(new TaskStateChangedEvent(oldState, newState));
        }
        return newState;
    }

    public void setCalculationFinishedWithException(TaskEntity task) {
        task.setCalculationFinishedWithException(true);
        taskDao.setCalculationFinishedWithException(task);
    }

    public void update(TaskEntity task) {
        taskDao.update(task);
    }

    public Optional<TaskEntity> findNextTask(int taskId) {
        return taskDao.findNextTask(taskId);
    }
}
