package ru.crp.lvms.model.dao;

import org.springframework.stereotype.Repository;
import ru.crp.lvms.model.entities.ParameterSample;
import ru.crp.lvms.model.entities.ParameterValues;
import ru.crp.lvms.model.mappers.ParameterValuesRowMapper;

import javax.annotation.PostConstruct;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import static ru.crp.lvms.managers.ParametersManager.DEFAULT_TEMPERATURE;
import static ru.crp.lvms.web.dto.DefaultParameterValue.extractValueForTemp;

@Repository
public class ParameterValuesDao extends BaseSpaceDao<ParameterValues> {
    private static final String SELECT_SQL = "SELECT * FROM parameters_materials_join";
    private static final String SELECT_ONE_SQL = SELECT_SQL + " WHERE id = ?";
    private static final String DELETE_PARAMETER_VALUES_FOR_MATERIAL_SQL = "DELETE FROM parameters_materials WHERE "
            + "material_id = ? AND parameter_id = ?";
    private static final String DELETE_ALL_PARAMETER_FILES_FOR_MATERIAL_SQL = "DELETE FROM parameter_files_materials WHERE "
            + "composite_material_id = ?";
    private static final String DELETE_ALL_PARAMETERS_FOR_MATERIAL_SQL = "DELETE FROM parameters_materials WHERE "
            + "material_id = ?";
    private static final String BATCH_INSERT_MATERIAL_PARAMETER_VALUES_SQL = "INSERT INTO parameters_materials "
            + "(value, func_value, sim_value, material_id, parameter_id, temp) VALUES (?, ?, ?, ?, ?, ?)";
    private static final String UPSERT_PARAM_SQL = "WITH updated AS (UPDATE parameters_materials SET "
            + "value = ?, func_value = ?, sim_value = ?"
            + " WHERE material_id = ? AND parameter_id = ? AND temp = ? RETURNING id) "
            + "INSERT INTO parameters_materials(value, func_value, sim_value, material_id, parameter_id, temp) "
            + "SELECT ?, ?, ?, ?, ?, ? WHERE NOT EXISTS(SELECT 1 FROM updated)";

    private static final String POINTS_COUNT_SQL = "SELECT COUNT(*) FROM parameters_materials WHERE id = ?";
    private static final String DELETE_POINT_SQL = "DELETE FROM parameters_materials WHERE id = ?";
    private static final String UPDATE_POINT_SQL = "UPDATE parameters_materials SET "
            + "value = ?, func_value = ?, sim_value = ? WHERE id = ?";

    @PostConstruct
    public void init() {
        setRowMapper(new ParameterValuesRowMapper());
        setFindSql(SELECT_SQL);
        setFindAllSql(SELECT_SQL);
        setFindOneSql(SELECT_ONE_SQL);
    }


    public List<ParameterValues> findByMaterialId(int materialId) {
        return find("WHERE material_id = ?", materialId);
    }


    public ParameterValues findByMaterialAndParameterIds(int materialId, int parameterId) {
        return findOne("WHERE material_id = ? AND id = ?", materialId, parameterId);
    }


    public BigDecimal findByMaterialAndParameterIdsWithDefaultTemperature(int materialId, int parameterId) {
        return extractValueForTemp(DEFAULT_TEMPERATURE,
                        findOne("WHERE material_id = ? AND id = ?", materialId, parameterId));
    }


    public void deleteAllParametersForMaterial(int materialId) {
        getJdbcTemplate().update(DELETE_ALL_PARAMETERS_FOR_MATERIAL_SQL, materialId);
        getJdbcTemplate().update(DELETE_ALL_PARAMETER_FILES_FOR_MATERIAL_SQL, materialId);
    }


    public void deleteParameterValuesForMaterial(int materialId, int parameterId) {
        getJdbcTemplate().update(DELETE_PARAMETER_VALUES_FOR_MATERIAL_SQL, materialId, parameterId);
    }


    public void insertParameterValuesForMaterial(int materialId, int parameterId,
                                                 List<ParameterSample> newParameterValues) {
        if (newParameterValues.isEmpty()) {
            return;
        }
        List<Object[]> args = new ArrayList<>();
        newParameterValues.forEach(parameterValue -> args.add(new Object[]{
                parameterValue.getParamValue(),
                parameterValue.getFuncParamValue(),
                parameterValue.getSimParamValue(),
                materialId,
                parameterId,
                parameterValue.getTemperature(),
        }));
        getJdbcTemplate().batchUpdate(BATCH_INSERT_MATERIAL_PARAMETER_VALUES_SQL, args);
    }

    public void upsertParameterValue(int materialId, int parameterId, ParameterSample newParameterValue) {
        BigDecimal temp = newParameterValue.getTemperature();
        BigDecimal paramValue = newParameterValue.getParamValue();
        BigDecimal funcParamValue = newParameterValue.getFuncParamValue();
        BigDecimal simParamValue = newParameterValue.getSimParamValue();
        getJdbcTemplate().update(UPSERT_PARAM_SQL,
                paramValue, funcParamValue, simParamValue, materialId, parameterId, temp,
                paramValue, funcParamValue, simParamValue, materialId, parameterId, temp);
    }


    public boolean pointExists(int pointId) {
        return !getJdbcTemplate()
                .queryForObject(POINTS_COUNT_SQL, new Object[]{pointId}, Integer.class).equals(0);
    }


    public void deletePoint(int pointId) {
        getJdbcTemplate().update(DELETE_POINT_SQL, pointId);
    }


    public void updatePoint(ParameterSample newParameterValue) {
        getJdbcTemplate().update(UPDATE_POINT_SQL, newParameterValue.getParamValue(),
                newParameterValue.getFuncParamValue(),
                newParameterValue.getSimParamValue(), newParameterValue.getId());
    }
}
