package ru.crp.lvms.model.dao;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Primary;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmlb.model.utils.EntityUtils;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;

import java.util.*;

import static java.util.stream.Collectors.toList;

@Transactional(propagation = Propagation.REQUIRED)
@Repository
@Primary
public class TaskDao extends BaseSpaceDao<TaskEntity> {
    private static final String TASK_QUEUE_ORDERING_SQL = "ORDER BY queue_priority, task_position";

    @Autowired
    private NodesDao nodesDao;

    @Autowired
    private ClustersManager clustersManager;

    public List<TaskEntity> findRunning() {
        Set<CalculationStatus> workingStatuses = CalculationStatus.getWorkingStatuses();
        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        return find("where status_id IN " + INSql + " ", params.toArray());
    }

    public TaskEntity findRunning(ResultsParent resultsParent) {
        return findOne("WHERE status_id = ? AND  " + resultsParent.getType().getColName() + " = ? ",
                CalculationStatus.Solving.getId(),resultsParent.getId());
    }

    public PageImpl<TaskEntity> findRunning(Pageable pageable) {
        Set<CalculationStatus> workingStatuses = CalculationStatus.getWorkingStatuses();
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        return find("where status_id IN " + INSql + " ", pageable, params.toArray());
    }

    public List<TaskEntity> getWholeTaskQueue() {
        return find("WHERE status_id = ? " + TASK_QUEUE_ORDERING_SQL, CalculationStatus.Waiting.getId());
    }

    public List<TaskEntity> getUserTaskQueue(int currentUserId) {
        return find("WHERE owner_id = ? AND  status_id = ? " + TASK_QUEUE_ORDERING_SQL, currentUserId, CalculationStatus.Waiting.getId());
    }

    public PageImpl<TaskEntity> getTaskQueue(Map<String, Object> filters, Pageable pageable) {
        Sort sort = new Sort(
                new Sort.Order(Sort.Direction.ASC, "queue_priority"),
                new Sort.Order(Sort.Direction.ASC, "task_position"));
        PageRequest pageRequest = new PageRequest(pageable.getPageNumber(), pageable.getPageSize(), sort);

        return find("WHERE status_id = ?", filters, pageRequest, CalculationStatus.Waiting.getId());
    }

    public void swapTasks(TaskEntity firstTask, TaskEntity secondTask) {
        int firstTaskId = firstTask.getId();
        Integer firstTaskPos = firstTask.getTaskPosition();
        int secondTaskId = secondTask.getId();
        Integer secondTaskPos = secondTask.getTaskPosition();
        getJdbcTemplate().update("UPDATE tasks SET task_position =" +
                "(CASE WHEN id = ? THEN ?" +
                " WHEN id = ? THEN ? END)" +
                "WHERE id IN (?, ?)", firstTaskId, secondTaskPos, secondTaskId, firstTaskPos, firstTaskId, secondTaskId);
    }

    public TaskEntity findRunningOnNode(int nodeId) {
        Set<CalculationStatus> workingStatuses = CalculationStatus.getWorkingStatuses();
        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        NodeEntity nodeToCheck = nodesDao.findOne(nodeId);
        ClusterType nodeClusterType = clustersManager.getClusterType(nodeToCheck.getClusterId());

        List<TaskEntity> taskEntities;
        if (nodeClusterType == ClusterType.Slurm) {
            params.add(nodeToCheck.getName());
            taskEntities = find("JOIN slurm_tasks_info sti " +
                    "ON sti.task_id = " + getAlias() + ".id " +
                    "WHERE " + getAlias() + ".status_id IN " + INSql + " " +
                    "AND ? = ANY(sti.nodes)", params.toArray());
        } else {
            params.add(0, nodeId);
            taskEntities = find("where node_id=? AND status_id IN " + INSql, params.toArray());
        }
        if (taskEntities.size() > 1) {
            logger.error("More than 1 running task on node with id: " + nodeId);
        }
        return taskEntities.isEmpty() ? null : taskEntities.get(0);
    }

    public boolean isNodeFree(int nodeId) {
        return findRunningOnNode(nodeId) == null;
    }

    public List<TaskEntity> findTasksForCheckingStatus() {
        Set<CalculationStatus> workingStatuses = CalculationStatus.getWorkingStatuses();
        // we don't need to check task status if it's not in solver
        // after successful transferring all task data task will change it's state to Starting
        workingStatuses.remove(CalculationStatus.Sending);
        workingStatuses.remove(CalculationStatus.ImportingResults);

        // slurm handling tasks queue by itself so we need to check waiting tasks too
        workingStatuses.add(CalculationStatus.Waiting);

        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        return find("where status_id IN " + INSql + " ", params.toArray());

    }

    public List<TaskEntity> findComplete() {
        Set<CalculationStatus> workingStatuses = EnumSet.of(CalculationStatus.CalculationFinished);
        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        return find("WHERE status_id IN " + INSql, params.toArray());
    }

    public PageImpl<TaskEntity> getFinishedTasksPage(Map<String, Object> filters, Pageable pageable) {
        Set<CalculationStatus> terminalStatuses = CalculationStatus.getTerminalStatuses();
        List<Object> params = new ArrayList<>(terminalStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        String INSql = EntityUtils.sqlForListParams(terminalStatuses);
        return find("WHERE status_id IN " + INSql, filters, pageable, params.toArray());
    }

    public void setCalculationFinishedWithException(TaskEntity task) {
        getJdbcTemplate().update("UPDATE tasks SET calculation_finished_with_exception = TRUE WHERE id = ?", task.getId());
    }

    public List<TaskEntity> getLastSuccessfulChainOfTasks(ResultsParent resultsParent) {
        String request = "WITH RECURSIVE r(id) AS (" +
                "  SELECT (SELECT tj1.id" +
                "          FROM tasks_join tj1" +
                "          WHERE tj1.prev_task_id IS NULL" +
                "                AND " + resultsParent.getType().getColName() + " = ?" +
                "                AND status_id = ?" +
                "          ORDER BY date_modified DESC" +
                "          LIMIT 1) AS chain_head" +
                "  UNION" +
                "  SELECT tj2.id" +
                "  FROM tasks_join tj2, r" +
                "  WHERE r.id = tj2.prev_task_id" +
                " )" +
                " SELECT *" +
                " FROM tasks_join" +
                " WHERE id IN (SELECT id FROM r);";
        return jdbcTemplate().query(request, getRowMapper(), resultsParent.getId(), CalculationStatus.Finished.getId());
    }

    public TaskEntity getLastPerformedTask(ResultsParent resultsParent) {
        String request = "SELECT *" +
                "          FROM tasks_join " +
                "          WHERE  " + resultsParent.getType().getColName() + " = ?" +
                "                AND status_id NOT IN (?,?)" +
                "          ORDER BY date_modified DESC" +
                "          LIMIT 1;";
        return jdbcTemplate().query(request, getRowMapper(), resultsParent.getId(),
                CalculationStatus.Draft.getId(),CalculationStatus.Waiting.getId()).get(0);
    }

    public Optional<TaskEntity> findNextTask(int taskId) {
        return Optional.ofNullable(findOne("WHERE prev_task_id = ?", taskId));
    }
}
