package ru.crp.lvms.optimizations.base.controllers;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationFlow;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.web.controllers.misc.ControllerImageNew;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.dto.OptimizationsGraphEntry;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;
import ru.crp.lvms.web.dto.pagination.SortingOrder;

import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import static ru.crp.lvms.utils.URLs.OPTIMIZATIONS_URL;

@RestController
@RequestMapping(value = OPTIMIZATIONS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptimizationController extends CrudControllerSuggest<OptimizationEntity, OptimizationEntity, OptimizationEntity> implements ControllerImageNew<OptimizationEntity> {

    @Autowired
    private EventManager eventManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private TaskManager taskManager;


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.APPLICATION_JSON_VALUE)
    public OptimizationEntity createOptimization(HttpServletRequest request,
                                                 @RequestBody @Valid OptimizationEntity newEntity,
                                                 @CurrentUser PortalUser currentUser) {
        OptimizationEntity newOptimization = optimizationManager.createOptimization(newEntity, currentUser.getUser());
        extendOptimizationIfNeed(newOptimization, currentUser);
        return newOptimization;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public OptimizationEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public OptimizationEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                            @Valid @RequestBody OptimizationEntity updateVO,
                                            @CurrentUser PortalUser user) {
        return super.update(request,id,updateVO,user);
    }
    @RequestMapping(value = "/{referenceId}/copy",method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public OptimizationEntity cloneOptimization( @RequestParam("name") String name,
                                                 @RequestParam("isChild") String isChild,
                                                 @PathVariable("referenceId") Integer referenceId,
                                                 @CurrentUser PortalUser currentUser) throws CloneNotSupportedException {
        OptimizationEntity referenceOptimization = optimizationManager.findOne(referenceId, currentUser.getUser());
        OptimizationEntity newOptimization = optimizationManager.copyOptimization(name,referenceOptimization,isChild,currentUser.getUser());
        return newOptimization;
    }


    @NotNull
    @Override
    public List<OptimizationEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        List<OptimizationEntity> all = super.getAll(request, currentUser);
        all.forEach(optimization -> extendOptimizationIfNeed(optimization, currentUser));
        return all;
    }


    private void extendOptimizationIfNeed(OptimizationEntity optimization, PortalUser currentUser) {
        optimizationManager.extendOptimization(optimization, currentUser.getUser());
    }


    @NotNull
    @Override
    public PageingResponse<OptimizationEntity> page(HttpServletRequest request, @RequestBody DataRequest dataRequest,
                                                    @CurrentUser PortalUser currentUser) {
        PageingResponse<OptimizationEntity> page = super.page(request, dataRequest, currentUser);
        page.getContent().forEach(optimization -> extendOptimizationIfNeed(optimization, currentUser));
        return page;
    }


    @NotNull
    @Override
    public OptimizationEntity get(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = super.get(request, id, currentUser);
        extendOptimizationIfNeed(optimization, currentUser);
        return optimization;
    }


    @RequestMapping(value = "/{id}/siblings/list", method = RequestMethod.POST)
    public PageingResponse<OptimizationEntity> getSiblingsPage(@PathVariable("id") int optimizationId,
                                                               @RequestBody DataRequest dataRequest,
                                                               @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        dataRequest.getFilters().put("pid", optimization.getPid());
        PageImpl<OptimizationEntity> pageImpl = optimizationManager.find(dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(pageImpl.getContent(), dataRequest.getFilters(), pageable, pageImpl.getTotalElements());
    }


    @RequestMapping(value = "/{id}/graph", method = RequestMethod.GET)
    public List<OptimizationsGraphEntry> getOptimizationsGraph(@PathVariable("id") int optimizationId,
                                                               @CurrentUser PortalUser currentUser) {
        List<OptimizationsGraphEntry> graph = optimizationManager.getOptimizationsGraph(optimizationId, currentUser.getUser());
        if (graph == null) {
            throw new NotFoundException(optimizationId, OptimizationEntity.class);
        }
        return graph;
    }


    @PreAuthorize("hasRole('ROLE_GUEST')")
    @RequestMapping(value = "/{id}/flow", method = RequestMethod.GET)
    public OptimizationFlow getSimulationFlow(@PathVariable("id") int id,
                                              @CurrentUser PortalUser currentUser) throws IOException {
        return optimizationManager.getOptimizationFlow(id, currentUser.getUser());
    }


    @RequestMapping(value = "/{optimizationId}/tasks/list", method = RequestMethod.POST)
    @Transactional(readOnly = true)
    public PageingResponse<TaskEntity> getOptimizationTasks(@PathVariable("optimizationId") Integer optimizationId,
                                                            @RequestBody DataRequest dataRequest,
                                                            @CurrentUser PortalUser currentUser) {
        Sort sort = null;
        if (dataRequest.getSort() != null && !dataRequest.getSort().isEmpty()) {
            sort = new Sort(SortingOrder.toSpringSortabble(dataRequest.getSort()));
        }
        Pageable pageable = dataRequest.getPageable().toSpringPageable(sort);
        Map<String, Object> filters = dataRequest.getFilters();
        filters.put("parentId", optimizationId);
        filters.put("parentType", TaskParentType.optimization.toString());
        PageImpl<TaskEntity> response = taskManager.find(filters, pageable, currentUser.getUser());
        return new PageingResponse<>(response.getContent(), dataRequest.getFilters(), pageable, response.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_GUEST')")
    @RequestMapping(value = "/{id}/refreshResults", method = RequestMethod.GET)
    public void refreshResults(@PathVariable("id") int id,
                               @CurrentUser PortalUser currentUser) {
        ResultsParent resultsParent = Mapper.map(optimizationManager.findOne(id, currentUser.getUser()), ResultsParent.class);
        taskManager.refreshResultsHandling(resultsParent, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/downloadCurrentResults", method = RequestMethod.GET)
    public void downloadCurrentResults(@PathVariable("id") int id,
                               @CurrentUser PortalUser currentUser) throws Exception {
        ResultsParent resultsParent = Mapper.map(optimizationManager.findOne(id, currentUser.getUser()), ResultsParent.class);
        taskManager.downloadCurrentTaskResults(resultsParent, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/extractLogs", method = RequestMethod.GET)
    public void extractLogs(@PathVariable("id") int id,
                                       @CurrentUser PortalUser currentUser) throws Exception {
        ResultsParent resultsParent = Mapper.map(optimizationManager.findOne(id, currentUser.getUser()), ResultsParent.class);
        taskManager.extractLogs(resultsParent, currentUser.getUser());
    }


    @Override
    protected BaseManagerSuggest<OptimizationEntity> getManager() {
        return optimizationManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }


    @Override
    public ImageHandler imageHandler() {
        return optimizationManager;
    }


    @Override
    public OptimizationEntity getEntity(HttpServletRequest request, Integer entityId, PortalUser currentUser) {
        return get(request, entityId, currentUser);
    }


    @Override
    public FileManager getFileManager() {
        return fileManager;
    }
}
