package ru.crp.lvms.optimizations.base.controllers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.MaterialManager;
import ru.crp.lvms.managers.TechnologyManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationDependency;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationDependenciesManager;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import java.util.List;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.URLs.OPTIMIZATIONS_URL;

@RestController
@RequestMapping(value = OPTIMIZATIONS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptimizationDependenciesController {

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private OptimizationDependenciesManager optimizationDependenciesManager;

    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private FileManager fileManager;


    /**
     * TECHNOLOGIES
     **/
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/technologies", method = RequestMethod.POST)
    @ResponseStatus(HttpStatus.OK)
    public void addTechnology(@PathVariable("id") int optimizationId,
                              @RequestParam("entityId") int technologyId, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        TechnologyEntity technology = technologyManager
                .checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser.getUser());
        optimizationDependenciesManager.addTechnology(optimization, technology, currentUser.getUser());
    }

    @RequestMapping(value = "/{id}/technologies", method = RequestMethod.GET)
    public List<OptimizationDependency> getTechnologies(@PathVariable("id") int optimizationId,
                                                        @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        return optimizationDependenciesManager.getTechnologies(optimization, currentUser.getUser());
    }

    @RequestMapping(value = "/{id}/technologies/list", method = RequestMethod.POST)
    public PageingResponse<OptimizationDependency> getTechnologiesPage(@PathVariable("id") int optimizationId,
                                                                       @RequestBody DataRequest dataRequest,
                                                                       @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        PageImpl<OptimizationDependency> pageImpl = optimizationDependenciesManager
                .getTechnologiesPage(optimization, pageable, currentUser.getUser());
        return new PageingResponse<>(pageImpl.getContent(), dataRequest.getFilters(), pageable, pageImpl.getTotalElements());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/technologies/{technologyId}", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteTechnology(@PathVariable("id") int optimizationId,
                                 @PathVariable("technologyId") int technologyId, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        optimizationDependenciesManager.deleteTechnology(optimization, technologyId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/technologies", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteTechnologies(@PathVariable("id") int optimizationId,
                                   @RequestBody List<Integer> entitiesToDelete, @CurrentUser PortalUser currentUser) {
        entitiesToDelete.forEach(id -> deleteTechnology(optimizationId, id, currentUser));
    }


    /**
     * MATERIALS
     **/
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/materials", method = RequestMethod.POST)
    @ResponseStatus(HttpStatus.OK)
    public void addMaterial(@PathVariable("id") int optimizationId,
                            @RequestParam("entityId") int materialId, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        MaterialEntity material = materialManager
                .checkEntityExistsAndAccesible(materialId, AccessRights.view, currentUser.getUser());
        optimizationDependenciesManager.addMaterial(optimization, material, currentUser.getUser());
    }

    @RequestMapping(value = "/{id}/materials", method = RequestMethod.GET)
    public List<OptimizationDependency> getMaterials(@PathVariable("id") int optimizationId,
                                                     @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        return optimizationDependenciesManager.getMaterials(optimization, currentUser.getUser());
    }

    @RequestMapping(value = "/{id}/materials/list", method = RequestMethod.POST)
    public PageingResponse<OptimizationDependency> getMaterialsPage(@PathVariable("id") int optimizationId,
                                                                    @RequestBody DataRequest dataRequest,
                                                                    @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        PageImpl<OptimizationDependency> pageImpl = optimizationDependenciesManager
                .getMaterialsPage(optimization, pageable, currentUser.getUser());
        return new PageingResponse<>(pageImpl.getContent(), dataRequest.getFilters(), pageable, pageImpl.getTotalElements());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/materials/{materialId}", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteMaterial(@PathVariable("id") int optimizationId,
                               @PathVariable("materialId") int materialId, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        optimizationDependenciesManager.deleteMaterial(optimization, materialId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/materials", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteMaterials(@PathVariable("id") int optimizationId,
                                @RequestBody List<Integer> entitiesToDelete, @CurrentUser PortalUser currentUser) {
        entitiesToDelete.forEach(id -> deleteMaterial(optimizationId, id, currentUser));
    }


    /**
     * Load conditions
     **/
    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/loadConditions", method = RequestMethod.POST)
    public List<OptimizationDependency> addLoadConditions(HttpServletRequest request, @PathVariable("optimizationId") Integer optimizationId,
                                                   @RequestPart("file") List<MultipartFile> newLoadConditions, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        return newLoadConditions.stream()
                .map(lc-> new OptimizationDependency(optimizationManager.addModel(optimization, lc, FileType.LoadCondition, currentUser.getUser()), optimizationId))
                .collect(toList());

    }

    @RequestMapping(value = "/{optimizationId}/loadConditions", method = RequestMethod.GET)
    public List<OptimizationDependency> getAllLoadConditions(@PathVariable("optimizationId") Integer optimizationId,
                                                             @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        return optimizationManager.getOptimizationFiles(optimization, currentUser.getUser()).stream()
                .filter(file -> file.getFileTypeId() == FileType.LoadCondition.getId())
                .map(file -> new OptimizationDependency(file, optimizationId))
                .collect(toList());
    }

    @RequestMapping(value = "/{optimizationId}/loadConditions/list", method = RequestMethod.POST)
    public PageingResponse<OptimizationDependency> getLoadConditionsPage(@PathVariable("optimizationId") Integer optimizationId,
                                                                         @RequestBody DataRequest dataRequest,
                                                                         @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        dataRequest.getFilters().put("fileTypeId", FileType.LoadCondition.getId());
        Pageable pageRequest = dataRequest.getPageable().toSpringPageable();
        PageImpl<FileEntity> page = optimizationManager.getOptimizationFilesPage(optimization, dataRequest.getFilters(),
                pageRequest, currentUser.getUser());
        return new PageingResponse<>(
                page.getContent().stream().map(file -> new OptimizationDependency(file, optimizationId)).collect(toList()),
                dataRequest.getFilters(), pageRequest, page.getTotalElements());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/loadConditions/{id}", method = RequestMethod.DELETE)
    public void deleteLoadCondition(@PathVariable("optimizationId") Integer optimizationId,
                                    @PathVariable("id") Integer fileId,
                                    @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        FileEntity fileToDelete = fileManager.findOne(fileId, currentUser.getUser());
        if (fileToDelete == null) {
            throw new NotFoundException(fileId, FileEntity.class);
        }
        optimizationManager.deleteModel(optimization, fileToDelete, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/loadConditions", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteLoadConditions(@PathVariable("id") int optimizationId,
                                     @RequestBody List<Integer> entitiesToDelete, @CurrentUser PortalUser currentUser) {
        entitiesToDelete.forEach(id -> deleteLoadCondition(optimizationId, id, currentUser));
    }
}
