package ru.crp.lvms.optimizations.base.dao;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.stereotype.Repository;
import ru.crp.lvms.model.dao.BaseCompositeDao;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.entities.OptimizationDependency;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntityTable;
import ru.crp.lvms.optimizations.base.entities.OptimizationFlow;
import ru.crp.lvms.optimizations.base.enums.OptimizationCad;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.enums.OptimizationSolver;
import ru.crp.lvms.optimizations.base.mappers.OptimizationFlowResultRowMapper;
import ru.crp.lvms.optimizations.base.mappers.OptimizationFlowRowMapper;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

@Repository
public class OptimizationDao extends BaseCompositeDao<OptimizationEntityTable, OptimizationEntity> {

    private static final String CAO_TABLE_NAME = "caos";
    private static final String CAE_TABLE_NAME = "caes";
    private static final String CAD_TABLE_NAME = "cads";

    @Component
    public static class TableDao extends BaseSpaceDao<OptimizationEntityTable> {

    }

    @Autowired
    private TableDao tableDao;

    @PostConstruct
    public void init() {
        getJdbcTemplate().query("SELECT id, max_value, min_value, display_name FROM optimization_base_setting_types", rs -> {
            int id = rs.getInt("id");
            int maxValue = rs.getInt("max_value");
            int minValue = rs.getInt("min_value");
            String displayName = rs.getString("display_name");
            MFBaseSettingType type = MFBaseSettingType.valueOf(id);
            type.setMaxValue(maxValue);
            type.setMinValue(minValue);
            type.setDisplayName(displayName);
        });

        getJdbcTemplate().query("SELECT id, max_value, min_value, display_name FROM optimus_optimization_base_setting_types", rs -> {
            int id = rs.getInt("id");
            int maxValue = rs.getInt("max_value");
            int minValue = rs.getInt("min_value");
            String displayName = rs.getString("display_name");
            OptimusBaseSettingType type = OptimusBaseSettingType.valueOf(id);
            type.setMaxValue(maxValue);
            type.setMinValue(minValue);
            type.setDisplayName(displayName);
        });

        fillSolverEnum(CAO_TABLE_NAME, OptimizationCao::fromUniqueName);
        fillSolverEnum(CAE_TABLE_NAME, OptimizationCae::fromUniqueName);
        fillSolverEnum(CAD_TABLE_NAME, OptimizationCad::fromUniqueName);
    }

    @Override
    public TableDao tableDao() {
        return tableDao;
    }

    private void fillSolverEnum(String tableName, Function<String, OptimizationSolver> finder) {
        getJdbcTemplate().query("SELECT id, unique_name FROM " + tableName, rs -> {
            int id = rs.getInt("id");
            String uniqueName = rs.getString("unique_name");
            OptimizationSolver solver = finder.apply(uniqueName);
            solver.setId(id);
        });
    }


    public List<OptimizationEntity> findByFileId(Integer fileId) {
        return find("JOIN optimization_files of ON of.optimization_id = " + getAlias() + ".id WHERE of.file_id = ?", fileId);
    }

    public void appendFile(int optimizationId, int fileId) {
        jdbcTemplate().update("INSERT INTO optimization_files(optimization_id, file_id, ordering)" +
                " SELECT ?, ?," +
                " coalesce(MAX(ordering),-1) + 1 AS ordering" +
                " FROM optimization_files" +
                " WHERE optimization_id = ?", optimizationId, fileId, optimizationId);
    }

    public void deleteFile(int optimizationId, int fileId) {
        jdbcTemplate().update("DELETE FROM optimization_files WHERE optimization_id = ? AND file_id = ?", optimizationId, fileId);
    }

    public void clearFiles(int optimizationId) {
        jdbcTemplate().update("DELETE FROM optimization_files WHERE optimization_id=?", optimizationId);
    }

    public void replaceFiles(Integer optimizationId, List<Integer> fileIds) {
        clearFiles(optimizationId);
        List<Object[]> paramsInsert = new ArrayList<>();
        int ordering = 0;
        for (Integer fileId : fileIds) {
            paramsInsert.add(new Object[]{optimizationId, fileId, ordering});
            ordering++;
        }
        if (!paramsInsert.isEmpty()) {
            jdbcTemplate().batchUpdate(
                    "INSERT INTO optimization_files (optimization_id, file_id, ordering) VALUES (?,?,?)",
                    paramsInsert);
        }
    }

    public List<OptimizationEntity> findByPid(int pid) {
        return find("WHERE pid = ?", pid);
    }

    public List<OptimizationEntity> getOptimizationsByConstructionElementId(int constructionElementId) {
        return find("JOIN optimization_folders of ON of.id = " + getAlias() + ".pid " +
                "JOIN construction_elements ce ON ce.id = of.construction_element_id " +
                "WHERE ce.id = ?", constructionElementId);
    }

    public void addTechnology(int optimizationId, int technologyId) {
        jdbcTemplate().update("INSERT INTO optimization_technologies(optimization_id, technology_id) VALUES (?,?)",
                optimizationId, technologyId);
    }

    public List<OptimizationDependency> getTechnologies(int optimizationId) {
        return jdbcTemplate().query("SELECT ot.*, t.name" +
                " FROM optimization_technologies ot" +
                " JOIN technologies t ON ot.technology_id = t.id" +
                " WHERE ot.optimization_id = ?", (rs, rowNum) -> new OptimizationDependency(
                rs.getInt("technology_id"),
                rs.getInt("optimization_id"),
                rs.getString("name")), optimizationId);
    }

    public List<OptimizationDependency> getTechnologiesPage(int optimizationId, int pageNum, int pageSize) {
        return jdbcTemplate().query("SELECT ot.*, t.name" +
                " FROM optimization_technologies ot" +
                " JOIN technologies t ON ot.technology_id = t.id" +
                " WHERE ot.optimization_id = ?" +
                " OFFSET ? LIMIT ?", (rs, rowNum) -> new OptimizationDependency(
                rs.getInt("technology_id"),
                rs.getInt("optimization_id"),
                rs.getString("name")), optimizationId, pageNum * pageSize, pageSize);
    }

    public int getTotalTechnologies(int optimizationId) {
        return jdbcTemplate().queryForObject("SELECT COUNT(*)" +
                " FROM optimization_technologies ot" +
                " WHERE ot.optimization_id = ?", Integer.class, optimizationId);
    }

    public boolean technologyIsUsed(int optimizationId, int technologyId) {
        return jdbcTemplate().queryForObject("SELECT EXISTS(SELECT id FROM optimization_technologies" +
                " WHERE optimization_id = ? AND technology_id = ?)", Boolean.class, optimizationId, technologyId);
    }

    public void deleteTechnology(int optimizationId, int technologyId) {
        jdbcTemplate().update("DELETE FROM optimization_technologies WHERE optimization_id = ? AND technology_id = ?",
                optimizationId, technologyId);
    }

    public boolean materialIsUsed(int optimizationId, int materialId) {
        return jdbcTemplate().queryForObject("SELECT EXISTS(SELECT id FROM optimization_materials" +
                " WHERE optimization_id = ? AND material_id = ?)", Boolean.class, optimizationId, materialId);
    }

    public void addMaterial(int optimizationId, int materialId) {
        jdbcTemplate().update("INSERT INTO optimization_materials(optimization_id, material_id) VALUES (?,?)",
                optimizationId, materialId);
    }

    public void deleteMaterial(int optimizationId, int materialId) {
        jdbcTemplate().update("DELETE FROM optimization_materials WHERE optimization_id = ? AND material_id = ?",
                optimizationId, materialId);
    }

    public List<OptimizationDependency> getMaterials(int optimizationId) {
        return jdbcTemplate().query("SELECT om.*, m.name" +
                " FROM optimization_materials om" +
                " JOIN materials m ON om.material_id = m.id" +
                " WHERE om.optimization_id = ?", (rs, rowNum) -> new OptimizationDependency(
                rs.getInt("material_id"),
                rs.getInt("optimization_id"),
                rs.getString("name")), optimizationId);
    }

    public List<OptimizationDependency> getMaterialsPage(int optimizationId, int pageNumber, int pageSize) {
        return jdbcTemplate().query("SELECT om.*, m.name" +
                " FROM optimization_materials om" +
                " JOIN materials m ON om.material_id = m.id" +
                " WHERE om.optimization_id = ?" +
                " OFFSET ? LIMIT ?", (rs, rowNum) -> new OptimizationDependency(
                rs.getInt("material_id"),
                rs.getInt("optimization_id"),
                rs.getString("name")), optimizationId, pageNumber * pageSize, pageSize);
    }

    public int getTotalMaterials(int optimizationId) {
        return jdbcTemplate().queryForObject("SELECT COUNT(*)" +
                " FROM optimization_materials om" +
                " WHERE om.optimization_id = ?", Integer.class, optimizationId);
    }

    public OptimizationFlow getOptimizationFlow(int optimizationId) {
        return getJdbcTemplate()
                .queryForObject("SELECT * FROM optimization_flow_join WHERE optimization_id = ?",
                        new OptimizationFlowRowMapper(), optimizationId);
    }

    public List<OptimizationFlow.ResultsInfo> getOptimizationResultsInfo(int optimizationId) {
        return getJdbcTemplate()
                .query("SELECT * FROM optimization_flow_results_join WHERE optimization_id = ?",
                        new OptimizationFlowResultRowMapper(), optimizationId);
    }

    public boolean isFileUsedByAnyOptimization(int fileId) {
        return getJdbcTemplate().queryForObject("SELECT EXISTS(SELECT id FROM optimization_files WHERE file_id = ?)",
                Boolean.class, fileId);
    }

    public boolean isFileUsedByOptimization(int optimizationId, String modelName) {
        return getJdbcTemplate().queryForObject("SELECT EXISTS(SELECT f.id FROM optimization_files of" +
                        " INNER JOIN files f ON f.id = of.file_id WHERE of.optimization_id = ? AND f.name = ?)", Boolean.class,
                optimizationId, modelName);
    }


    public boolean isFileUsedByOptimization(int optimizationId, int fileId) {
        return getJdbcTemplate().queryForObject("SELECT EXISTS(SELECT of.id FROM optimization_files of" +
                        " WHERE of.optimization_id = ? AND of.file_id = ?)", Boolean.class,
                optimizationId, fileId);
    }

    @Nullable
    public Integer findSelectedAlgorithmId(int optimizationId) {
        List<Integer> ids = getJdbcTemplate().queryForList("SELECT object_id FROM optimization_selected_does_and_algorithms" +
                        " WHERE optimization_id = ? AND object_type_id = ?",
                Integer.class, optimizationId, ParametersDictionaryType.OptimizationMethod.getId());
        return ids.isEmpty() ? null : ids.get(0);
    }

    @Nullable
    public Integer findSelectedDoeId(int optimizationId) {
        List<Integer> ids = getJdbcTemplate().queryForList("SELECT object_id FROM optimization_selected_does_and_algorithms" +
                        " WHERE optimization_id = ? AND object_type_id = ?",
                Integer.class, optimizationId, ParametersDictionaryType.DoEType.getId());
        return ids.isEmpty() ? null : ids.get(0);
    }

    @Nullable
    public Integer findSelectedOptimizationMethodId(int optimizationId) {
        List<Integer> ids = getJdbcTemplate().queryForList("SELECT object_id FROM optimization_selected_does_and_algorithms" +
                        " WHERE optimization_id = ? AND object_type_id = ?",
                Integer.class, optimizationId, ParametersDictionaryType.ToskaOptimizationMethod.getId());
        return ids.isEmpty() ? null : ids.get(0);
    }

    public void selectAlgorithm(int optimizationId, @Nullable Integer selectedAlgorithmId) {
        getJdbcTemplate().update("DELETE FROM optimization_selected_does_and_algorithms WHERE optimization_id = ? AND object_type_id = ?",
                optimizationId, ParametersDictionaryType.OptimizationMethod.getId());
        if (selectedAlgorithmId != null) {
            getJdbcTemplate().update("INSERT INTO optimization_selected_does_and_algorithms(object_type_id, object_id, optimization_id)" +
                    " VALUES(?, ?, ?)", ParametersDictionaryType.OptimizationMethod.getId(), selectedAlgorithmId, optimizationId);
        }
    }

    public void selectDoe(int optimizationId, Integer selectedDoeId) {
        getJdbcTemplate().update("DELETE FROM optimization_selected_does_and_algorithms WHERE optimization_id = ? AND object_type_id = ?",
                optimizationId, ParametersDictionaryType.DoEType.getId());
        if (selectedDoeId != null) {
            getJdbcTemplate().update("INSERT INTO optimization_selected_does_and_algorithms(object_type_id, object_id, optimization_id)" +
                    " VALUES(?, ?, ?)", ParametersDictionaryType.DoEType.getId(), selectedDoeId, optimizationId);
        }
    }

    public void selectOptimizationMethod(int optimizationId, Integer selectedOptMethodId) {
        getJdbcTemplate().update("DELETE FROM optimization_selected_does_and_algorithms WHERE optimization_id = ? AND object_type_id = ?",
                optimizationId, ParametersDictionaryType.ToskaOptimizationMethod.getId());
        if (selectedOptMethodId != null) {
            getJdbcTemplate().update("INSERT INTO optimization_selected_does_and_algorithms(object_type_id, object_id, optimization_id)" +
                    " VALUES(?, ?, ?)", ParametersDictionaryType.ToskaOptimizationMethod.getId(), selectedOptMethodId, optimizationId);
        }
    }

    public List<OptimizationEntity> findByPathNames(String pathNames) {
        return find("WHERE path_names = ?", pathNames);
    }
}
