package ru.crp.lvms.optimizations.base.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationDependency;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;

import java.util.List;

@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
@Service
public class OptimizationDependenciesManager {
    @Autowired
    private OptimizationDao optimizationDao;

    public void addTechnology(OptimizationEntity optimization, TechnologyEntity technology, UserEntity currentUserUser) {
        if (optimizationDao.technologyIsUsed(optimization.getId(), technology.getId())) {
            throw new BadRequestException("Технология уже используется в оптимизации");
        }
        optimizationDao.addTechnology(optimization.getId(), technology.getId());
    }

    public List<OptimizationDependency> getTechnologies(OptimizationEntity optimization, UserEntity currentUser) {
        return optimizationDao.getTechnologies(optimization.getId());
    }
    public PageImpl<OptimizationDependency> getTechnologiesPage(OptimizationEntity optimization,
                                                                Pageable pageable,
                                                                UserEntity currentUser) {
        List<OptimizationDependency> content = optimizationDao.getTechnologiesPage(optimization.getId(), pageable.getPageNumber(), pageable.getPageSize());
        return new PageImpl<>(content, pageable, (long)optimizationDao.getTotalTechnologies(optimization.getId()));
    }

    public void deleteTechnology(OptimizationEntity optimization, int technologyId, UserEntity currentUserUser) {
        if (!optimizationDao.technologyIsUsed(optimization.getId(), technologyId)) {
            throw new BadRequestException("Технология уже используется в оптимизации");
        }
        optimizationDao.deleteTechnology(optimization.getId(), technologyId);
    }

    public void addMaterial(OptimizationEntity optimization, MaterialEntity material, UserEntity currentUser) {
        if (optimizationDao.materialIsUsed(optimization.getId(), material.getId())) {
            throw new BadRequestException("Материал уже используется в оптимизации");
        }
        optimizationDao.addMaterial(optimization.getId(), material.getId());
    }

    public void deleteMaterial(OptimizationEntity optimization, int materialId, UserEntity currentUser) {
        if (!optimizationDao.materialIsUsed(optimization.getId(), materialId)) {
            throw new BadRequestException("Технология уже используется в оптимизации");
        }
        optimizationDao.deleteMaterial(optimization.getId(), materialId);
    }

    public List<OptimizationDependency> getMaterials(OptimizationEntity optimization, UserEntity currentUser) {
        return optimizationDao.getMaterials(optimization.getId());
    }

    public PageImpl<OptimizationDependency> getMaterialsPage(OptimizationEntity optimization, Pageable pageable, UserEntity currentUser) {
        List<OptimizationDependency> content = optimizationDao.getMaterialsPage(optimization.getId(), pageable.getPageNumber(), pageable.getPageSize());
        return new PageImpl<>(content, pageable, (long)optimizationDao.getTotalMaterials(optimization.getId()));
    }
}
