package ru.crp.lvms.optimizations.mf.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.dao.ParametersDictionaryDao;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.mf.dao.*;
import ru.crp.lvms.optimizations.mf.entities.*;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toList;

@Component
public class MFDefaultParamsCreator {
    public static final int SIMPLEX_METHOD_ID = 7;
    public static final int MOGA_METHOD_ID = 8;
    public static final int DOE_SEQUENCE_METHOD_ID = 9;

    public static final int RANDOM_DOE_ID = 10;
    public static final int SOBOL_DOE_ID = 11;
    public static final int ULH_DOE_ID = 12;

    // value must be synchronized with cml-solver:MFZipFileTransformer.TIME_OUT_PARAM_NAME
    public static final String TIME_OUT_PARAM_NAME = "Тайм-аут";

    @Autowired
    private MFOptimizationConstraintsDao MFOptimizationConstraintsDao;

    @Autowired
    private MFOptimizationTargetFuncsDao MFOptimizationTargetFuncsDao;

    @Autowired
    private MFOptimizationBaseSettingsDao MFOptimizationBaseSettingsDao;

    @Autowired
    private MFOptimizationVaryingParamsDao MFOptimizationVaryingParamsDao;

    @Autowired
    private MFOptimizationCADVaryingParamsDao mfOptimizationCADVaryingParamsDao;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    @Autowired
    private ParametersDictionaryDao parametersDictionaryDao;

    @Autowired
    private MFOptimizationSettingsManager MFOptimizationSettingsManager;


    public void addDefaultSettings(OptimizationEntity newOptimization) {
        List<MFOptimizationBaseSetting> initialBasicSettings = createFrontierDefaultSettings(newOptimization);
        initialBasicSettings.forEach(MFOptimizationBaseSettingsDao::insert);

        List<MFOptimizationConstraint> initialConstraints = createFrontierConstraints(newOptimization);
        initialConstraints.forEach(MFOptimizationConstraintsDao::insert);

        List<MFOptimizationTargetFunc> initialTargetFuncs = createFrontierTargetFuncs(newOptimization);
        initialTargetFuncs.forEach(MFOptimizationTargetFuncsDao::insert);

        List<MFOptimizationVaryingParam> initialVaryingParams = createFrontierInitialVaryingParams(newOptimization);
        initialVaryingParams.forEach(MFOptimizationVaryingParamsDao::insert);

        if (newOptimization.getSolverCadId() != null) {
            List<MFOptimizationCADVaryingParam> initialCADVaryingParams = createFrontierInitialCADVaryingParams(newOptimization);
            initialCADVaryingParams.forEach(mfOptimizationCADVaryingParamsDao::insert);
        }

        List<OptimizationJsonSetting> runParams = createFrontierRunParams(newOptimization);
        runParams.forEach(optimizationJsonSettingsDao::createSettings);

        List<OptimizationJsonSetting> doeParams = createFrontrierDoeParams(newOptimization);
        doeParams.forEach(optimizationJsonSettingsDao::createSettings);

        List<OptimizationJsonSetting> algorithmParams = createFrontierAlgorithmParams(newOptimization);
        algorithmParams.forEach(optimizationJsonSettingsDao::createSettings);

        Integer numTargetFuncs = initialBasicSettings.stream()
                .filter(s -> s.getTypeId() == MFBaseSettingType.NUM_TARGET_FUNCS.getId())
                .map(MFOptimizationBaseSetting::getCurrentValue)
                .findFirst().orElse(null);
        if (numTargetFuncs == null) {
            numTargetFuncs = 1;
        }
        MFOptimizationSettingsManager.selectDefaultAlgorithm(newOptimization.getId(), numTargetFuncs);
    }

    public List<OptimizationJsonSetting> createFrontierAlgorithmParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        Map<Integer, List<ParametersDictionary>> methods = parametersDictionaryDao
                .findByTypeName(ParametersDictionaryType.OptimizationMethod).stream().collect(groupingBy(ParametersDictionary::getId));

        List<ParametersDictionary> methodsList = methods.getOrDefault(SIMPLEX_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'SIMPLEX' method");
        }
        ParametersDictionary simplexMethod = methodsList.get(0);

        methodsList = methods.getOrDefault(MOGA_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'MOGA-II' method");
        }
        ParametersDictionary moga2Method = methodsList.get(0);

        methodsList = methods.getOrDefault(DOE_SEQUENCE_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'DOE Sequence' method");
        }

        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simplexMethod.getId(), "Максимальное число итераций (1…9999)", "simplex.parameter.num.max.iterations", "200"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, simplexMethod.getId(), "Автоматический рестарт", "simplex.parameter.restart", true));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simplexMethod.getId(), "Критерий сходимости(1.0Е-10…1)", "simplex.parameter.convergence.limit", "0.00001"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, simplexMethod.getId(), "Учет нарушения ограничений при сходимости", "simplex.parameter.feasibility", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, simplexMethod.getId(), "Указывать оператор, применяемый на данной итерации", "simplex.parameter.categorize.operators", false));

        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "Число генераций наборов DoE (1…5000)", "moga.parameter.num.generations", "200"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "Directional Cross-Over (0.0…1.0)", "moga.parameter.prob.directional.crossover", "0.5"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "Selection(0.0…1.0)", "moga.parameter.prob.selection", "0.05"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "Mutation(0.0…1.0)", "moga.parameter.prob.mutation", "0.1"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "DNA String Mutation Ratio(0.0…1.0)", "moga.parameter.dna.mutation.ratio", "0.05"));
        settings.add(OptimizationJsonSetting.newListSetting(optId, moga2Method.getId(), "Elitism", "moga.parameter.elitism",
                asList(new OptimizationListSetting(true, "Enabled", "moga.parameter.elitism.enabled"), new OptimizationListSetting(false, "Disabled", "moga.parameter.elitism.disabled"))));
        settings.add(OptimizationJsonSetting.newListSetting(optId, moga2Method.getId(), "Treat Constraint", "moga.parameter.treat.constraints",
                asList(new OptimizationListSetting(true, "Penalising Objectives", "moga.parameter.treat.constraints.penalising"), new OptimizationListSetting(false, "As One Extra Objective", "moga.parameter.treat.constraints.extraobj"))));
        settings.add(OptimizationJsonSetting.newListSetting(optId, moga2Method.getId(), "Тип алгоритма", "moga.parameter.algorithm.type",
                asList(new OptimizationListSetting(true, "Generational Evolution", "moga.parameter.algorithm.type.generational"), new OptimizationListSetting(false, "Steady Evolution", "moga.parameter.algorithm.type.steady"),
                        new OptimizationListSetting(false, "Adaptive Evolution", "moga.parameter.algorithm.type.adaptive"))));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, moga2Method.getId(), "Генератор случайных чисел(0...999)", "moga.parameter.seed", "346"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, moga2Method.getId(), "Указывать номер очередной генерации", "moga.parameter.categorize.generations", false));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, moga2Method.getId(), "Указывать операторы генетического алгоритма", "moga.parameter.categorize.operators", false));

        return settings;
    }

    public List<OptimizationJsonSetting> createFrontrierDoeParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        Map<Integer, List<ParametersDictionary>> does = parametersDictionaryDao
                .findByTypeName(ParametersDictionaryType.DoEType).stream().collect(groupingBy(ParametersDictionary::getId));

        List<ParametersDictionary> doesList = does.getOrDefault(10, emptyList());
        if (doesList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'Random' doe");
        }
        ParametersDictionary randomDoe = doesList.get(0);
        doesList = does.getOrDefault(11, emptyList());
        if (doesList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'Sobol' doe");
        }
        ParametersDictionary sobolDoe = doesList.get(0);

        doesList = does.getOrDefault(12, emptyList());
        if (doesList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'Uniform Latin Hypercube' doe");
        }
        ParametersDictionary ulhDoe = doesList.get(0);

        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, randomDoe.getId(), "Число точек (1…256000)", "doerand.parameter.num.rows", "9"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, randomDoe.getId(), "Удалять точки не удовлетворяющие ограничениям", "doerand.parameter.reject", true));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, randomDoe.getId(), "Генератор случайных чисел (0…999)", "doerand.parameter.seed", "0"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, randomDoe.getId(), "Удалять повторяющиеся точки", "common.parameter.repeat.design", true));

        settings.add(OptimizationJsonSetting.newPlainSetting(optId, sobolDoe.getId(), "Число точек (1…256000)", "sobol.parameter.num.rows", "11"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, sobolDoe.getId(), "Удалять точки не удовлетворяющие ограничениям", "sobol.parameter.reject", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, sobolDoe.getId(), "Удалять повторяющиеся точки", "common.parameter.repeat.design", true));

        settings.add(OptimizationJsonSetting.newPlainSetting(optId, ulhDoe.getId(), "Число точек (1…256000)", "ulh.parameter.designs", "10"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, ulhDoe.getId(), "Минимальная корреляция входных переменных", "ulh.parameter.distance.correlation", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, ulhDoe.getId(), "Критерий MAXIMIN", "ulh.parameter.distance.maximin", true));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, ulhDoe.getId(), "Генератор случайных чисел (0…999)", "ulh.parameter.seed", "0"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, ulhDoe.getId(), "Минимизировать число точек, не удовлетворяющих ограничениям", "ulh.parameter.reject", true));
        // default value is used for parallelism settings (see ConfigGenerator::formDoe)
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, ulhDoe.getId(), "Удалять повторяющиеся точки", "common.parameter.repeat.design", true));

        return settings;
    }

    public List<OptimizationJsonSetting> createFrontierRunParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();

        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, "Сохранять ошибочные итерации в БД", "SAVE_ERROR_DESIGNS", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, "Производить расчет итерации, если в БД уже есть итерация с такими же параметрами", "CHECK_REPEATED_DESIGNS", false));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, "Производить расчет итераций, в которых нарушаются ограничения на значения варьируемых параметров", "EVAL_UNFEASIBLE_DESIGNS", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, "Присваивать всем переменным значения NaN(not-a-number), если итерация завершается с ошибкой", "FORCE_ERROR_DESIGNS_TO_NAN", true));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, "Сохранять повторяющиеся итерации в БД", "SAVE_REPEATED_DESIGNS", true));

        settings.add(OptimizationJsonSetting.newListSetting(optId, "Очищение рабочей директории расчета", "CLEAR_DESIGN_DIR",
                asList(new OptimizationListSetting(false, "Никогда", "never"), new OptimizationListSetting(true, "Всегда", "always"), new OptimizationListSetting(false, "Без ошибок", "not_on_failure"))));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, "Неудаляемые файлы", "KEEP_FILES", null));
        settings.add(OptimizationJsonSetting.newListSetting(optId, "Степень детализации файлов журнала", "LOG_LEVEL",
                asList(new OptimizationListSetting(true, "Fine", "FINE"), new OptimizationListSetting(false, "Info", "INFO"),
                        new OptimizationListSetting(false, "Warning", "WARNING"), new OptimizationListSetting(false, "Severe", "SEVERE"))));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, "Число одновременных вычислений", "NUM_CONC_DESIGNS", "1"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, TIME_OUT_PARAM_NAME, null, ""));
        return settings;
    }


    public List<MFOptimizationVaryingParam> createFrontierInitialVaryingParams(OptimizationEntity newOptimization) {
        return IntStream.range(1, MFBaseSettingType.NUM_CAO_VARYING_PARAMS.getMaxValue() + 1)
                .mapToObj(num -> new MFOptimizationVaryingParam(newOptimization.getId(), "par" + num))
                .collect(toList());
    }

    public List<MFOptimizationCADVaryingParam> createFrontierInitialCADVaryingParams(OptimizationEntity newOptimization) {
        return IntStream.range(1, MFBaseSettingType.NUM_CAD_VARYING_PARAMS.getMaxValue() + 1)
                .mapToObj(num -> new MFOptimizationCADVaryingParam(newOptimization.getId(), "gpar" + num))
                .collect(toList());
    }

    public List<MFOptimizationTargetFunc> createFrontierTargetFuncs(OptimizationEntity newOptimization) {
        return IntStream.range(1, MFBaseSettingType.NUM_TARGET_FUNCS.getMaxValue() + 1)
                .mapToObj(num -> new MFOptimizationTargetFunc(newOptimization.getId(), "obj" + num, 1))
                .collect(toList());
    }

    public List<MFOptimizationConstraint> createFrontierConstraints(OptimizationEntity newOptimization) {
        return IntStream.range(1, MFBaseSettingType.NUM_CONSTRAINTS.getMaxValue() + 1)
                .mapToObj(num -> new MFOptimizationConstraint(newOptimization.getId(), "c" + num, 4))
                .collect(toList());
    }

    public List<MFOptimizationBaseSetting> createFrontierDefaultSettings(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        return Stream.of(MFBaseSettingType.values())
                // cad varying params is required only for optimizations with cad solver
                .filter(s -> newOptimization.getSolverCadId() != null || s != MFBaseSettingType.NUM_CAD_VARYING_PARAMS)
                .map(type -> new MFOptimizationBaseSetting(optId, type.getId(), 1))
                .collect(toList());
    }
}
