package ru.crp.lvms.optimizations.mf.results;

import ru.crp.lvms.chart.Point;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toList;

public class ModFrontierParamsRow {
    private int iteration;
    private String rid;
    private boolean isError;
    private boolean isVirtual;
    private boolean isFeasible;
    private boolean isMarked;
    private int category;
    private List<Double> varyingParams;
    private List<Double> outputParams;
    private List<Double> targetFuncs;
    private List<Double> constraints;

    public ModFrontierParamsRow(String[] resultRow, Map<Integer, Integer> varyingParamsIndices, Map<Integer, Integer> outputParamsIndices, Map<Integer, Integer> targetFunctionsIndices, Map<Integer, Integer> constraintsIndices) {
        this.iteration = Integer.parseInt(resultRow[0]);
        this.rid = resultRow[1];

        int startParamsIdx = 2;
        int endParamIdx = 2 + varyingParamsIndices.size();
        List<Double> varyingParamsRaw = asList(resultRow).subList(startParamsIdx, endParamIdx).stream().map(Double::parseDouble).collect(toList());
        varyingParams = placeItemsInCorrectOrder(varyingParamsRaw, varyingParamsIndices);

        startParamsIdx = endParamIdx;
        endParamIdx = endParamIdx + outputParamsIndices.size();
        List<Double> outputParamsRaw = asList(resultRow).subList(startParamsIdx, endParamIdx).stream().map(Double::parseDouble).collect(toList());
        outputParams = placeItemsInCorrectOrder(outputParamsRaw, outputParamsIndices);

        startParamsIdx = endParamIdx;
        endParamIdx = endParamIdx + targetFunctionsIndices.size();
        List<Double> targetFuncsRaw = asList(resultRow).subList(startParamsIdx, endParamIdx).stream().map(Double::parseDouble).collect(toList());
        targetFuncs = placeItemsInCorrectOrder(targetFuncsRaw, targetFunctionsIndices);

        startParamsIdx = endParamIdx;
        endParamIdx = endParamIdx + constraintsIndices.size();
        List<Double> constraintsRaw = asList(resultRow).subList(startParamsIdx, endParamIdx).stream().map(Double::parseDouble).collect(toList());
        constraints = placeItemsInCorrectOrder(constraintsRaw, constraintsIndices);

        this.isError = Boolean.parseBoolean(resultRow[endParamIdx]);
        this.isVirtual = Boolean.parseBoolean(resultRow[endParamIdx + 1]);
        this.isFeasible = Boolean.parseBoolean(resultRow[endParamIdx + 2]);
        this.isMarked = Boolean.parseBoolean(resultRow[endParamIdx + 3]);
        this.category = Integer.parseInt(resultRow[endParamIdx + 4]);
    }

    private List<Double> placeItemsInCorrectOrder(List<Double> raw, Map<Integer, Integer> indices) {
        if (raw.size() != indices.size()) {
            throw new IllegalArgumentException("Count of indices is not equal to count of items.");
        }
        return IntStream.range(0, raw.size())
                .mapToObj(indices::get)
                .map(raw::get)
                .collect(Collectors.toList());
    }

    public Point toTargetFunctionPoint(int targetFuncIndex) {
        return isFeasible ? Point.commonPoint(iteration, targetFuncs.get(targetFuncIndex)) : Point.improperPoint(iteration, targetFuncs.get(targetFuncIndex));
    }

    public Point toVaryingParamPoint(int varyingParamsIndex, Integer outputVariableIdx) {
        return isFeasible ?
                Point.commonPoint(varyingParams.get(varyingParamsIndex), outputParams.get(outputVariableIdx)) :
                Point.improperPoint(varyingParams.get(varyingParamsIndex), outputParams.get(outputVariableIdx));
    }

    public Point toConstraintPoint(int constraintIndex) {
        return isFeasible ? Point.commonPoint(iteration, constraints.get(constraintIndex)) : Point.improperPoint(iteration, constraints.get(constraintIndex));
    }

    public boolean isAcceptable() {
        return isFeasible && !isError;
    }

    public boolean isError() {
        return isError;
    }

    public Double getTargetFunctionValue(int targetFuncIndex) {
        return targetFuncs.get(targetFuncIndex);
    }

    public boolean isDominated(ModFrontierParamsRow anotherRow, List<Comparator<Double>> comparators) {
        if (targetFuncs.isEmpty() || targetFuncs.size() != comparators.size()) {
            throw new IllegalArgumentException("Unexpected target funcs count");
        }
        final int UNKNOWN_STATE = 0;
        final int THIS_WIN = 1;
        final int ANOTHER_WIN = 2;
        final int IN_PARETO_RELATION = 3;

        int finalState = IntStream.range(0, comparators.size()).reduce(UNKNOWN_STATE, (state, curIndex) -> {
            if (state == IN_PARETO_RELATION) return state;
            int curState = comparators.get(curIndex).compare(targetFuncs.get(curIndex), anotherRow.targetFuncs.get(curIndex));
            if ((state == THIS_WIN && curState < 0) || (state == ANOTHER_WIN && curState > 0))
                return IN_PARETO_RELATION;
            return curState < 0 ? ANOTHER_WIN : curState > 0 ? THIS_WIN : curState;
        });
        return finalState == ANOTHER_WIN;
    }

    public List<Double> getVaryingParams() {
        return varyingParams;
    }

    public List<Double> getTargetFuncs() {
        return targetFuncs;
    }

    public List<Double> getConstraints() {
        return constraints;
    }

    public int getIteration() {
        return iteration;
    }
}
