package ru.crp.lvms.optimizations.mf.results;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FilenameUtils;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.Chart;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.resultsHandlers.KeyResultHandler;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;

import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

import static java.util.stream.Collectors.toList;

public class ModeFrontierParamsHandler extends KeyResultHandler {

    private ObjectMapper objectMapper;
    private final Function<Integer, List<MFOptimizationTargetFunc>> targetFunctionsSupplier;
    private final Function<Integer, List<MFOptimizationConstraint>> constraintsSupplier;
    private final Function<Integer, List<MFOptimizationVaryingParam>> varParamsSupplier;

    public ModeFrontierParamsHandler(Path pathToStorage, ObjectMapper objectMapper,
                                     Function<Integer, List<MFOptimizationTargetFunc>> targetFunctionsSupplier,
                                     Function<Integer, List<MFOptimizationConstraint>> constraintsSupplier,
                                     Function<Integer, List<MFOptimizationVaryingParam>> varParamsSupplier) {
        super(pathToStorage);
        this.objectMapper = objectMapper;
        this.targetFunctionsSupplier = targetFunctionsSupplier;
        this.constraintsSupplier = constraintsSupplier;
        this.varParamsSupplier = varParamsSupplier;
    }


    @Override
    public boolean canHandle(ResultType resultType, Path pathToFile) {
        return resultType == ResultType.modfrontierParams && Files.isRegularFile(pathToFile);
    }


    @Override
    protected List<ResultEntity> doHandleResultsFile(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        List<ResultEntity> extractedResults = new ArrayList<>();

        ResultEntity rawParams = initCommonResultProperties(pathToFile, resultsParent, createdBy);
        rawParams.setName(pathToFile.getFileName().toString());
        rawParams.setType(ResultType.modfrontierParams);
        extractedResults.add(rawParams);

        List<MFOptimizationTargetFunc> targetFuncs = targetFunctionsSupplier.apply(resultsParent.getId());
        List<MFOptimizationConstraint> constraintsSettings = constraintsSupplier.apply(resultsParent.getId());
        List<MFOptimizationVaryingParam> mfOptimizationVaryingParams = varParamsSupplier.apply(resultsParent.getId());
        try {
            if (targetFuncs.size() == 1) {
                extractedResults.addAll(handleSingleTargetFunction(pathToFile, resultsParent, createdBy, targetFuncs, constraintsSettings, mfOptimizationVaryingParams));
            }
            return extractedResults;
        } catch (IOException ex) {
            throw new ResultProcessingException(ex);
        }
    }


    private List<ResultEntity> handleSingleTargetFunction(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy,
                                                          List<MFOptimizationTargetFunc> mfOptimizationTargetFuncs, List<MFOptimizationConstraint> constraintsSettings,
                                                          List<MFOptimizationVaryingParam> mfOptimizationVaryingParams) throws IOException {
        ModFrontierParams modFrontierParams = parseModFrontierParams(pathToFile);
        List<ResultEntity> extractedResults = new ArrayList<>();
        for (int i = 0; i < modFrontierParams.getNumTargetFunctions(); ++i) {
            MFOptimizationTargetFunc currentTargetFunc = mfOptimizationTargetFuncs.get(i);
            Chart targetFunctionChart = modFrontierParams.createTargetFunctionChart(i, currentTargetFunc, mfOptimizationVaryingParams.size());
            String curveFileSuffix = String.format("%s-target-function", currentTargetFunc.getName());
            Path pathToJson = saveCurve(targetFunctionChart.toChartsSet(), curveFileSuffix, pathToFile);
            ResultEntity resultCurve = initCommonResultProperties(pathToJson, resultsParent, createdBy);
            resultCurve.setName(String.format("Целевая функция %s", currentTargetFunc.getName()));
            resultCurve.setType(ResultType.curves);
            extractedResults.add(resultCurve);
        }

        for (int i = 0; i < modFrontierParams.getNumConstraints(); ++i) {
            MFOptimizationConstraint currentConstraint = constraintsSettings.get(i);
            Chart constraintChart = modFrontierParams.createConstraintChart(i, currentConstraint);

            String curveFileSuffix = String.format("%s-constraint", currentConstraint.getName());
            Path pathToJson = saveCurve(constraintChart.toChartsSet(), curveFileSuffix, pathToFile);
            ResultEntity resultCurve = initCommonResultProperties(pathToJson, resultsParent, createdBy);
            resultCurve.setName(String.format("Ограничение %s", currentConstraint.getName()));
            resultCurve.setType(ResultType.curves);
            extractedResults.add(resultCurve);
        }
        return extractedResults;
    }


    public static ModFrontierParams parseModFrontierParams(Path pathToFile) throws IOException {
        final boolean[] resultsStarted = new boolean[]{false};
        List<String[]> resultData = Files.lines(pathToFile, Charset.forName("CP1251")).filter(str -> {
            if (resultsStarted[0]) return true;
            if (str.startsWith("<ID>")) {
                resultsStarted[0] = true;
                return true;
            }
            return false;
        })
                .map(str -> str.split("\\s+"))
                .collect(toList());
        ModFrontierParams modFrontierParams = new ModFrontierParams(resultData.get(0));
        resultData.stream().skip(1).forEach(modFrontierParams::addResultRow);
        return modFrontierParams;
    }


    private Path saveCurve(ChartsSet targetFunctionCurve, String fileNameSuffix, Path pathToRawFile) throws IOException {
        String jsonFileName = FilenameUtils.removeExtension(pathToRawFile.getFileName().toString());
        jsonFileName = String.format("%s-%s.json", jsonFileName, fileNameSuffix);
        Path pathToJsonFile = pathToRawFile.resolveSibling(jsonFileName);
        objectMapper.writeValue(pathToJsonFile.toFile(), targetFunctionCurve);
        return pathToJsonFile;
    }
}
