package ru.crp.lvms.optimizations.misc;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;

import javax.annotation.PostConstruct;
import java.util.List;

import static java.util.stream.Collectors.toList;

abstract public class AbstractOptimizationSettingsController<T extends AbstractOptimizationSettings, SETTING_TYPE> {

    @Autowired
    protected OptimizationManager optimizationManager;

    protected SETTING_TYPE settingsType;

    @PostConstruct
    public void setup() {
        settingsType = getSettingsType();
    }

    protected abstract SETTING_TYPE getSettingsType();

    protected abstract Integer getSettingsCount(int optimizationId);

    protected abstract AbstractOptimizationSettingsManager<SETTING_TYPE> getSettingsManager();

    protected abstract OptimizationCao getOptimizationSolverType();

    @RequestMapping(value = "", method = RequestMethod.GET)
    public List<? extends AbstractOptimizationSettings> getAllSettings(@PathVariable("optimizationId") int optimizationId,
                                                                       @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager()
                .findOptimizationSettings(optimizationId, settingsType, currentUser.getUser()).stream()
                .limit(getSettingsCount(optimizationId))
                .collect(toList());
    }

    @RequestMapping(value = "/{id}", method = RequestMethod.GET)
    public AbstractOptimizationSettings getSettingsById(@PathVariable("optimizationId") int optimizationId,
                                                        @PathVariable("id") int settingsId,
                                                        @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().findOptimizationSettingsById(settingsId, settingsType, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public AbstractOptimizationSettings updateSettings(@PathVariable("optimizationId") int optimizationId,
                                                       @PathVariable("id") int settingsId,
                                                       @RequestBody T updatedSettings,
                                                       @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        updatedSettings.setId(settingsId);
        updatedSettings.setOptimizationId(optimizationId);
        return getSettingsManager().updateSettings(updatedSettings, settingsType);
    }

    protected OptimizationEntity checkOptimizationSolverType(int optimizationId, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimizationEntity = optimizationManager.findOne(optimizationId, currentUser.getUser());
        if (optimizationEntity == null) {
            throw new BadRequestException("Оптимизация с таким идентификатором не найдена");
        }
        OptimizationCao type = OptimizationCao.valueOf(optimizationEntity.getSolverCaoId());
        if (getOptimizationSolverType() != type) {
            throw new BadRequestException("Тип оптимизации не соответствует типу запрашиваемого параметра");
        }
        return optimizationEntity;
    }
}
