package ru.crp.lvms.optimizations.optimus.controllers;

import com.fasterxml.jackson.annotation.JsonView;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.AlgorithmType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.controllers.JsonSettingsController;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.base.entities.SettingsView;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.optimus.config.OptimusConfigGenerator;
import ru.crp.lvms.optimizations.optimus.enums.OptimusAlgorithmType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.validators.task.optimization.CompositeOptimizationTaskParamsValidator;
import ru.crp.lvms.web.controllers.misc.FileHelper;

import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.URLs.MF_SETTINGS_URL;
import static ru.crp.lvms.utils.URLs.OPTIMUS_SETTINGS_URL;

@RestController
@RequestMapping(value = OPTIMUS_SETTINGS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptimusJsonSettingsController extends JsonSettingsController {

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private CompositeOptimizationTaskParamsValidator optimizationTaskParamsValidator;

    @Autowired
    private OptimusConfigGenerator optimusConfigGenerator;

    @RequestMapping(value = "/configFile", method = RequestMethod.GET)
    public void generateSettingsXMl(@PathVariable int optimizationId, HttpServletResponse response,
                                    @CurrentUser PortalUser currentUser) throws IOException, JAXBException {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        if (optimization.getSolverCaoId() != OptimizationCao.Optimus.getId()) {
            throw new BadRequestException(String.format("Настройки '%s' недоступны", OptimizationCao.Optimus));
        }
        TaskEntity task = new TaskEntity();
        task.setParentId(optimizationId);
        task.setSolverName(OptimizationCao.Optimus.getSolverConfigName());
        task.setParentType(TaskParentType.optimization);
        Errors errors = new BeanPropertyBindingResult(task, task.getClass().getName());
        optimizationTaskParamsValidator.validate(task, errors);
        if (errors.hasErrors() && !(errors.getAllErrors().size() == 1 && errors.getAllErrors().get(0).getCode().equals("solvingAlreadyStarted"))) {
            throw new TaskStartingException("Не удалось создать файл с 'settings.xml'", errors);
        }
        Path settings = optimusConfigGenerator.generateConfig(optimization);
        FileHelper.writeFileToResponse(settings, "settings.seq", response);
        Files.deleteIfExists(settings);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/runParams", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getRunParams(@PathVariable("optimizationId") int optimizationId,
                                                      @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findRunSettings(optimizationId);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/runParams/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updateRunParam(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                  @PathVariable("optimizationId") int optimizationId,
                                                  @PathVariable("paramId") int paramId,
                                                  @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }

    @RequestMapping(value = "/methods", method = RequestMethod.GET)
    public List<OptimizationListSetting> getAlgorithms(@PathVariable("optimizationId") int optimizationId,
                                                       @CurrentUser PortalUser currentUser) {
        Integer numTargetFuncs = optimusOptimizationSettingsManager.findBaseSettings(optimizationId, currentUser.getUser())
                .get(OptimusBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue();
        Integer selectedAlgorithmId = optimusOptimizationSettingsManager.findSelectedAlgorithmId(optimizationId);
        return optimusOptimizationSettingsManager.findByTypeName(ParametersDictionaryType.OptimusOptimizationMethod).stream()
                .filter(method -> method.getId() != OptimusAlgorithmType.EGO.getId() || numTargetFuncs == 1)
                .map(method -> new OptimizationListSetting(false, method.getId(), method.getName(), method.getName()))
                .peek(a -> {
                    if (selectedAlgorithmId != null && selectedAlgorithmId.equals(a.getId())) a.setSelected(true);
                })
                .collect(toList());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/methods/{methodId}", method = RequestMethod.PUT)
    public List<OptimizationListSetting> selectAlgorithm(@Valid @RequestBody OptimizationListSetting selectedAlgorithm,
                                                         @PathVariable("optimizationId") int optimizationId,
                                                         @PathVariable("methodId") int methodId,
                                                         @CurrentUser PortalUser currentUser) {
        optimusOptimizationSettingsManager.selectAlgorithm(optimizationId, methodId);
        return getAlgorithms(optimizationId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/selected", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getSelectedAlgorithmParams(@PathVariable("optimizationId") int optimizationId,
                                                                    @CurrentUser PortalUser currentUser) {
        Integer selectedAlgorithmId = optimusOptimizationSettingsManager.findSelectedAlgorithmId(optimizationId);
        if (selectedAlgorithmId == null) {
            throw new InternalException("Метод не выбран");
        }
        return getAlgorithmParams(optimizationId, selectedAlgorithmId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/{methodId}/params", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getAlgorithmParams(@PathVariable("optimizationId") int optimizationId,
                                                            @PathVariable("methodId") int methodId,
                                                            @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findSettingsByParent(optimizationId, methodId);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/params/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updateAlgorithmParam(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                        @PathVariable("optimizationId") int optimizationId,
                                                        @PathVariable("paramId") int paramId,
                                                        @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }


    @RequestMapping(value = "/does", method = RequestMethod.GET)
    public List<OptimizationListSetting> getDoes(@PathVariable("optimizationId") int optimizationId,
                                                 @CurrentUser PortalUser currentUser) {
        Integer selectedDoeId = optimusOptimizationSettingsManager.findSelectedDoeId(optimizationId);
        return optimusOptimizationSettingsManager.findByTypeName(ParametersDictionaryType.OptimusDoEType).stream()
                .map(method -> new OptimizationListSetting(false, method.getId(), method.getName(), method.getName()))
                .peek(a -> {
                    if (selectedDoeId != null && selectedDoeId.equals(a.getId())) a.setSelected(true);
                })
                .collect(toList());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/does/{doeId}", method = RequestMethod.PUT)
    public List<OptimizationListSetting> selectDoe(@Valid @RequestBody OptimizationListSetting selectedDoe,
                                                   @PathVariable("optimizationId") int optimizationId,
                                                   @PathVariable("doeId") int doeId,
                                                   @CurrentUser PortalUser currentUser) {
        optimusOptimizationSettingsManager.selectDoe(optimizationId, doeId);
        return getDoes(optimizationId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/does/selected", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> selectDoe(@PathVariable("optimizationId") int optimizationId,
                                                   @CurrentUser PortalUser currentUser) {
        Integer selectedDoeId = optimusOptimizationSettingsManager.findSelectedDoeId(optimizationId);
        if (selectedDoeId == null) {
            throw new InternalException("DoE не выбран");
        }
        return getDoeParams(optimizationId, selectedDoeId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/does/{doeId}/params", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getDoeParams(@PathVariable("optimizationId") int optimizationId,
                                                      @PathVariable("doeId") int doeId,
                                                      @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findSettingsByParent(optimizationId, doeId);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/does/params/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updateDoeParam(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                  @PathVariable("optimizationId") int optimizationId,
                                                  @PathVariable("paramId") int paramId,
                                                  @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }
}
