package ru.crp.lvms.optimizations.optimus.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.dao.ParametersDictionaryDao;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.optimus.dao.*;
import ru.crp.lvms.optimizations.optimus.entities.*;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toList;

@Component
public class OptimusDefaultParamsCreator {
    public static final int EGO_METHOD_ID = 22;
    public static final int SIMAN_METHOD_ID = 23;
    public static final int DOE_METHOD_ID = 24;

    public static final int LATIN_DOE_ID = 20;
    public static final int RANDOM_DOE_ID = 21;

    // value must be synchronized with cml-solver:MFZipFileTransformer.TIME_OUT_PARAM_NAME
    public static final String TIME_OUT_PARAM_NAME = "Тайм-аут";

    @Autowired
    private OptimusOptimizationConstraintsDao optimusOptimizationConstraintsDao;

    @Autowired
    private OptimusOptimizationTargetFuncsDao optimusOptimizationTargetFuncsDao;

    @Autowired
    private OptimusOptimizationBaseSettingsDao optimusOptimizationBaseSettingsDao;

    @Autowired
    private OptimusOptimizationVaryingParamsDao optimusOptimizationVaryingParamsDao;

    @Autowired
    private OptimusOptimizationCADVaryingParamsDao optimusOptimizationCADVaryingParamsDao;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    @Autowired
    private ParametersDictionaryDao parametersDictionaryDao;

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;


    public void addDefaultSettings(OptimizationEntity newOptimization) {
        List<OptimusOptimizationBaseSetting> initialBasicSettings = createOptimusDefaultSettings(newOptimization);
        initialBasicSettings.forEach(optimusOptimizationBaseSettingsDao::insert);

        List<OptimusOptimizationConstraint> initialConstraints = createOptimusConstraints(newOptimization);
        initialConstraints.forEach(optimusOptimizationConstraintsDao::insert);

        List<OptimusOptimizationTargetFunc> initialTargetFuncs = createOptimusTargetFuncs(newOptimization);
        initialTargetFuncs.forEach(optimusOptimizationTargetFuncsDao::insert);

        List<OptimusOptimizationVaryingParam> initialVaryingParams = createOptimusInitialVaryingParams(newOptimization);
        initialVaryingParams.forEach(optimusOptimizationVaryingParamsDao::insert);

        if (newOptimization.getSolverCadId() != null) {
            List<OptimusOptimizationCADVaryingParam> initialCADVaryingParams = createOptimusInitialCADVaryingParams(newOptimization);
            initialCADVaryingParams.forEach(optimusOptimizationCADVaryingParamsDao::insert);
        }

        List<OptimizationJsonSetting> runParams = createOptimusRunParams(newOptimization);
        runParams.forEach(optimizationJsonSettingsDao::createSettings);

        List<OptimizationJsonSetting> doeParams = createOptimusDoeParams(newOptimization);
        doeParams.forEach(optimizationJsonSettingsDao::createSettings);

        List<OptimizationJsonSetting> algorithmParams = createOptimusAlgorithmParams(newOptimization);
        algorithmParams.forEach(optimizationJsonSettingsDao::createSettings);

        Integer numTargetFuncs = initialBasicSettings.stream()
                .filter(s -> s.getTypeId() == OptimusBaseSettingType.NUM_TARGET_FUNCS.getId())
                .map(OptimusOptimizationBaseSetting::getCurrentValue)
                .findFirst().orElse(null);
        if (numTargetFuncs == null) {
            numTargetFuncs = 1;
        }
        optimusOptimizationSettingsManager.selectDefaultAlgorithm(newOptimization.getId(), numTargetFuncs);
    }

    public List<OptimizationJsonSetting> createOptimusAlgorithmParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        Map<Integer, List<ParametersDictionary>> methods = parametersDictionaryDao
                .findByTypeName(ParametersDictionaryType.OptimusOptimizationMethod).stream().collect(groupingBy(ParametersDictionary::getId));

        List<ParametersDictionary> methodsList = methods.getOrDefault(EGO_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'EGO' method");
        }
        ParametersDictionary egoMethod = methodsList.get(0);

        methodsList = methods.getOrDefault(SIMAN_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'SIMAN' method");
        }
        ParametersDictionary simanMethod = methodsList.get(0);

        methodsList = methods.getOrDefault(DOE_METHOD_ID, emptyList());
        if (methodsList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'SIMAN' method");
        }
        ParametersDictionary doeMethod = methodsList.get(0);


        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Число точек DoE", "nrexpfirstlh", "50"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Случайное число (неотрицательное)", "seed", "0"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Максимальное количество итераций", "maxnumeval", "500"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Критерий сходимости", "tolerance", "0.001"));
        settings.add(OptimizationJsonSetting.newListSetting(optId, egoMethod.getId(), "Тип метамодели", "modeltype",
                asList(new OptimizationListSetting(true, "Krigging", "Krigging",false,"0")
                    //    new OptimizationListSetting(false, "RBF", null,false,"1")
                )));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Фактор сглаживания", "nugget", "0"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, egoMethod.getId(), "Определение наилушчего Theta", "intensivesearch", false));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Частота обновления модели", "updatefreq", "1"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, egoMethod.getId(), "Фактор локализации поиска", "nrsigma", "3"));
        settings.add(OptimizationJsonSetting.newBoolSetting(optId, egoMethod.getId(), "Cохранение метамодели","savelastkrigingmodel", true));

        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Начальная температура", "inittemp", "10"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Скорость снижения температуры (0-1)", "annealfactor", "0.9"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Вариативность шага (Реком. 1.05-1.5)", "stepwidthadjfactor", "1.25"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Количество циклов (Реком. 3-5)", "numcycles", "4"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Параметр итераций (Реком. 3-5)", "iterattconst", "4"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Критерий сходимости по DV (0-1)", "avgstopwidth", "0.01"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, simanMethod.getId(), "Максимальное количество итераций", "maxeval", "1000"));
        return settings;
    }

    public List<OptimizationJsonSetting> createOptimusDoeParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        Map<Integer, List<ParametersDictionary>> does = parametersDictionaryDao
                .findByTypeName(ParametersDictionaryType.OptimusDoEType).stream().collect(groupingBy(ParametersDictionary::getId));

        List<ParametersDictionary> doesList = does.getOrDefault(20, emptyList());
        if (doesList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'LATIN' doe");
        }
        ParametersDictionary latinDoe = doesList.get(0);

        doesList = does.getOrDefault(21, emptyList());
        if (doesList.size() != 1) {
            throw new InternalException("Wrong does config. Can't find 'RANDOM' doe");
        }
        ParametersDictionary randomDoe = doesList.get(0);


        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, latinDoe.getId(), "Число точек (1-100)","numpoints", "3"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, latinDoe.getId(), "Отсчетная точка", "seed", "0"));

        settings.add(OptimizationJsonSetting.newPlainSetting(optId, randomDoe.getId(), "Число точек (1-100)", "numpoints", "3"));
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, randomDoe.getId(), "Отсчетная точка", "seed", "0"));
        return settings;
    }

    public List<OptimizationJsonSetting> createOptimusRunParams(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        List<OptimizationJsonSetting> settings = new ArrayList<>();
        settings.add(OptimizationJsonSetting.newPlainSetting(optId, TIME_OUT_PARAM_NAME, "timeout", ""));
        return settings;
    }


    public List<OptimusOptimizationVaryingParam> createOptimusInitialVaryingParams(OptimizationEntity newOptimization) {
        return IntStream.range(1, OptimusBaseSettingType.NUM_CAO_VARYING_PARAMS.getMaxValue() + 1)
                .mapToObj(num -> new OptimusOptimizationVaryingParam(newOptimization.getId(), "par" + num))
                .collect(toList());
    }

    public List<OptimusOptimizationCADVaryingParam> createOptimusInitialCADVaryingParams(OptimizationEntity newOptimization) {
        return IntStream.range(1, OptimusBaseSettingType.NUM_CAD_VARYING_PARAMS.getMaxValue() + 1)
                .mapToObj(num -> new OptimusOptimizationCADVaryingParam(newOptimization.getId(), "gpar" + num))
                .collect(toList());
    }

    public List<OptimusOptimizationTargetFunc> createOptimusTargetFuncs(OptimizationEntity newOptimization) {
        return IntStream.range(1, OptimusBaseSettingType.NUM_TARGET_FUNCS.getMaxValue() + 1)
                .mapToObj(num -> new OptimusOptimizationTargetFunc(newOptimization.getId(), "obj" + num, 1))
                .collect(toList());
    }

    public List<OptimusOptimizationConstraint> createOptimusConstraints(OptimizationEntity newOptimization) {
        return IntStream.range(1, OptimusBaseSettingType.NUM_CONSTRAINTS.getMaxValue() + 1)
                .mapToObj(num -> new OptimusOptimizationConstraint(newOptimization.getId(), "c" + num, 4))
                .collect(toList());
    }

    public List<OptimusOptimizationBaseSetting> createOptimusDefaultSettings(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        return Stream.of(OptimusBaseSettingType.values())
                // cad varying params is required only for optimizations with cad solver
                .filter(s -> newOptimization.getSolverCadId() != null || s != OptimusBaseSettingType.NUM_CAD_VARYING_PARAMS)
                .map(type -> new OptimusOptimizationBaseSetting(optId, type.getId(), 1))
                .collect(toList());
    }
}
