package ru.crp.lvms.optimizations.optistruct.managers;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructAdditionalParametersTypeDao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructOptimizationAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructOptimizationAdditionalParameter;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class OptistructAdditionalParametersManager {

    public final static int OPTIMIZATION = 1;
    public final static int POSTPROCESSING = 2;

    public final static Integer PLAIN_TYPE_ID = 1;

    @Autowired
    private OptistructAdditionalParametersDao optistructAdditionalParametersDao;

    @Autowired
    private OptistructAdditionalParametersTypeDao optistructAdditionalParametersTypeDao;

    @Autowired
    private OptistructOptimizationAdditionalParametersDao optistructOptimizationAdditionalParametersDao;

    public void createDefaultParameters(int optimizationId) {
        optistructAdditionalParametersDao.findAll()
                .stream()
                .map(parameter ->
                        Pair.of(parameter,
                                optistructAdditionalParametersTypeDao.findAllowedForParameter(parameter.getId()).get(0)))
                .forEach(pair -> {
                    OptistructOptimizationAdditionalParameter p = new OptistructOptimizationAdditionalParameter();
                    p.setOptimizationId(optimizationId);
                    p.setName(pair.getKey().getName());
                    p.setParameterId(pair.getKey().getId());
                    p.setParameterTypeId(pair.getValue().getId());
                    p.setValue(null);
                    optistructOptimizationAdditionalParametersDao.insert(p);
                });
    }

    public List<OptimizationJsonSetting> getAdditionalParameters(int optimizationId, int groupId) {
        return optistructOptimizationAdditionalParametersDao
                .findForOptimization(optimizationId)
                .stream()
                .filter(ent -> ent.getGroupId() == groupId)
                .map(this::additionalParameterToJsonSetting)
                .sorted(Comparator.comparing(OptimizationJsonSetting::getId))
                .collect(Collectors.toList());
    }

    public OptimizationJsonSetting updateAdditionalParameter(OptimizationJsonSetting updated, int optimizationId, int parameterId, int groupId) {
        updated.setId(parameterId);
        OptistructOptimizationAdditionalParameter parameter = jsonSettingToAdditionalParameter(updated);
        parameter.setGroupId(groupId);
        if (parameter.getParameterTypeId() != PLAIN_TYPE_ID) {
            parameter.setValue(null);
        }
        OptistructOptimizationAdditionalParameter afterUpdate = optistructOptimizationAdditionalParametersDao.updateAndRead(parameter);
        return additionalParameterToJsonSetting(afterUpdate);
    }

    private OptimizationJsonSetting additionalParameterToJsonSetting(OptistructOptimizationAdditionalParameter optimizationParameter) {
        OptimizationJsonSetting ret;
        if (optimizationParameter.getParameterTypeId() == PLAIN_TYPE_ID) {
            ret = OptimizationJsonSetting.newPlainSetting(optimizationParameter.getOptimizationId(), optimizationParameter.getName(), "unused", optimizationParameter.getValue());
        } else {
            List<OptimizationListSetting> settings = optistructAdditionalParametersTypeDao
                    .findAllowedForParameter(optimizationParameter.getParameterId())
                    .stream()
                    .map(type -> new OptimizationListSetting(false, type.getId(), type.getName(), type.getName()))
                    .peek(setting -> {
                        if (optimizationParameter.getParameterTypeId() == setting.getId()) setting.setSelected(true);
                    })
                    .collect(Collectors.toList());
            ret = OptimizationJsonSetting.newListSetting(optimizationParameter.getOptimizationId(), optimizationParameter.getName(), "unused", settings);
        }
        ret.setId(optimizationParameter.getId());
        return ret;
    }

    private OptistructOptimizationAdditionalParameter jsonSettingToAdditionalParameter(OptimizationJsonSetting setting) {
        OptistructOptimizationAdditionalParameter optimizationParameter = optistructOptimizationAdditionalParametersDao.findOne(setting.getId());
        if (optimizationParameter.getParameterTypeId() == PLAIN_TYPE_ID) {
            optimizationParameter.setValue(setting.getPlainValue());
        } else {
            Optional.ofNullable(setting.getListValue()).orElse(Collections.emptyList())
                    .stream()
                    .filter(OptimizationListSetting::isSelected)
                    .findAny()
                    .map(OptimizationListSetting::getId)
                    .ifPresent(optimizationParameter::setParameterTypeId);
            optimizationParameter.setValue(null);
        }
        return optimizationParameter;
    }
}
