package ru.crp.lvms.optimizations.optistruct.managers;

import org.springframework.stereotype.Component;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructCollector;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructLnSet;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructSet;
import ru.crp.lvms.optimizations.optistruct.enums.ComponentType;
import ru.crp.lvms.optimizations.optistruct.enums.LnSetType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructCollectorType;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.util.regex.Pattern.MULTILINE;

@Component
public class OptistructFemParser {

    private final static Pattern SET_NAME_AND_ID_PATTERN = Pattern.compile("^\\$HMNAME PROP.*(.)\"(.*)\".*$", MULTILINE);
    // order is important because some names are prefixes of another (ex. PBAR - PBARL)
    private final static Pattern SET_TYPE_PATTERN = Pattern.compile("^.*(PCOMP|PSHELL|PSOLID|PBARL|PROD|PBAR|PBEAML|PBEAM).*$", MULTILINE);

    private final static Pattern LNSET_NAME_AND_ID_PATTERN = Pattern.compile("^\\$HMSET\\s*([0-9]+)\\s*(.)\\s*\"(.*)\".*$", MULTILINE);

    private final static Pattern COLLECTOR_NAME_AND_ID_PATTERN = Pattern.compile("^\\$HMNAME LOADCOL\\s*([0-9]+)\"(.*)\".*$", MULTILINE);
    private final static Pattern FIXATION_INFO_PATTERN = Pattern.compile("^.*SPC\\s*([0-9]+).*$", MULTILINE);
    private final static Pattern INERTIA_INFO_PATTERN = Pattern.compile("^.*MPC\\s*([0-9]+).*$", MULTILINE);
    private final static Pattern MODAL_INFO_PATTERN = Pattern.compile("^.*EIGRL\\s*([0-9]+).*$", MULTILINE);
    private final static Pattern LOAD_INFO_PATTERN = Pattern.compile("^.*(FORCE|MOMENT|PLOAD|GRAV|LOAD)\\s*([0-9]+).*$", MULTILINE);

    public List<OptistructSet> extractSets(int optimizationId, Path pathToFile) throws IOException {
        List<String> lines = Files.readAllLines(pathToFile);
        List<OptistructSet> sets = new ArrayList<>();
        for (int cur = 0; cur < lines.size() - 2; cur++) {
            Matcher nameMatcher = SET_NAME_AND_ID_PATTERN.matcher(lines.get(cur));
            Matcher typeMatcher = SET_TYPE_PATTERN.matcher(lines.get(cur + 2));
            if (nameMatcher.find() && typeMatcher.find()) {
                sets.add(new OptistructSet(optimizationId, ComponentType.valueOf(typeMatcher.group(1)), nameMatcher.group(2)));
            }
        }
        return sets;
    }

    public List<OptistructLnSet> extractLnSets(int optimizationId, Path pathToFile) throws IOException {
        List<String> lines = Files.readAllLines(pathToFile);
        List<OptistructLnSet> lnsets = new ArrayList<>();
        for (int cur = 0; cur < lines.size() - 2; cur++) {
            Matcher nameMatcher = LNSET_NAME_AND_ID_PATTERN.matcher(lines.get(cur));
            if (nameMatcher.find()) {
                lnsets.add(new OptistructLnSet(optimizationId, LnSetType.valueOf(Integer.parseInt(nameMatcher.group(2))), nameMatcher.group(3)));
            }
        }
        return lnsets;
    }

    public List<OptistructCollector> extractCollectors(int optimizationId, Path pathToFile) throws IOException {
        Map<Integer, OptistructCollector> steps = new HashMap<>();
        try (BufferedReader reader = Files.newBufferedReader(pathToFile)) {
            String line;
            while ((line = reader.readLine()) != null) {
                Matcher nameMatcher = COLLECTOR_NAME_AND_ID_PATTERN.matcher(line);
                if (nameMatcher.find()) {
                    int stepId = Integer.valueOf(nameMatcher.group(1));
                    steps.put(stepId, new OptistructCollector(optimizationId, stepId, nameMatcher.group(2)));
                }

                setDataType(steps, INERTIA_INFO_PATTERN.matcher(line), 1, OptistructCollectorType.Inertia);
                setDataType(steps, FIXATION_INFO_PATTERN.matcher(line), 1, OptistructCollectorType.Fixation);
                setDataType(steps, MODAL_INFO_PATTERN.matcher(line), 1, OptistructCollectorType.Modal);
                setDataType(steps, LOAD_INFO_PATTERN.matcher(line), 2, OptistructCollectorType.Load);
            }
        }
        return new ArrayList<>(steps.values());
    }

    private void setDataType(Map<Integer, OptistructCollector> steps, Matcher matcher,
                             Integer stepIdIndex, OptistructCollectorType dataType) {
        if (matcher.find()) {
            Integer stepId = Integer.valueOf(matcher.group(stepIdIndex));
            if (steps.containsKey(stepId)) {
                steps.get(stepId).setDataType(dataType.name());
            }
        }
    }
}
