package ru.crp.lvms.optimizations.optistruct.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettignsDao;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optistruct.dao.*;
import ru.crp.lvms.optimizations.optistruct.entities.*;
import ru.crp.lvms.optimizations.optistruct.enums.*;

import javax.annotation.Nullable;
import java.util.*;
import java.util.function.Function;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;
import static ru.crp.lvms.optimizations.optistruct.managers.OptistructAdditionalParametersManager.PLAIN_TYPE_ID;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class OptistructOptimizationSettingsManager extends AbstractOptimizationSettingsManager<OptistructSettingsType> {

    public final static int DEFAULT_TF = 1;
    public static final int DEFAULT_TECH_CONSTRAINT = 1;

    @Autowired
    private OptistructOptimizationBaseSettingsDao optistructOptimizationBaseSettingsDao;

    @Autowired
    private OptistructAdditionalParametersManager optistructAdditionalParametersManager;

    @Autowired
    private OptistructOptimizationTFDao optistructOptimizationTFDao;

    @Autowired
    private OptistructOptimizationConstraintDao optistructOptimizationConstraintDao;

    @Autowired
    private OptistructOptimizationTechConstraintDao optistructOptimizationTechConstraintDao;

    @Autowired
    private OptistructOptimizationLnSetDao optistructOptimizationLnSetDao;

    @Autowired
    private OptistructOptimizationSetDao optistructOptimizationSetDao;

    @Autowired
    private OptistructSuggestManager optistructSuggestManager;

    @Autowired
    private OptistructCollectorDao optistructCollectorDao;

    @Autowired
    private OptistructLoadStepDao optistructLoadStepDao;

    @Autowired
    private OptistructOptimizationTechnologicalConstraintManager optistructOptimizationTechnologicalConstraintManager;

    @Autowired
    private OptistructTechnologicalConstraintTreeValueManager optistructTechnologicalConstraintTreeValueManager;

    @Autowired
    private OptistructOptimizationAdditionalParametersDao optistructOptimizationAdditionalParametersDao;

    @Override
    protected Map<OptistructSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> getSettingsDaos() {
        Map<OptistructSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>>
                settingsDaos = new EnumMap<>(OptistructSettingsType.class);
        settingsDaos.put(OptistructSettingsType.BaseSetting, optistructOptimizationBaseSettingsDao);
        settingsDaos.put(OptistructSettingsType.TargetFunction, optistructOptimizationTFDao);
        settingsDaos.put(OptistructSettingsType.Constraint, optistructOptimizationConstraintDao);
        settingsDaos.put(OptistructSettingsType.TechConstraint, optistructOptimizationTechConstraintDao);
        settingsDaos.put(OptistructSettingsType.Set, optistructOptimizationSetDao);
        settingsDaos.put(OptistructSettingsType.LNSet, optistructOptimizationLnSetDao);
        settingsDaos.put(OptistructSettingsType.LoadStep, optistructLoadStepDao);
        return settingsDaos;
    }

    public Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> findBaseSettings(int optimizationId,
                                                                                              UserEntity currentUser) {
        return optistructOptimizationBaseSettingsDao
                .findForOptimization(optimizationId).stream()
                .collect(toMap(s -> OptistructBaseSettingType.valueOf(s.getTypeId()), Function.identity()));
    }

    public void addDefaultSettings(OptimizationEntity newOptimization) {
//        TODO: Find out real amount of target functions and constraints
        createDefaultSettings(newOptimization).forEach(optistructOptimizationBaseSettingsDao::insert);
        optistructAdditionalParametersManager.createDefaultParameters(newOptimization.getId());
        createTechConstraints(newOptimization).forEach(optistructOptimizationTechConstraintDao::insert);
        createTargetFuncs(newOptimization).forEach(optistructOptimizationTFDao::insert);
        createConstraints(newOptimization).forEach(optistructOptimizationConstraintDao::insert);
        createLoadSteps(newOptimization).forEach(optistructLoadStepDao::insert);
    }

    private Stream<OptistructOptimizationLoadStep> createLoadSteps(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, OptistructBaseSettingType.NUM_LOAD_STEPS.getMaxValue()).mapToObj(num -> {
            OptistructOptimizationLoadStep ools = new OptistructOptimizationLoadStep();
            ools.setOptimizationId(newOptimization.getId());
            ools.setName("step" + num);
            ools.setCalculationType(OptistructLoadStepCalculationType.MODAL);
            ools.setScriptId(num);
            ools.setIsFixed(true);
            return ools;
        });
    }

    private Stream<OptistructOptimizationTF> createTargetFuncs(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, OptistructBaseSettingType.NUM_TARGET_FUNCS.getMaxValue())
                .mapToObj(num -> {
                    OptistructOptimizationTF tf = new OptistructOptimizationTF(newOptimization.getId(), "tf" + num, DEFAULT_TF);
                    tf.setAttaId(firstIdOrNull(optistructSuggestManager.findAllowedAtta(DEFAULT_TF)));
                    tf.setAttiId(firstIdOrNull(optistructSuggestManager.findAllowedAtti(DEFAULT_TF)));
                    tf.setOptimizationTypeId(firstIdOrNull(optistructSuggestManager.findAllTFOptimizationTypes()));
                    return tf;
                });
    }

    private Stream<OptistructOptimizationConstraint> createConstraints(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, OptistructBaseSettingType.NUM_CONSTRAINTS.getMaxValue())
                .mapToObj(num -> {
                    OptistructOptimizationConstraint c = new OptistructOptimizationConstraint(newOptimization.getId(), "c" + num, DEFAULT_TF);
                    c.setAttaId(firstIdOrNull(optistructSuggestManager.findAllowedAtta(DEFAULT_TF)));
                    c.setAttiId(firstIdOrNull(optistructSuggestManager.findAllowedAtti(DEFAULT_TF)));
                    return c;
                });
    }

    private Integer firstIdOrNull(List<? extends Entity> entities) {
        return entities.isEmpty() ? null : entities.get(0).getId();
    }

    private Stream<OptistructOptimizationTechnologicalConstraint> createTechConstraints(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, OptistructBaseSettingType.NUM_TECH_CONSTRAINTS.getMaxValue())
                .mapToObj(num -> new OptistructOptimizationTechnologicalConstraint(newOptimization.getId(), DEFAULT_TECH_CONSTRAINT, "tc" + num));
    }

    private Stream<OptistructOptimizationBaseSetting> createDefaultSettings(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        return Stream.of(OptistructBaseSettingType.values())
                .filter(type -> type.getId() != OptistructBaseSettingType.NUM_SETS.getId())
                .filter(type -> type.getId() != OptistructBaseSettingType.NUM_LNSETS.getId())
                .map(type -> new OptistructOptimizationBaseSetting(optId, type.getId(), 1));
    }

    public OptistructOptimizationBaseSetting getBaseSettingsByType(int optimizationId, OptistructBaseSettingType type) {
        OptistructOptimizationBaseSetting baseSetting = optistructOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, type.getId());
        if (baseSetting == null) {
            throw new BadRequestException("Базовые настройки с типом '" + type.getDisplayName() + "' для оптимизации " + optimizationId + " не найдены");
        }
        return baseSetting;
    }

    @Override
    public List<? extends AbstractOptimizationSettings> findOptimizationSettings(int optimizationId, OptistructSettingsType settingsType, UserEntity currentUser) {
        if (settingsType == OptistructSettingsType.TechConstraint) {
            return optistructOptimizationTechnologicalConstraintManager.findOptimizationTCs(optimizationId);
        } else {
            return super.findOptimizationSettings(optimizationId, settingsType, currentUser);
        }
    }

    @Override
    public AbstractOptimizationSettings findOptimizationSettingsById(int settingsId, OptistructSettingsType settingsType, UserEntity user) {
        if (settingsType == OptistructSettingsType.TechConstraint) {
            return optistructOptimizationTechnologicalConstraintManager.findOne(settingsId, null);
        } else {
            return super.findOptimizationSettingsById(settingsId, settingsType, user);
        }
    }

    @Override
    public <T extends AbstractOptimizationSettings> AbstractOptimizationSettings updateSettings(T updatedSettings, OptistructSettingsType settingsType) {
        if (settingsType == OptistructSettingsType.TargetFunction || settingsType == OptistructSettingsType.Constraint) {
            OptistructOptimizationTFOrConstraint updated = (OptistructOptimizationTFOrConstraint) updatedSettings;
            updated = deleteParameterValuesIfTypeNotPlain(updated);
            updated = updateAllowedParametersIfTfTypeChanged(updated, settingsType);
            return super.updateSettings(updated, settingsType);
        } else if (settingsType == OptistructSettingsType.TechConstraint) {
            return optistructOptimizationTechnologicalConstraintManager.updateAndRead((OptistructOptimizationTechnologicalConstraint) updatedSettings, null);
        } else {
            return super.updateSettings(updatedSettings, settingsType);
        }
    }

    private OptistructOptimizationTFOrConstraint updateAllowedParametersIfTfTypeChanged(OptistructOptimizationTFOrConstraint updated, OptistructSettingsType settingsType) {
        OptistructOptimizationTFOrConstraint old = (OptistructOptimizationTFOrConstraint) findOptimizationSettingsById(updated.getId(), settingsType, null);
        if (!Objects.equals(old.getTfId(), null) && !Objects.equals(old.getTfId(), updated.getTfId())) {
            updated.setAttaId(firstIdOrNull(optistructSuggestManager.findAllowedAtta(updated.getTfId())));
            updated.setAttiId(firstIdOrNull(optistructSuggestManager.findAllowedAtti(updated.getTfId())));
            updated.setAttaValue(null);
            updated.setAttiValue(null);
            if (settingsType == OptistructSettingsType.TargetFunction) {
                updated.setOptimizationTypeId(firstIdOrNull(optistructSuggestManager.findAllTFOptimizationTypes()));
            }
        }
        return updated;
    }

    private OptistructOptimizationTFOrConstraint deleteParameterValuesIfTypeNotPlain(OptistructOptimizationTFOrConstraint updated) {
        if (!Objects.equals(updated.getAttaId(), PLAIN_TYPE_ID)) {
            updated.setAttaValue(null);
        }
        if (!Objects.equals(updated.getAttiId(), PLAIN_TYPE_ID)) {
            updated.setAttiValue(null);
        }
        return updated;
    }

    public List<OptistructSet> findSelectedSets(Integer optimizationId, ComponentType componentType) {
        Integer componentTypeId = null;
        if (componentType != null) {
            componentTypeId = componentType.getId();
        }
        return optistructOptimizationSetDao.findSelectedSets(optimizationId, componentTypeId);
    }

    public List<OptistructLnSet> findSelectedLnSets(Integer optimizationId, LnSetType lnSetType) {
        Integer lnSetTypeId = null;
        if (lnSetType != null) {
            lnSetTypeId = lnSetType.getId();
        }
        return optistructOptimizationLnSetDao.findSelectedLnSets(optimizationId, lnSetTypeId);
    }

    public List<OptistructSet> suggestSets(Integer optimizationId, SuggestRequest suggestRequest, UserEntity currentUser) {
        suggestRequest.upsertQuery("optimization_id", optimizationId + "");
        List<Integer> alreadySelected = optistructOptimizationSetDao.findSelectedSets(optimizationId, null).stream()
                .map(OptistructSet::getId)
                .collect(toList());
        suggestRequest.setExcludedIds(alreadySelected);
        return optistructOptimizationSetDao.suggest(suggestRequest, currentUser.getId()).stream()
                .sorted(Comparator.comparing(AbstractOptimizationSettings::getName))
                .collect(toList());
    }

    public List<OptistructLnSet> suggestLnSets(Integer optimizationId, SuggestRequest suggestRequest, UserEntity currentUser) {
        suggestRequest.upsertQuery("optimization_id", optimizationId + "");
        List<Integer> alreadySelected = optistructOptimizationLnSetDao.findSelectedLnSets(optimizationId, null).stream()
                .map(OptistructLnSet::getId)
                .collect(toList());
        suggestRequest.setExcludedIds(alreadySelected);
        return optistructOptimizationLnSetDao.suggest(suggestRequest, currentUser.getId()).stream()
                .sorted(Comparator.comparing(AbstractOptimizationSettings::getName))
                .collect(toList());
    }

    public OptistructSet createSet(OptistructSet newSet) {
        return optistructOptimizationSetDao.insertAndRead(newSet);
    }

    public OptistructLnSet createLnSet(OptistructLnSet newLnSet) {
        return optistructOptimizationLnSetDao.insertAndRead(newLnSet);
    }

    public void resetSets(Integer optimizationId, List<OptistructSet> newSets) {
        optistructOptimizationSetDao.deleteAllForOptimization(optimizationId);
        optistructOptimizationSetDao.addSetsForOptimization(optimizationId, newSets);
    }

    public void resetLnSets(Integer optimizationId, List<OptistructLnSet> newLnSets) {
        optistructOptimizationLnSetDao.deleteAllForOptimization(optimizationId);
        optistructOptimizationLnSetDao.addLnSetsForOptimization(optimizationId, newLnSets);
    }

    public void resetCollectors(Integer optimizationId, List<OptistructCollector> newSteps) {
        optistructCollectorDao.deleteAllForOptimization(optimizationId);
        optistructCollectorDao.addCollectorsForOptimization(optimizationId, newSteps);
    }

    public List<OptistructCollector> findCollectorsForOptimizationByType(int optimizationId, @Nullable OptistructCollectorType collectorType) {
        return optistructCollectorDao.findAllForOptimizationByType(optimizationId, collectorType);
    }

    public void copySettings(int referenceOptimizationId, int newOptimizationId) {
        List<OptistructOptimizationBaseSetting> OptistructBaseSettings=optistructOptimizationBaseSettingsDao.findForOptimization(referenceOptimizationId);
        OptistructBaseSettings.forEach(tbs ->{
            tbs.setOptimizationId(newOptimizationId);
            optistructOptimizationBaseSettingsDao.insert(tbs);
        });
        List<OptistructOptimizationConstraint> OptistructConstraints=optistructOptimizationConstraintDao.findForOptimization(referenceOptimizationId);
        OptistructConstraints.forEach(tc ->{
            tc.setOptimizationId(newOptimizationId);
            optistructOptimizationConstraintDao.insert(tc);
        });
        List<OptistructOptimizationTF> OptistructTargetFunctions=optistructOptimizationTFDao.findForOptimization(referenceOptimizationId);
        OptistructTargetFunctions.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            optistructOptimizationTFDao.insert(ttf);
        });
        List<OptistructSet> OptistructSets=optistructOptimizationSetDao.findForOptimization(referenceOptimizationId);
        OptistructSets.forEach(ts ->{
            ts.setOptimizationId(newOptimizationId);
            optistructOptimizationSetDao.insert(ts);
        });
        List<OptistructLnSet> OptistructLnSets=optistructOptimizationLnSetDao.findForOptimization(referenceOptimizationId);
        OptistructLnSets.forEach(ts ->{
            ts.setOptimizationId(newOptimizationId);
            optistructOptimizationLnSetDao.insert(ts);
        });
        List<OptistructOptimizationTechnologicalConstraint> OptistructTechnologicalConstraints=optistructOptimizationTechConstraintDao.findForOptimization(referenceOptimizationId);
        OptistructTechnologicalConstraints.forEach(ttc ->{
            List<TechnologicalConstraintTreeNodeValue> technologicalConstraintTreeNodeValues = optistructTechnologicalConstraintTreeValueManager.findValuesOfTC(ttc.getId());
            ttc.setOptimizationId(newOptimizationId);
            OptistructOptimizationTechnologicalConstraint newOptistructTechCons =optistructOptimizationTechnologicalConstraintManager.createAndRead(ttc,null);
            technologicalConstraintTreeNodeValues.forEach(ttctv->{
                ttctv.setTechnologicalConstraintId(newOptistructTechCons.getId());
                optistructTechnologicalConstraintTreeValueManager.create(ttctv,null);
            });
        });
        List<OptistructOptimizationLoadStep> optistructLoadSteps=optistructLoadStepDao.findForOptimization(referenceOptimizationId);
        optistructLoadSteps.forEach(ts ->{
            ts.setOptimizationId(newOptimizationId);
            optistructLoadStepDao.insert(ts);
        });
        List<OptistructOptimizationAdditionalParameter> additionalParameters = optistructOptimizationAdditionalParametersDao.findForOptimization(referenceOptimizationId);
        additionalParameters.forEach((ts ->{
            ts.setOptimizationId(newOptimizationId);
            optistructOptimizationAdditionalParametersDao.insert(ts);
        }));
        List<OptistructCollector> collectors = optistructCollectorDao.findAllForOptimizationByType(referenceOptimizationId,null);
        collectors.forEach((ts ->{
            ts.setOptimizationId(newOptimizationId);
            optistructCollectorDao.insert(ts);
        }));
    }
}
