package ru.crp.lvms.optimizations.tosca.controllers;

import com.fasterxml.jackson.annotation.JsonView;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.controllers.JsonSettingsController;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.base.entities.SettingsView;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.enums.OptimizationTypes;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.tosca.config.ToscaConfigGenerator;
import ru.crp.lvms.optimizations.tosca.enums.ToscaJsonParametersParents;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.tasks.fileGenerators.ToscaInputFileGenerator;
import ru.crp.lvms.validators.task.optimization.CompositeOptimizationTaskParamsValidator;
import ru.crp.lvms.web.controllers.misc.FileHelper;

import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.URLs.TOSCA_SETTINGS_URL;

@RestController
@RequestMapping(value = TOSCA_SETTINGS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class ToscaJsonSettingsController extends JsonSettingsController {

    @Autowired
    private MFOptimizationSettingsManager MFOptimizationSettingsManager;

    @Autowired
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private CompositeOptimizationTaskParamsValidator optimizationTaskParamsValidator;

    @Autowired
    private ToscaConfigGenerator toscaConfigGenerator;


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/configFile", method = RequestMethod.GET)
    public void generateSettingsXMl(@PathVariable("optimizationId") int optimizationId, HttpServletResponse response,
                                    @CurrentUser PortalUser currentUser) throws IOException, JAXBException {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        if (optimization.getSolverCaoId() != OptimizationCao.Tosca.getId()) {
            throw new BadRequestException(String.format("Настройки '%s' недоступны", OptimizationCao.Tosca));
        }
        TaskEntity task = new TaskEntity();
        task.setParentId(optimizationId);
        task.setSolverName(OptimizationCao.Tosca.getSolverConfigName());
        task.setParentType(TaskParentType.optimization);
        Errors errors = new BeanPropertyBindingResult(task, task.getClass().getName());
        optimizationTaskParamsValidator.validate(task, errors);
        if (errors.hasErrors() && !(errors.getAllErrors().size() == 1 && errors.getAllErrors().get(0).getCode().equals("solvingAlreadyStarted"))) {
            throw new TaskStartingException(String.format("Не удалось создать файл '%s'", ToscaInputFileGenerator.SETTINGS_FILE_NAME), errors);
        }
        Path settings = toscaConfigGenerator.generateConfig(optimization);
        FileHelper.writeFileToResponse(settings, ToscaInputFileGenerator.SETTINGS_FILE_NAME, response);
        Files.deleteIfExists(settings);
    }

    @RequestMapping(value = "/methods", method = RequestMethod.GET)
    public List<OptimizationListSetting> getOptimizationMethods(@PathVariable("optimizationId") int optimizationId,
                                                                @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser.getUser());
        OptimizationTypes optType = OptimizationTypes.valueOf(optimization.getOptimizationTypeId());
        Integer optimizationMethodId = toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimizationId);
        return MFOptimizationSettingsManager.findByTypeName(ParametersDictionaryType.ToskaOptimizationMethod).stream()
                .map(method -> new OptimizationListSetting(false, method.getId(), method.getName(), method.getName()))
                .filter(a -> ToscaJsonParametersParents.valueOf(a.getId()).isMethodAvailableForOptimizationType(optType))
                .peek(a -> {
                    if (optimizationMethodId != null && optimizationMethodId.equals(a.getId())) a.setSelected(true);
                })
                .collect(toList());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/methods/{omId}", method = RequestMethod.PUT)
    public List<OptimizationListSetting> selectOptimizationMethod(@PathVariable("optimizationId") int optimizationId,
                                                                  @PathVariable("omId") int omId,
                                                                  @CurrentUser PortalUser currentUser) {
        toscaOptimizationSettingsManager.selectOptimizationMethod(optimizationId, omId);
        return getOptimizationMethods(optimizationId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/selected", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> selectAdditionalParameters(@PathVariable("optimizationId") int optimizationId,
                                                                    @CurrentUser PortalUser currentUser) {
        Integer optimizationMethodId = toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimizationId);
        if (optimizationMethodId == null) {
            throw new InternalException("Метод оптимизации не выбран");
        }
        return getAdditionalParameters(optimizationId, optimizationMethodId, currentUser);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/{omId}/params", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getAdditionalParameters(@PathVariable("optimizationId") int optimizationId,
                                                                 @PathVariable("omId") int omId,
                                                                 @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findSettingsByParent(optimizationId, omId);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/methods/params/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updateAdditionalParameters(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                              @PathVariable("optimizationId") int optimizationId,
                                                              @PathVariable("paramId") int paramId,
                                                              @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/postprocessingParams", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getPostprocessingParams(@PathVariable("optimizationId") int optimizationId,
                                                                 @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findSettingsByParent(optimizationId, ToscaJsonParametersParents.PostprocessingParameters.getId());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/postprocessingParams/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updatePostprocessingParam(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                             @PathVariable("optimizationId") int optimizationId,
                                                             @PathVariable("paramId") int paramId,
                                                             @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }

    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/animationParams", method = RequestMethod.GET)
    public List<OptimizationJsonSetting> getAnimationParams(@PathVariable("optimizationId") int optimizationId,
                                                            @CurrentUser PortalUser currentUser) {
        return jsonSettingsDao.findSettingsByParent(optimizationId, ToscaJsonParametersParents.AnimationParameters.getId());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @JsonView(SettingsView.Short.class)
    @RequestMapping(value = "/animationParams/{paramId}", method = RequestMethod.PUT)
    public OptimizationJsonSetting updateAnimationParam(@Valid @RequestBody OptimizationJsonSetting updatedModel,
                                                        @PathVariable("optimizationId") int optimizationId,
                                                        @PathVariable("paramId") int paramId,
                                                        @CurrentUser PortalUser currentUser) {
        return updateParam(optimizationId, paramId, updatedModel);
    }
}
