package ru.crp.lvms.optimizations.tosca.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationType;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNode;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettignsDao;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettingsManager;
import ru.crp.lvms.optimizations.tosca.dao.*;
import ru.crp.lvms.optimizations.tosca.entities.*;
import ru.crp.lvms.optimizations.tosca.enums.SetType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaSettingsType;

import javax.annotation.Nullable;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;
import static ru.crp.lvms.optimizations.base.entities.OptimizationType.*;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class ToscaOptimizationSettingsManager extends AbstractOptimizationSettingsManager<ToscaSettingsType> {

    public static final int DEFAULT_PARAMETER_ID_VALUE = 1;

    public static final int DEFAULT_CONSTRAINT_ID_VALUE = 1;

    @Autowired
    private ToscaOptimizationTechnologicalConstraintManager toscaOptimizationTechnologicalConstraintManager;

    @Autowired
    private ToscaOptimizationBaseSettingsDao toskaOptimizationBaseSettingsDao;

    @Autowired
    private ToscaOptimizationConstraintDao toscaOptimizationConstraintDao;

    @Autowired
    private ToscaOptimizationTargetFunctionDao toscaOptimizationTargetFunctionDao;

    @Autowired
    private ToscaOptimizationTechnologicalConstraintDao toscaOptimizationTechnologicalConstraintDao;

    @Autowired
    private ToscaOptimizationSetsDao toscaOptimizationSetsDao;

    @Autowired
    private OptimizationDao optimizationDao;

    @Autowired
    private ToscaTechnologicalConstraintTreeValueManager toscaTechnologicalConstraintTreeValueManager;

    @Override
    protected Map<ToscaSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> getSettingsDaos() {
        Map<ToscaSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>>
                settingsDaos = new EnumMap<>(ToscaSettingsType.class);
        settingsDaos.put(ToscaSettingsType.BaseSetting, toskaOptimizationBaseSettingsDao);
        settingsDaos.put(ToscaSettingsType.Constraint, toscaOptimizationConstraintDao);
        // is not used for now
        settingsDaos.put(ToscaSettingsType.Set, toscaOptimizationSetsDao);
        settingsDaos.put(ToscaSettingsType.TargetFunc, toscaOptimizationTargetFunctionDao);
        settingsDaos.put(ToscaSettingsType.TechnologicalConstraint, toscaOptimizationTechnologicalConstraintDao);
        return settingsDaos;
    }

    public Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> findBaseSettings(int optimizationId,
                                                                                    UserEntity currentUser) {
        Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> settingsByType = toskaOptimizationBaseSettingsDao
                .findForOptimization(optimizationId).stream()
                .collect(toMap(s -> ToscaBaseSettingType.valueOf(s.getTypeId()), Function.identity()));
        Stream.of(ToscaBaseSettingType.values()).forEach(type -> settingsByType.putIfAbsent(type, null));
        return settingsByType;
    }

    public void addDefaultSettings(OptimizationEntity newOptimization) {
//        TODO: Find out real amount of target functions and constraints
        createDefaultSettings(newOptimization).forEach(toskaOptimizationBaseSettingsDao::insert);
        // TODO add extrating sets from file
        createSets(newOptimization).forEach(toscaOptimizationSetsDao::insert);
        createConstraints(newOptimization).forEach(toscaOptimizationConstraintDao::insert);
        createTechConstraints(newOptimization).forEach(toscaOptimizationTechnologicalConstraintDao::insert);
        createTargetFuncs(newOptimization).forEach(toscaOptimizationTargetFunctionDao::insert);
    }

    private Stream<ToscaSet> createSets(OptimizationEntity newOptimization) {
        return Stream.concat(
                IntStream.rangeClosed(1, ToscaBaseSettingType.NUM_SETS.getMaxValue())
                        .mapToObj(num -> new ToscaSet(newOptimization.getId(), SetType.Elements, "ElementsSet" + num)),
                IntStream.rangeClosed(1, ToscaBaseSettingType.NUM_SETS.getMaxValue())
                        .mapToObj(num -> new ToscaSet(newOptimization.getId(), SetType.Nodes, "NodesSet" + num))
        );
    }

    private Stream<ToscaConstraint> createConstraints(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, ToscaBaseSettingType.NUM_CONSTRAINTS.getMaxValue())
                .mapToObj(num -> new ToscaConstraint(newOptimization.getId(), DEFAULT_PARAMETER_ID_VALUE, "constr" + num));
    }

    private Stream<ToscaOptimizationTechnologicalConstraint> createTechConstraints(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, ToscaBaseSettingType.NUM_TECH_CONSTRAINTS.getMaxValue())
                .mapToObj(num -> new ToscaOptimizationTechnologicalConstraint(newOptimization.getId(), selectDefaultConstraintForOptimization(newOptimization), "tc" + num));
    }

    private int selectDefaultConstraintForOptimization(OptimizationEntity newOptimization) {
        switch (newOptimization.getOptimizationTypeId()) {
            case TOPOLOGICAL_OPTIMIZATION_TYPE_ID:
                return 1;
            case TOPOGRAPHIC_OPTIMIZATION_TYPE_ID:
                return 8;
            case SHAPE_OPTIMIZATION_TYPE_ID:
                return 7;
            default:
                return 1;
        }
    }

    private Stream<ToscaTargetFunction> createTargetFuncs(OptimizationEntity newOptimization) {
        return IntStream.rangeClosed(1, ToscaBaseSettingType.NUM_TARGET_FUNCS.getMaxValue())
                .mapToObj(num -> new ToscaTargetFunction(newOptimization.getId(), DEFAULT_PARAMETER_ID_VALUE, "tf" + num));
    }

    private Stream<ToscaOptimizationBaseSetting> createDefaultSettings(OptimizationEntity newOptimization) {
        int optId = newOptimization.getId();
        return Stream.of(ToscaBaseSettingType.values())
                .filter(type -> type.getId() != ToscaBaseSettingType.NUM_SETS.getId())
                .map(type -> {
                    if (type.getId() == ToscaBaseSettingType.NUM_TECH_CONSTRAINTS.getId()) {
                        return new ToscaOptimizationBaseSetting(optId, type.getId(), 0);
                    }
                    if (type.getId() == ToscaBaseSettingType.NUM_CONSTRAINTS.getId() && newOptimization.getOptimizationTypeId() == OptimizationType.SHAPE_OPTIMIZATION_TYPE_ID) {
                        return new ToscaOptimizationBaseSetting(optId, type.getId(), 0);
                    }
                    return new ToscaOptimizationBaseSetting(optId, type.getId(), 1);
                });
    }

    public ToscaOptimizationBaseSetting getBaseSettingsByType(int optimizationId, ToscaBaseSettingType type) {
        ToscaOptimizationBaseSetting baseSetting = toskaOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, type.getId());
        if (baseSetting == null) {
            throw new BadRequestException("Базовые настройки с типом '" + type.getDisplayName() + "' для оптимизации " + optimizationId + " не найдены");
        }
        return baseSetting;
    }

    @Nullable
    public Integer findSelectedOptimizationMethodId(int optimizationId) {
        return optimizationDao.findSelectedOptimizationMethodId(optimizationId);
    }

    public void selectOptimizationMethod(int optimizationId, @Nullable Integer optMethId) {
        optimizationDao.selectOptimizationMethod(optimizationId, optMethId);
    }

    @Override
    public List<? extends AbstractOptimizationSettings> findOptimizationSettings(int optimizationId, ToscaSettingsType toscaSettingsType, UserEntity currentUser) {
        if (toscaSettingsType == ToscaSettingsType.TechnologicalConstraint) {
            return toscaOptimizationTechnologicalConstraintManager.findOptimizationTCs(optimizationId);
        } else {
            return super.findOptimizationSettings(optimizationId, toscaSettingsType, currentUser);
        }
    }

    @Override
    public <T extends AbstractOptimizationSettings> AbstractOptimizationSettings updateSettings(T updatedSettings, ToscaSettingsType toscaSettingsType) {
        if (toscaSettingsType == ToscaSettingsType.TechnologicalConstraint) {
            return toscaOptimizationTechnologicalConstraintManager.updateAndRead((ToscaOptimizationTechnologicalConstraint) updatedSettings, null);
        } else {
            return super.updateSettings(updatedSettings, toscaSettingsType);
        }
    }

    @Override
    public AbstractOptimizationSettings findOptimizationSettingsById(int settingsId, ToscaSettingsType settingsType, UserEntity user) {
        if (settingsType == ToscaSettingsType.TechnologicalConstraint) {
            return toscaOptimizationTechnologicalConstraintManager.findOne(settingsId, null);
        } else {
            return super.findOptimizationSettingsById(settingsId, settingsType, user);
        }
    }

    public List<ToscaSet> findSelectedSets(Integer optimizationId, @Nullable SetType setType) {
        Integer setTypeId = null;
        if (setType != null) {
            setTypeId = setType.getId();
        }
        return toscaOptimizationSetsDao.findSelectedSets(optimizationId, setTypeId);
    }

    public List<ToscaSet> suggestSets(int optimizationId, SuggestRequest suggestRequest, UserEntity currentUser) {
        suggestRequest.upsertQuery("optimization_id", optimizationId + "");
        List<Integer> alreadySelected = toscaOptimizationSetsDao.findSelectedSets(optimizationId, null).stream()
                .map(ToscaSet::getId)
                .collect(toList());
        suggestRequest.setExcludedIds(alreadySelected);
        return toscaOptimizationSetsDao.suggest(suggestRequest, currentUser.getId()).stream()
                .sorted((s1, s2) -> s1.getName().compareTo(s2.getName()))
                .collect(toList());
    }

    public void resetSets(Integer optimizationId, List<ToscaSet> newSets) {
        toscaOptimizationSetsDao.deleteAllForOptimization(optimizationId);
        toscaOptimizationSetsDao.addSetsForOptimization(optimizationId, newSets);
    }

    public void copySettings(int referenceOptimizationId, int newOptimizationId) {
        List<ToscaOptimizationBaseSetting> toscaBaseSettings=toskaOptimizationBaseSettingsDao.findForOptimization(referenceOptimizationId);
        toscaBaseSettings.forEach(tbs ->{
            tbs.setOptimizationId(newOptimizationId);
            toskaOptimizationBaseSettingsDao.insert(tbs);
        });
        List<ToscaConstraint> toscaConstraints=toscaOptimizationConstraintDao.findForOptimization(referenceOptimizationId);
        toscaConstraints.forEach(tc ->{
            tc.setOptimizationId(newOptimizationId);
            toscaOptimizationConstraintDao.insert(tc);
        });
        List<ToscaTargetFunction> toscaTargetFunctions=toscaOptimizationTargetFunctionDao.findForOptimization(referenceOptimizationId);
        toscaTargetFunctions.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            toscaOptimizationTargetFunctionDao.insert(ttf);
        });
        List<ToscaSet> toscaSets=toscaOptimizationSetsDao.findForOptimization(referenceOptimizationId);
        toscaSets.forEach(ts ->{
            ts.setOptimizationId(newOptimizationId);
            toscaOptimizationSetsDao.insert(ts);
        });
        List<ToscaOptimizationTechnologicalConstraint> toscaTechnologicalConstraints=toscaOptimizationTechnologicalConstraintDao.findForOptimization(referenceOptimizationId);
        toscaTechnologicalConstraints.forEach(ttc ->{
            List<TechnologicalConstraintTreeNodeValue> technologicalConstraintTreeNodeValues = toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(ttc.getId());
            ttc.setOptimizationId(newOptimizationId);
            ToscaOptimizationTechnologicalConstraint newToscaTechCons =toscaOptimizationTechnologicalConstraintDao.insertAndRead(ttc);
            technologicalConstraintTreeNodeValues.forEach(ttctv->{
                ttctv.setTechnologicalConstraintId(newToscaTechCons.getId());
                toscaTechnologicalConstraintTreeValueManager.create(ttctv,null);
            });
        });

    }
}
