package ru.crp.lvms.optimizations.tosca.results;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FilenameUtils;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.ChartFactory;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.chart.Point;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.resultsHandlers.KeyResultHandler;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.tosca.dao.ToscaTargetFunctionParametersDao;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationBaseSetting;
import ru.crp.lvms.optimizations.tosca.entities.ToscaTargetFunction;
import ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaSettingsType;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import static java.util.Collections.singletonList;

public class ToscaRawDataHandler extends KeyResultHandler {
    public static final String X_LABEL = "Номер итерации";

    private ObjectMapper objectMapper;
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;
    private ToscaTargetFunctionParametersDao toscaTargetFunctionParametersDao;

    public ToscaRawDataHandler(Path pathToStorage, ObjectMapper objectMapper,
                               ToscaOptimizationSettingsManager toscaOptimizationSettingsManager, ToscaTargetFunctionParametersDao toscaTargetFunctionParametersDao) {
        super(pathToStorage);
        this.objectMapper = objectMapper;
        this.toscaOptimizationSettingsManager = toscaOptimizationSettingsManager;
        this.toscaTargetFunctionParametersDao = toscaTargetFunctionParametersDao;
    }

    @Override
    public boolean canHandle(ResultType resultType, Path pathToFile) {
        return resultType == ResultType.toscaRawData && Files.isRegularFile(pathToFile);
    }

    @Override
    protected List<ResultEntity> doHandleResultsFile(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        ResultEntity newResult = createResultEntity(pathToFile, resultsParent, createdBy);
        try {
            ChartsSet chartsSet = parseFile(pathToFile);
            initCurveLabels(chartsSet, resultsParent);
            Path pathToJsonFile = createJsonFile(pathToFile);
            objectMapper.writeValue(pathToJsonFile.toFile(), chartsSet);
        } catch (IOException | NumberFormatException ex) {
            throw new ResultProcessingException("Can't create json file for curves data", ex);
        }

        return singletonList(newResult);
    }

    private void initCurveLabels(ChartsSet chartsSet, ResultsParent resultsParent) {
        Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings = toscaOptimizationSettingsManager.findBaseSettings(resultsParent.getId(), null);
        if (isTargetFunction(chartsSet.getTitle())) {
            if (baseSettings.get(ToscaBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue() == 1) {
                List<? extends AbstractOptimizationSettings> optimizationSettings = toscaOptimizationSettingsManager.findOptimizationSettings(resultsParent.getId(), ToscaSettingsType.TargetFunc, null);
                ToscaTargetFunction tf = (ToscaTargetFunction) optimizationSettings.get(0);
                chartsSet.setyAxisLabel(toscaTargetFunctionParametersDao.findOne(tf.getParameterId()).getName());
            } else {
                chartsSet.setyAxisLabel("Целевая функция");
            }
        }
    }

    private boolean isTargetFunction(String fileName) {
        return fileName.contains("OBJ_FUNC_REPORT");
    }

    private Path createJsonFile(Path rawDataFile) throws IOException {
        String jsonFileName = FilenameUtils.removeExtension(rawDataFile.getFileName().toString()) + ".json";
        Path sibling = rawDataFile.resolveSibling(jsonFileName);
        Files.deleteIfExists(sibling);
        return Files.createFile(sibling);
    }

    private ResultEntity createResultEntity(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        ResultEntity newResult = initCommonResultProperties(pathToFile, resultsParent, createdBy);
        if (createdBy == CreatedBy.Auto) {
            extractDescription(pathToFile, newResult);
        }

        newResult.setName(pathToFile.getFileName().toString());
        newResult.setType(ResultType.curves);
        return newResult;
    }

    private ChartsSet parseFile(Path rawDataFile) throws IOException {
        ChartsSet chartsSet = new ChartsSet(rawDataFile.getFileName().toString(), X_LABEL, "", "Y", "");
        LineChart chart = ChartFactory.createCurve();
        chartsSet.addChart(chart);
        Files.lines(rawDataFile)
                .map(line -> line.split(","))
                .peek(splitted -> {
                    if (splitted.length < 3) {
                        throw new ResultProcessingException(String.format("Can't create json file for curves data. Unexpected line format:", Arrays.toString(splitted)));
                    }
                })
                .map(splitted -> Point.commonPoint(Double.parseDouble(splitted[0]), Double.parseDouble(splitted[2])))
                .forEach(chart::addPoint);
        return chartsSet;
    }
}
