package ru.crp.lvms.tasks.fileGenerators;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@Component
public class OptimizationInputFileGenerator extends BaseInputFileGenerator {
    private static final Logger LOGGER = LoggerFactory.getLogger(OptimizationInputFileGenerator.class);

    @Autowired
    private MFInputFileGenerator mfInputFileGenerator;

    @Autowired
    private OptimusInputFileGenerator optimusInputFileGenerator;

    @Autowired
    private ToscaInputFileGenerator toscaInputFileGenerator;

    @Autowired
    private OptistructInputFileGenerator optistructInputFileGenerator;

    @Autowired
    private AnsysInputFileGenerator ansysInputFileGenerator;

    @Autowired
    private LsDynaInputFileGenerator lsDynaInputFileGenerator;

    @Autowired
    private NXNastranInputFileGenerator nxNastranInputFileGenerator;


    public List<FileData> getInputFilesForOptimization(TaskEntity taskEntity)
            throws IOException, ResourcePreparationException {
        if (taskEntity.getParentType() != TaskParentType.optimization) {
            throw new TaskStartingException("Unexpected task type: " + taskEntity.getParentType());
        }
        OptimizationEntity optimization = optimizationManager.findOne(taskEntity.getParentId(), null);
        if (optimization == null) {
            throw new NotFoundException(taskEntity.getParentId(), OptimizationEntity.class);
        }
        List<FileData> inputFiles = new ArrayList<>();
        if (optimization.getSolverCaoId() != null) {
            inputFiles.addAll(getCaoInputFiles(taskEntity, optimization));
        } else if (optimization.getSolverCaeId() != null) {
            inputFiles.addAll(getCaeInputFiles(taskEntity, optimization));
        }
        return inputFiles;
    }

    private List<FileData> getCaoInputFiles(TaskEntity taskEntity, OptimizationEntity optimization) throws IOException {
        switch (OptimizationCao.valueOf(optimization.getSolverCaoId())) {
            case Tosca:
                return toscaInputFileGenerator.getInputFilesForOptimization(taskEntity, optimization);
            case ModeFrontier:
                return mfInputFileGenerator.getInputFilesForOptimization(optimization);
            case Optimus:
                return optimusInputFileGenerator.getInputFilesForOptimization(optimization);
            case Optistruct:
                return optistructInputFileGenerator.getInputFilesForOptimization(taskEntity.getType(), optimization);
            default:
                throw new TaskStartingException(String.format("Unexpected solver id: %d", optimization.getSolverCaoId()));
        }
    }

    private List<FileData> getCaeInputFiles(TaskEntity taskEntity, OptimizationEntity optimization) throws IOException {
        switch (OptimizationCae.valueOf(optimization.getSolverCaeId())) {
            case ANSYS:
                return ansysInputFileGenerator.getInputFilesForOptimization(optimization);
            case LSDYNA: // both slurm and local cluster dyna
                return lsDynaInputFileGenerator.getInputFilesForOptimization(optimization);
            case NXNASTRAN:
                return nxNastranInputFileGenerator.getInputFilesForOptimization(optimization);
            case ABAQUS:
            case NASTRAN:
                return optimizationManager.getInputFilesForOptimization(optimization, null);
            default:
                throw new TaskStartingException(String.format("Unexpected solver id: %d for cluster: %d",
                        optimization.getSolverCaoId(), taskEntity.getClusterId()));
        }
    }
}
