package ru.crp.lvms.tasks.slurm;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.calculations.solvers.xml.config.Command;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.cmls.calculations.solvers.xml.config.arguments.CommandResolver;
import ru.crp.cmls.calculations.solvers.xml.config.arguments.CompositeArgumentResolver;
import ru.crp.cmls.calculations.solvers.xml.config.files.CompositeFileResolver;
import ru.crp.cmls.calculations.solvers.xml.config.placeholder.BasePlaceholdersResolver;
import ru.crp.cmls.calculations.solvers.xml.config.placeholder.SimplePlaceholderResolver;
import ru.crp.lvms.cluster.model.SlurmClusterConfigEntity;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.entities.tasks.SlurmTaskInfo;
import ru.crp.lvms.model.entities.tasks.TaskEntity;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static ru.crp.cmls.calculations.solvers.xml.config.placeholder.PlaceholderResolver.*;

@Service
public class SlurmScriptCreator {
    private static final String SLURM_START_ANSYS_TEMPLATE = "ansys-start.ftl";
    private static final String SLURM_START_MPI_TEMPLATE = "slurm-start-mpi.ftl";
    private static final String SLURM_START_SIMPLE_TEMPLATE = "slurm-start-simple.ftl";
    public static final String SLURM_START_SCRIPT_NAME = "slurm-start-script.sh";
    private static final int MAX_AVAILABLE_CORES = 24;
    public static final String INPUT_FILE_NAME = "file.inp";
    public static final String OUTPUT_FILE_NAME = "file.out";

    @Autowired
    private Configuration configuration; //configuration for freemarker templates

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private SlurmPathsResolver slurmPathsResolver;

    @Autowired
    private SlurmTaskInfoDao slurmTaskInfoDao;

    public void createSlurmScript(TaskEntity task, SlurmClusterConfigEntity clusterConfig, String pathToConfig,
                                  List<FileData> taskFiles, Path pathToScript) throws IOException, TemplateException, ResourcePreparationException {
        SlurmTaskInfo slurmTaskInfo = slurmTaskInfoDao.getSlurmTaskInfo(task.getId());

        Template template = getTemplateForTask(task);
        Map<String, Object> params = new HashMap<>();
        params.put("partitionName", clusterConfig.getPartitionName());

        Integer numOfCores = task.getNumOfCores();
        if (numOfCores == null || numOfCores <= 0 || numOfCores >= MAX_AVAILABLE_CORES) {
            numOfCores = MAX_AVAILABLE_CORES;
        }
        params.put("numOfCores", numOfCores + "");

        Integer numOfNodes = slurmTaskInfo.getNumOfNodes();
        if (numOfNodes == null || numOfNodes <= 0) {
            numOfNodes = 1;
        }
        params.put("numOfNodes", numOfNodes + "");

        params.put("isParallel", slurmTaskInfo.getIsParallel());
        params.put("inputFileName", INPUT_FILE_NAME);
        params.put("outputFileName", OUTPUT_FILE_NAME);
        params.put("outputDir", slurmPathsResolver.getOutputDirForTask(clusterConfig, task));
        params.put("stdOutFile", SlurmPathsResolver.STDOUT_FILE_NAME);
        params.put("stdErrFile", SlurmPathsResolver.STDERR_FILE_NAME);
        params.put("taskName", "task-" + task.getId());
        params.put("commandLine", generateCommandLine(task, clusterConfig, pathToConfig, taskFiles));
        template.process(params, Files.newBufferedWriter(pathToScript));
    }


    private String generateCommandLine(TaskEntity task, SlurmClusterConfigEntity clusterConfig, String pathToConfig, List<FileData> taskFiles) throws ResourcePreparationException {
        SolverConfig config = solverConfigManager.getSolverByName(task.getSolverName());
        Map<String, String> argumentsList = getCommandsList(task, clusterConfig, pathToConfig, config, taskFiles);
        return argumentsList.entrySet().stream()
                .map(entry -> entry.getKey() + " " + entry.getValue()).collect(Collectors.joining(" "));
    }

    private Map<String, String> getCommandsList(TaskEntity task, SlurmClusterConfigEntity clusterConfig,
                                                String pathToConfig, SolverConfig config, List<FileData> taskFiles) throws ResourcePreparationException {
        Path outputDir = Paths.get(slurmPathsResolver.getOutputDirForTask(clusterConfig, task));
        Path inputDir = Paths.get(slurmPathsResolver.getInputDirForTask(clusterConfig, task));
        Type type = task.getType();

        BasePlaceholdersResolver placeholderResolver = new BasePlaceholdersResolver(task.getParentName(), inputDir, outputDir);
        placeholderResolver.addResolver(new SimplePlaceholderResolver(CONFIG_FILE_PLACEHOLDER, () -> pathToConfig));
        placeholderResolver.addResolver(new SimplePlaceholderResolver(ADDITIONAL_ARGUMENTS_PLACEHOLDER, task::getAdditionalStartupArguments));

        placeholderResolver.addResolver(new SimplePlaceholderResolver(MEMORY_PLACEHOLDER, () -> {
            Long memory = task.getMemory();
            if (memory == null) {
                memory = 0L;
            }
            return config.getLimitedMemory(type, memory);
        }));
        placeholderResolver.addResolver(new SimplePlaceholderResolver(NUM_OF_CORES_PLACEHOLDER, task::getNumOfCores));


        CompositeFileResolver fileResolver = new CompositeFileResolver(placeholderResolver, inputDir, outputDir, taskFiles);
        CompositeArgumentResolver argumentResolver = new CompositeArgumentResolver(placeholderResolver, fileResolver);

        CommandResolver commandResolver = new CommandResolver(argumentResolver);
        Map<String, String> resolvedCommands = new HashMap<>();
        for (Command command : config.getCommandsList(type)) {
            String resolvedArguments = commandResolver.resolve(command);
            if (resolvedArguments != null) {
                resolvedCommands.put(command.getValue(), resolvedArguments);
            }
        }
        return resolvedCommands;
    }

    private Template getTemplateForTask(TaskEntity task) throws IOException {
        if (task.getSolverName().toLowerCase().startsWith("ansys")) {
            return configuration.getTemplate(SLURM_START_ANSYS_TEMPLATE);
        } else if (task.getSolverName().toLowerCase().startsWith("ls-dyna")) {
            return configuration.getTemplate(SLURM_START_MPI_TEMPLATE);
        } else {
            return configuration.getTemplate(SLURM_START_SIMPLE_TEMPLATE);
        }
    }
}
