package ru.crp.lvms.tasks.slurm;

import org.postgresql.jdbc4.Jdbc4Array;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.support.JdbcDaoSupport;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmlb.model.utils.EntityUtils;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.entities.tasks.SlurmTaskInfo;
import ru.crp.lvms.optimizations.base.mappers.SlurmTaskInfoRowMapper;

import javax.annotation.Nullable;
import javax.sql.DataSource;
import java.sql.Array;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import static java.util.stream.Collectors.toList;

@Transactional(propagation = Propagation.REQUIRED)
@Repository
public class SlurmTaskInfoDao extends JdbcDaoSupport {
    private static final String SELECT_SLRUM_TASK_INFO_SQL = "SELECT * FROM slurm_tasks_info WHERE task_id = ?";

    @Nullable
    public SlurmTaskInfo getSlurmTaskInfo(int taskId) {
        Object[] args = {taskId};
        List<SlurmTaskInfo> taskInfo = getJdbcTemplate().query(SELECT_SLRUM_TASK_INFO_SQL, args, new SlurmTaskInfoRowMapper());
        return taskInfo.isEmpty() ? null : taskInfo.get(0);
    }

    public void updateSlurmTaskInfo(int taskId, @Nullable SlurmTaskInfo slurmTaskInfo) {
        if (slurmTaskInfo == null) {
            getJdbcTemplate().update("DELETE FROM slurm_tasks_info WHERE task_id = ?", taskId);
        } else {
            getJdbcTemplate().update(con -> {
                PreparedStatement ps = con.prepareStatement(
                        "UPDATE slurm_tasks_info " +
                                "SET slurm_task_id = ?, nodes = ?,  num_of_nodes = ?, is_parallel = ?" +
                                "WHERE task_id = ?");
                Array tArray = con.createArrayOf("varchar", slurmTaskInfo.getNodeNames().toArray());
                ps.setInt(1, slurmTaskInfo.getSlurmTaskId());
                ps.setArray(2, tArray);
                ps.setInt(3, slurmTaskInfo.getNumOfNodes() == null ? 1 : slurmTaskInfo.getNumOfNodes().intValue());
                ps.setBoolean(4, slurmTaskInfo.getIsParallel() == null ? false : slurmTaskInfo.getIsParallel());
                ps.setInt(5, taskId);
                return ps;
            });
        }
    }

    public void createSlurmTaskInfo(SlurmTaskInfo slurmTaskInfo) {
        getJdbcTemplate().update(con -> {
            PreparedStatement ps = con.prepareStatement("INSERT INTO slurm_tasks_info(task_id, nodes, num_of_nodes, is_parallel)" +
                    " VALUES(?, ?, ?, ?)");
            Array tArray = con.createArrayOf("varchar", new String[0]);
            ps.setInt(1, slurmTaskInfo.getTaskId());
            ps.setArray(2, tArray);
            ps.setInt(3, slurmTaskInfo.getNumOfNodes() == null ? 1 : slurmTaskInfo.getNumOfNodes().intValue());
            ps.setBoolean(4, slurmTaskInfo.getIsParallel() == null ? false : slurmTaskInfo.getIsParallel());
            return ps;
        });
    }


    @Transactional(readOnly = true)
    public List<String> getBusySlurmNodeNames(int clusterId) {
        Set<CalculationStatus> workingStatuses = CalculationStatus.getWorkingStatuses();
        String INSql = EntityUtils.sqlForListParams(workingStatuses);
        List<Object> params = new ArrayList<>(workingStatuses.stream().map(CalculationStatus::getId).collect(toList()));
        params.add(0, clusterId);

        List<Jdbc4Array> allBusyNodeNames = getJdbcTemplate().queryForList(
                "SELECT sti.nodes FROM slurm_tasks_info sti JOIN tasks t ON t.id = sti.task_id " +
                        "WHERE t.cluster_id = ? AND t.status_id IN " + INSql,
                Jdbc4Array.class, params.toArray());
        return allBusyNodeNames.stream()
                .map(postgresArray -> {
                    try {
                        return (String[]) postgresArray.getArray();
                    } catch (SQLException ex) {
                        throw new InternalException("Unexpected slurm_task_info table format", ex);
                    }
                })
                .flatMap(Arrays::stream)
                .distinct()
                .distinct()
                .collect(toList());
    }

    @Autowired
    public void setDataSourceAutowired(DataSource dataSource) {
        super.setDataSource(dataSource);
    }
}
