package ru.crp.lvms.tasks.slurm.commands;

import freemarker.template.TemplateException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Service;
import org.springframework.util.concurrent.ListenableFuture;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.config.ConfigData;
import ru.crp.cmls.calculations.solvers.config.ConfigFileCreator;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.cmls.calculations.solvers.xml.config.placeholder.BasePlaceholdersResolver;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.cluster.model.SlurmClusterConfigEntity;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.tasks.SlurmTaskInfo;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.TaskFilesCollector;
import ru.crp.lvms.tasks.misc.BaseTaskCommand;
import ru.crp.lvms.tasks.slurm.SlurmConnectionManager;
import ru.crp.lvms.tasks.slurm.SlurmPathsResolver;
import ru.crp.lvms.tasks.slurm.SlurmScriptCreator;
import ru.crp.lvms.tasks.slurm.requests.ActionRequest;
import ru.crp.lvms.tasks.slurm.requests.StartTaskRequest;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import static java.util.Collections.emptyList;
import static java.util.Collections.singletonList;
import static java.util.stream.Collectors.toList;

@Service
public class SlurmStartTaskCommand extends BaseTaskCommand<Void, Void> {
    public static final Logger LOGGER = LoggerFactory.getLogger(SlurmStartTaskCommand.class);

    @Autowired
    private SlurmConnectionManager slurmConnectionManager;

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private SlurmPathsResolver slurmPathsResolver;

    @Autowired
    private ClustersManager clustersManager;

    @Autowired
    private TaskFilesCollector taskFilesCollector;

    @Autowired
    private SlurmScriptCreator slurmScriptCreator;

    @Autowired
    private OptimizationManager optimizationManager;

    @Override
    public boolean isTaskReadyForProcessing(TaskEntity task, ClusterEntity taskCluster) {
        return super.isTaskReadyForProcessing(task, taskCluster) && !isSlurmTaskStarted(task);
    }


    // all exceptions thrown by this method will be caught in ListenableFuture callback
    @Async("slurmStartTaskExecutor")
    @Override
    public ListenableFuture<Void> processTask(TaskEntity task, Object unused) throws Exception {
        SlurmClusterConfigEntity clusterConfig = clustersManager.getSlurmClusterConfig(task.getClusterId());
        prepareWorkingDir(task, clusterConfig);

        List<FileData> taskFiles = taskFilesCollector.collectTaskFiles(task);

        Path tempFile = Files.createTempFile(null, null);
        String configFileName = createConfigFile(task, clusterConfig, tempFile);
        taskFiles.add(FileData.fromFile(configFileName, DataType.SUBMODEL, tempFile.toFile()));

        String pathToConfigFileOnServer = slurmPathsResolver.getInputDirForTask(clusterConfig, task) + "/" + configFileName;
        Path slurmStartScript = generateSlurmStartScript(task, clusterConfig, pathToConfigFileOnServer, taskFiles);
        taskFiles.add(FileData.fromFile(SlurmScriptCreator.SLURM_START_SCRIPT_NAME, DataType.SUBMODEL, slurmStartScript.toFile()));

        for (FileData file : taskFiles) {
            Path pathToFile = file.getPathToFile();
            String name = file.getName();
            String pathOnServer = slurmPathsResolver.getInputDirForTask(clusterConfig, task) + "/" + name;
            slurmConnectionManager.uploadFile(clusterConfig, pathToFile, pathOnServer);
        }


        String pathToSlurmScriptOnServer = slurmPathsResolver.getInputDirForTask(clusterConfig, task) + "/" + SlurmScriptCreator.SLURM_START_SCRIPT_NAME;
        Integer slurmTaskId = slurmConnectionManager.execCommand(clusterConfig, new StartTaskRequest(pathToSlurmScriptOnServer));
        SlurmTaskInfo slurmTaskInfo = slurmTaskInfoDao.getSlurmTaskInfo(task.getId());
        slurmTaskInfo.setSlurmTaskId(slurmTaskId);
        slurmTaskInfoDao.updateSlurmTaskInfo(task.getId(), slurmTaskInfo);
        taskStatusUpdater.update(task);
        return new AsyncResult<>(null);
    }


    private void prepareWorkingDir(TaskEntity task, SlurmClusterConfigEntity clusterConfig) {
        String inputDir = slurmPathsResolver.getInputDirForTask(clusterConfig, task);
        String outputDir = slurmPathsResolver.getOutputDirForTask(clusterConfig, task);
        slurmConnectionManager.execCommand(clusterConfig, new ActionRequest(String.format("mkdir -p %s && mkdir -p %s", inputDir, outputDir)));
    }


    private String createConfigFile(TaskEntity task, SlurmClusterConfigEntity clusterConfig, Path destinationFile) throws ResourcePreparationException {
        SolverConfig config = solverConfigManager.getSolverByName(task.getSolverName());
        Type type = task.getType();

        Path outputDir = Paths.get(slurmPathsResolver.getOutputDirForTask(clusterConfig, task));
        Path inputDir = Paths.get(slurmPathsResolver.getInputDirForTask(clusterConfig, task));
        BasePlaceholdersResolver placeholderResolver = new BasePlaceholdersResolver(task.getParentName(), inputDir, outputDir);

        if (config.shouldCreateConfigFile(type)) {
            String unresolvedPathToConfig = config.getPathToConfigFile(type);
            String pathToConfig = placeholderResolver.resolve(unresolvedPathToConfig);
            if (pathToConfig == null) {
                throw new ResourcePreparationException(String.format("Unresolved path to config: %s", unresolvedPathToConfig));
            }
            try {
                ConfigData configData = taskAsConfigData(task);
                new ConfigFileCreator(config, type, placeholderResolver)
                        .createSolverConfigFile(destinationFile, configData);
            } catch (IOException ex) {
                throw new ResourcePreparationException(String.format("Can't create config file for task %d", task.getId()), ex);
            }
        }
        return config.getConfigFileName(type);
    }


    public Path generateSlurmStartScript(TaskEntity task, SlurmClusterConfigEntity clusterConfig, String pathToConfig, List<FileData> taskFiles) throws ResourcePreparationException {
        try {
            Path tempFile = Files.createTempFile(null, null);
            slurmScriptCreator.createSlurmScript(task, clusterConfig, pathToConfig, taskFiles, tempFile);
            return tempFile;
        } catch (IOException | TemplateException ex) {
            throw new ResourcePreparationException(String.format("Can't generate config file for task %d", task.getId()), ex);
        }
    }


    private ConfigData taskAsConfigData(TaskEntity task) throws IOException {
        List<FileData> inputFiles;

        if (task.getParentType() == TaskParentType.optimization) {
            inputFiles = optimizationManager.getInputFilesForOptimization(optimizationManager.findOne(task.getParentId(), null), singletonList(FileType.CAEFile));
        } else {
            inputFiles = emptyList();
        }
        return new ConfigData() {
            @Override
            public String getParentName() {
                return task.getParentName();
            }

            @Override
            public List<String> getSubmodelFileNames() {
                return inputFiles.stream().map(FileData::getName).collect(toList());
            }

            @Override
            public List<FileData> getFiles() {
                return inputFiles;
            }
        };
    }
}
