package ru.crp.lvms.tasks.space.commands;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Component;
import org.springframework.util.concurrent.ListenableFuture;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tasks.TaskStatusResponse;
import ru.crp.lvms.tasks.misc.BaseTaskCommand;
import ru.crp.lvms.tasks.space.exceptions.TaskNotFoundException;
import ru.crp.lvms.utils.Mapper;

import java.util.Date;
import java.util.Set;

@Component
public class SpaceCheckTaskCommand extends BaseTaskCommand<TaskStatusResponse, Void> {
    private static final Logger LOGGER = LoggerFactory.getLogger(SpaceCheckTaskCommand.class);

    @Override
    public boolean isTaskReadyForProcessing(TaskEntity task, ClusterEntity taskCluster) {
        if (!super.isTaskReadyForProcessing(task, taskCluster)) {
            return false;
        }
        // for space cluster waiting tasks are handled by start task commands
        if (CalculationStatus.valueOf(task.getStatusId()) == CalculationStatus.Waiting) {
            return false;
        }
        if (!checkNodeAlive(task)) {
            return false;
        }
        return true;
    }

    // all exceptions throwed by this method will be catched by ListenableFuture callback
    @Async("taskProgressExecutor")
    @Override
    public ListenableFuture<TaskStatusResponse> processTask(TaskEntity task, Object unused) throws Exception {
        int taskId = task.getId();
        NodeEntity node = nodesManager.findOne(task.getNodeId(), null);
        try {
            // status and progress may left the same but we need to change modification date to know
            // that node is still alive
            task.setModificationDate(new Date());
            TaskStatusResponse serviceTask = solverConnectionManager.sendCheckTaskRequest(node.getUrl(), taskId);
            CalculationStatus previousStatus = CalculationStatus.valueOf(task.getStatusId());
            CalculationStatus newStatus = CalculationStatus.valueOf(serviceTask.getStatus().getId());
            if (!isValidTransition(previousStatus, newStatus)) {
                LOGGER.error("Trying to apply invalid transition for task state. Task id = {}" +
                        " previous state = {}. New state = {}", taskId, previousStatus, newStatus);
                taskStatusUpdater.update(task);
                return new AsyncResult<>(serviceTask);
            }
            taskStatusUpdater.changeTaskStatus(task, newStatus);
            return new AsyncResult<>(serviceTask);
        } catch (TaskNotFoundException ex) {
            LOGGER.warn("Can't check status for task with id " + taskId, ex);
            taskStatusUpdater.changeTaskStatus(task, CalculationStatus.Failed);
            return new AsyncResult<>(Mapper.map(task, TaskStatusResponse.class));
        } finally {
            // if we considered node not alive but has any response from it update it's alive status
            if (!node.getIsAlive()) {
                nodesManager.setNodeAliveStatus(node, true);
            }
        }
    }


    // todo add more restrictions
    private boolean isValidTransition(CalculationStatus previousStatus, CalculationStatus newStatus) {
        Set<CalculationStatus> finalStates = CalculationStatus.getTerminalStatuses();
        if (previousStatus == CalculationStatus.ImportingResults && !finalStates.contains(newStatus)) {
            return false;
        }
        return true;
    }
}
