package ru.crp.lvms.tasks.space.commands;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Component;
import org.springframework.util.concurrent.ListenableFuture;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.api.ServiceTask;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.exceptions.solving.SolverException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.TaskFilesCollector;
import ru.crp.lvms.tasks.misc.BaseTaskCommand;
import ru.crp.lvms.tasks.space.exceptions.NodeNotResponseException;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.Utils;
import ru.crp.lvms.validators.task.optimization.CompositeOptimizationTaskParamsValidator;

import java.io.IOException;
import java.nio.file.Files;
import java.util.Objects;
import java.util.Optional;

import static ru.crp.cmls.api.TaskParentType.optimization;


@Component
public class SpaceStartTaskCommand extends BaseTaskCommand<Void, Void> {
    private static final Logger LOGGER = LoggerFactory.getLogger(SpaceStartTaskCommand.class);

    @Autowired
    private TaskManager taskManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private CompositeOptimizationTaskParamsValidator optimizationTaskParamsValidator;

    @Autowired
    private TaskFilesCollector taskFilesCollector;

    @Override
    public boolean isTaskReadyForProcessing(TaskEntity waitingTask, ClusterEntity taskCluster) {
        if (!super.isTaskReadyForProcessing(waitingTask, taskCluster)) {
            return false;
        }
        if (!findSuitableNodeForTask(waitingTask)) {
            return false;
        }
        NodeEntity workerNode = nodesManager.findOne(waitingTask.getNodeId(), null);
        if (!workerNode.getIsAlive() || !taskManager.isNodeFree(workerNode.getId())) {
            return false;
        }
        return true;
    }


    // all exceptions thrown by this method will be caught in ListenableFuture callback
    @Async("startTaskExecutor")
    @Override
    public ListenableFuture<Void> processTask(TaskEntity task, Object unused) throws Exception {
        int taskId = task.getId();
        NodeEntity workerNode = nodesManager.findOne(task.getNodeId(), null);
        verifyRequestedResources(task, workerNode);
        ServiceTask extendedTask = Mapper.map(task, ServiceTask.class);
        try {
            initTaskFiles(task, extendedTask);
        } catch (IOException e) {
            throw new SolverException("Init task files failed", e);
        }
        if (task.getParentType() == TaskParentType.optimization) {
            extendedTask.setCustomParameters(optimizationManager.getCustomParametersForTask(task));
        }

        String nodeUrl = workerNode.getUrl();
        solverConnectionManager.sendPrepareTaskRequest(nodeUrl, extendedTask);
        solverConnectionManager.sendTaskFiles(nodeUrl, extendedTask.getId(), extendedTask.getFiles());
        solverConnectionManager.sendStartTaskRequest(nodeUrl, extendedTask.getId());
        extendedTask.getFiles()
                .stream()
                .filter(f -> f.getType() == DataType.GENERATED)
                .forEach(f -> {
                    try {
                        Files.delete(f.getPathToFile());
                    } catch (IOException e) {
                        LOGGER.error("Unable to delete temporal file", e);
                    }
                });
        return new AsyncResult<>(null);
    }


    private void verifyRequestedResources(TaskEntity task, NodeEntity workerNode) throws NodeNotResponseException {
        NodeStatus nodeStatus = nodesManager.getNodeStatus(workerNode.getId());
        if (nodeStatus == null) {
            throw new NodeNotResponseException("No response from node");
        }
        if (task.getNumOfCores() == null || task.getNumOfCores() <= 0 || task.getNumOfCores() > nodeStatus.getAvailableProcessors()) {
            task.setNumOfCores(nodeStatus.getAvailableProcessors());
        }

        Long nodeMemory = Math
                .round((Utils.memoryInMegabytes(nodeStatus.getTotalPhysicalMemorySize()) * 0.8));

        if (task.getMemory() == null || task.getMemory() <= 0 || task.getMemory() > nodeMemory) {
            task.setMemory(nodeMemory);
        }
        taskManager.update(task, null);
    }


    private void initTaskFiles(TaskEntity task, ServiceTask extendedTask) throws ResourcePreparationException, IOException {
        TaskParentType parentType = extendedTask.getParentType();
        if (parentType == optimization) {
            Errors errors = new BeanPropertyBindingResult(task, task.getClass().getName());
            optimizationTaskParamsValidator.validate(task, errors);
            if (errors.hasErrors() && !(errors.getAllErrors().size() == 1 && errors.getAllErrors().get(0).getCode().equals("solvingAlreadyStarted"))) {
                throw new TaskStartingException("Exception while starting task", errors);
            }
            OptimizationEntity optimization = optimizationManager.findOne(task.getParentId(), null);
            if (Objects.equals(optimization.getSolverCaoId(), OptimizationCao.Tosca.getId())) {
                extendedTask.setAdditionalStartupArguments(optimizationManager.getAdditionalStartupArgumentsForTosca(optimization));
            }
        }
        extendedTask.addAll(taskFilesCollector.collectTaskFiles(task));
    }

    // returns true if task already has node id or there are free appropriate node for it
    private boolean findSuitableNodeForTask(TaskEntity waitingTask) {
        if (!waitingTask.getAutoSelectNode()) {
            return true;
        }
        Optional<NodeEntity> mostSuitableNodeForTask = nodesManager.getMostSuitableNodeForTask(waitingTask);
        if (!mostSuitableNodeForTask.isPresent()) {
            return false;
        }
        mostSuitableNodeForTask.ifPresent(node -> waitingTask.setNodeId(node.getId()));
        return true;
    }
}
