package ru.crp.lvms.utils;

import org.jetbrains.annotations.Contract;
import org.modelmapper.*;
import org.modelmapper.convention.MatchingStrategies;
import org.modelmapper.spi.MappingContext;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.cmls.api.ServiceTask;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.dto.SolverPresenceRequest;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.SimulationSampleFile;
import ru.crp.lvms.model.entities.*;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.nodes.ExtendedNode;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.tasks.SlurmTaskInfo;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tasks.TaskIn;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.base.enums.JsonSettingsType;
import ru.crp.lvms.optimizations.base.managers.OptimizationTypeManager;
import ru.crp.lvms.optimizations.mf.config.xml.VariableType;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationCADVaryingParam;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.mf.managers.MFPropertyTypeConverter;
import ru.crp.lvms.optimizations.optimus.config.xml.*;
//import ru.crp.lvms.optimizations.optimus.config.xml.DependencyList;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationCADVaryingParam;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationConstraint;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationTargetFunc;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationVaryingParam;
import ru.crp.lvms.optimizations.optimus.managers.OptimusPropertyTypeConverter;
import ru.crp.lvms.pdf.reports.ECReportInput;
import ru.crp.lvms.pdf.reports.SimulationReportInput;
import ru.crp.lvms.pdf.reports.vo.ReportInnerECParameters;
import ru.crp.lvms.pdf.reports.vo.ReportInnerSimulationParameters;
import ru.crp.lvms.web.dto.CompositeMaterialIn;
import ru.crp.lvms.web.dto.MaterialIn;
import ru.crp.lvms.web.dto.OptimizationsGraphEntry;
import ru.crp.lvms.web.dto.user.UserCreate;
import ru.crp.lvms.web.dto.user.UserIn;
import ru.crp.lvms.web.dto.user.UserOut;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.joining;

@Component
public final class Mapper {
    private static ModelMapper modelMapper;

    @Autowired
    private MappingHelpers mappingHelpers;

    @Autowired
    private OptimizationTypeManager optimizationTypeManager;

    @Autowired
    private MFPropertyTypeConverter mfPropertyTypeConverter;

    @Autowired
    private OptimusPropertyTypeConverter optimusPropertyTypeConverter;

    // Private constructor prevents instantiation from other classes
    private Mapper() {
    }

    public AbstractConverter<Map<?, Optional<?>>, String> mapToString = new AbstractConverter<Map<?, Optional<?>>, String>() {
        @Override
        protected String convert(Map<?, Optional<?>> source) {
            return source.entrySet().stream()
                    .filter((entry) -> entry.getValue().isPresent())
                    .map(entry -> entry.getKey() + ": " + entry.getValue())
                    .collect(joining(","));
        }
    };

    @PostConstruct
    void init() {
        modelMapper = new ModelMapper();
        modelMapper.getConfiguration().setMatchingStrategy(MatchingStrategies.STRICT);
//        modelMapper.getConfiguration().setPropertyCondition(mappingHelpers.skipSecureFields);

        modelMapper.addMappings(new PropertyMap<UserCreate, UserEntity>() {
            @Override
            protected void configure() {
                skip().setPassword(null);
                map().setActive(true);
                map().setNotified(true);
                skip().setRole(null);
                skip().setCreationDate(null);
                skip().setId(0);
            }
        });

        modelMapper.addMappings(new PropertyMap<UserIn, UserEntity>() {
            @Override
            protected void configure() {
                skip().setPassword(null);
                skip().setLogin(null);
                skip().setActive(false);
                skip().setRole(null);
                skip().setCreationDate(null);
                skip().setId(0);
            }
        });

        modelMapper.addMappings(new PropertyMap<UserEntity, UserOut>() {
            @Override
            protected void configure() {
                using(mappingHelpers.idToUserRoleName).map(source.getRoleId(), destination.getRole());
                using(mappingHelpers.idToUserRoleDisplayName).map(source.getRoleId(), destination.getRoleDisplayName());
            }
        });

        modelMapper.addMappings(new PropertyMap<MaterialIn, MaterialEntity>() {
            @Override
            protected void configure() {
                skip().setModificationDate(null);
                skip().setCreationDate(null);
                skip().setOwnerId(0);
                skip().setId(0);
                skip().setOwner(null);
                skip().setMaterialFiles(emptyList());
                skip().setParameterValues(emptyList());
                skip().setPathNames(null);
                skip().setType(null);
                skip().setPath(emptyList());
            }
        });

        modelMapper.addMappings(new PropertyMap<CompositeMaterialIn, CompositeMaterialEntity>() {
            @Override
            protected void configure() {
                skip().setModificationDate(null);
                skip().setCreationDate(null);
                skip().setOwnerId(0);
                skip().setId(0);
                skip().setOwner(null);
                skip().setMaterialFiles(null);
                skip().setParameterValues(null);
                skip().setPathNames(null);
                skip().setPath(emptyList());
                skip().setMatrix(null);
                skip().setFiber(null);
                skip().setTechnology(null);
                skip().setEquipment(null);
                skip().setLayers(null);
                skip().setSampleFiles(null);
                skip().setTechnologyType(null);
                skip().setEpoxide(null);
            }
        });
        modelMapper.addMappings(new PropertyMap<OptimizationEntity, OptimizationsGraphEntry>() {
            @Override
            protected void configure() {
                map(source.getOptimizationType(), destination.getType());
            }
        });
        modelMapper.addMappings(new PropertyMap<ConstructionElementEntity, OptimizationsGraphEntry>() {
            @Override
            protected void configure() {
                map().setType("Исходный ЭК");
            }
        });

        modelMapper.addMappings(new PropertyMap<FileEntity, SimulationSampleFile>() {
            @Override
            protected void configure() {
                skip().setStandardIds(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<SolverConfig, SolverPresenceRequest>() {
            @Override
            protected void configure() {
                map(source.getName(), destination.getSolverName());
                map(source.getEnvVariables(), destination.getEnvVariables());
            }
        });

        modelMapper.addMappings(new PropertyMap<SolverConfig, Solver>() {
            @Override
            protected void configure() {
                using(mapToString).map(source.getMemoryLimits(), destination.getMemoryLimit());
                map(source.getTaskTypes(), destination.getTaskTypes());
                map(source.getName(), destination.getId());
            }
        });

        modelMapper.addMappings(new PropertyMap<NodeEntity, ExtendedNode>() {
            @Override
            protected void configure() {
                skip().setHasSolvingTask(null);
                skip().setStatus(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<SimulationEntity, ResultsParent>() {
            @Override
            protected void configure() {
                map().setType(TaskParentType.simulation);
            }
        });

        modelMapper.addMappings(new PropertyMap<ECEntity, ResultsParent>() {
            @Override
            protected void configure() {
                map().setType(TaskParentType.effectiveCharacteristic);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimizationEntity, ResultsParent>() {
            @Override
            protected void configure() {
                map().setType(TaskParentType.optimization);
            }
        });

        modelMapper.addMappings(new PropertyMap<TaskEntity, ResultsParent>() {
            @Override
            protected void configure() {
                skip().setName(null);
                map(source.getParentType(), destination.getType());
                map(source.getParentId(), destination.getId());
            }
        });

        modelMapper.addMappings(new PropertyMap<TaskEntity, ServiceTask>() {
            @Override
            protected void configure() {
                skip().setCustomParameters(new HashMap<>());
            }
        });


        modelMapper.addMappings(new PropertyMap<TaskIn, TaskEntity>() {
            @Override
            protected void configure() {
                skip().setOwnerId(0);
                skip().setModificationDate(null);
                skip().setCreationDate(null);
                skip().setId(0);
                skip().setStatusId(0);
                skip().setQueuePriority(0);
                skip().setTaskPosition(null);
                skip().setProgress(null);
                skip().setStartExecDate(null);
                skip().setParentName(null);
                skip().setStatus(null);
                skip().setNode(null);
                skip().setOwner(null);
                skip().setAutoSelectNode(false);
                skip().setCalculationFinishedWithException(false);
                skip().setPrevTaskId(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<TaskIn, SlurmTaskInfo>() {
            @Override
            protected void configure() {
                skip().setTaskId(0);
                skip().setSlurmTaskId(null);
                skip().setNodeNames(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<ECReportInput, ReportInnerECParameters>() {
            @Override
            protected void configure() {
                using(mappingHelpers.toCompositeParameters).map(source.getMicroParams(), destination.getParammicro());
                using(mappingHelpers.toCompositeParameters).map(source.getMesoParams(), destination.getParammeso());
                map(source.getEcName(), destination.getName());
                using(mappingHelpers.toReportDate).map(source.getCreationDate(), destination.getDate());
                map(source.getUserName(), destination.getUser());
                using(mappingHelpers.fromECDefectEntity).map(source.getDefects(), destination.getDefects());
                using(mappingHelpers.fromPairsList).map(source.getEcMezostructure(), destination.getEcmeso());
                using(mappingHelpers.fromPairsList).map(source.getEcMicrostructure(), destination.getEcmicro());
                using(mappingHelpers.pathToStringAbsolute).map(source.getPathToMicrostructurePicture(), destination.getMicroPic());
                using(mappingHelpers.pathToStringAbsolute).map(source.getPathToMesostructurePicture(), destination.getMesoPic());
            }
        });

        modelMapper.addMappings(new PropertyMap<SimulationReportInput, ReportInnerSimulationParameters>() {
            @Override
            protected void configure() {
                map(source.getTypeName(), destination.getTypeName());
                map(source.getStandartName(), destination.getStandartName());
                map(source.getMaterialName(), destination.getMaterialName());
                using(mappingHelpers.toReportDate).map(source.getCreationDate(), destination.getCreationDate());
                map(source.getUserName(), destination.getUserName());
                using(mappingHelpers.pathToStringAbsolute).map(source.getPicture(), destination.getPicture());
                using(mappingHelpers.pathToStringAbsolute).map(source.getFiberPic(), destination.getFiberPic());
                using(mappingHelpers.pathToStringAbsolute).map(source.getMatrixPic(), destination.getMatrixPic());
                map(String.valueOf(source.getManualSizeH()), destination.getManualSizeH());
                map(String.valueOf(source.getManualSizeB()), destination.getManualSizeB());
                map(String.valueOf(source.getManualSizeL()), destination.getManualSizeL());
                using(mappingHelpers.fromPairsList).map(source.getResults(), destination.getResults());
            }
        });

        // Pay attention that name is intentionally null
        modelMapper.addMappings(new PropertyMap<MFOptimizationVaryingParam, VariableType>() {
            @Override
            protected void configure() {
                map().setClazz("it.esteco.procint.design.DesignInputVariable");
                map().setName(null);
                map().setKind("VARIABLE");
                using(mappingHelpers.toString).map(source.getLowerBound(), destination.getLowerbound());
                using(mappingHelpers.toString).map(source.getUpperBound(), destination.getUpperbound());
                using(mappingHelpers.toString).map(source.getSampling(), destination.getBase());
                map().setDistribution("NONE");
                map().setArrangement("ORDERED");
                map().setTolerance("0.0");
                skip().setValue(null);
                skip().setFormat(null);
                skip().setScale(null);
                skip().setExpression(null);
                skip().setValueAttribute(null);
                skip().setShape1(null);
                skip().setShape2(null);
                skip().setError(null);
                skip().setLimit(null);
            }
        });

        //OPTIMUS
        // Pay attention that name is intentionally null
        modelMapper.addMappings(new PropertyMap<OptimusOptimizationVaryingParam, FactorList>() {
            @Override
            protected void configure() {
                map().setName(null);
                map().setRange("LOW/HIGH");
                map().setType("REAL");
                map(source.getLowerBound(), destination.getLow());
                map(source.getUpperBound(), destination.getHigh());
                map(source.getUpperBound(), destination.getNominal());
                map().setStep(null);
                map().setVar(" ");


            }
        });

        // Pay attention that name is intentionally null
        modelMapper.addMappings(new PropertyMap<MFOptimizationCADVaryingParam, VariableType>() {
            @Override
            protected void configure() {
                map().setClazz("it.esteco.procint.design.DesignInputVariable");
                map().setName(null);
                map().setKind("VARIABLE");
                using(mappingHelpers.toString).map(source.getLowerBound(), destination.getLowerbound());
                using(mappingHelpers.toString).map(source.getUpperBound(), destination.getUpperbound());
                using(mappingHelpers.toString).map(source.getSampling(), destination.getBase());
                map().setDistribution("NONE");
                map().setArrangement("ORDERED");
                map().setTolerance("0.0");
                skip().setValue(null);
                skip().setFormat(null);
                skip().setScale(null);
                skip().setExpression(null);
                skip().setValueAttribute(null);
                skip().setShape1(null);
                skip().setShape2(null);
                skip().setError(null);
                skip().setLimit(null);
            }
        });

  //OPTIMUS
        // Pay attention that name is intentionally null
        modelMapper.addMappings(new PropertyMap<OptimusOptimizationCADVaryingParam, FactorList>() {
            @Override
            protected void configure() {
                map().setName(null);
                map().setRange(null);
                map().setType("REAL");
                map(source.getLowerBound(), destination.getLow());
                map(source.getUpperBound(), destination.getHigh());
                map(source.getSampling(), destination.getNominal());
                skip().setVar(null);
                skip().setStep(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<MFOptimizationTargetFunc, VariableType>() {
            @Override
            protected void configure() {
                map().setName(null);
                map().setExpression(null);
                map().setClazz("it.esteco.frontier.design.DesignObjective");
                using(mappingHelpers.targetFuncToXml).map(source.getFuncId(), destination.getKind());
                skip().setFormat(null);
                skip().setLimit(null);
                skip().setScale(null);
                skip().setValueAttribute(null);
                skip().setLowerbound(null);
                skip().setUpperbound(null);
                skip().setBase(null);
                skip().setDistribution(null);
                skip().setShape1(null);
                skip().setShape2(null);
                skip().setArrangement(null);
                skip().setTolerance(null);
                skip().setError(null);
                skip().setValue(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimusOptimizationTargetFunc, ResponseList>() {
            @Override
            protected void configure() {
                map().setName(null);
                skip().setLow(null);
                skip().setHigh(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<MFOptimizationConstraint, VariableType>() {
            @Override
            protected void configure() {
                map().setName(null);
                map().setExpression(null);
                map().setClazz("it.esteco.frontier.design.DesignConstraint");
                using(mappingHelpers.constraintToXml).map(source.getFuncId(), destination.getKind());
                using(mappingHelpers.toString).map(source.getValue(), destination.getLimit());
                map().setTolerance("0.0");
                skip().setValue(null);
                skip().setFormat(null);
                skip().setScale(null);
                skip().setValueAttribute(null);
                skip().setLowerbound(null);
                skip().setUpperbound(null);
                skip().setBase(null);
                skip().setDistribution(null);
                skip().setShape1(null);
                skip().setShape2(null);
                skip().setArrangement(null);
                skip().setError(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimusOptimizationConstraint, ResponseList>() {
            @Override
            protected void configure() {
                map().setName(null);
                map().setLow(null);
                map().setHigh(null);

            }
        });


        modelMapper.addMappings(new PropertyMap<OptimusOptimizationVaryingParam, VarList>() {
            @Override
            protected void configure() {
                map().setVar(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimusOptimizationVaryingParam, DepVarList>() {
            @Override
            protected void configure() {
                map().setDepvar(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimusOptimizationCADVaryingParam, VarList>() {
            @Override
            protected void configure() {
                map().setVar(null);
            }
        });

        modelMapper.addMappings(new PropertyMap<OptimusOptimizationCADVaryingParam, DepVarList>() {
            @Override
            protected void configure() {
                map().setDepvar(null);
            }
        });


        modelMapper.addConverter(new Converter<LinkedHashMap, OptimizationListSetting>() {
            @Override
            public OptimizationListSetting convert(MappingContext<LinkedHashMap, OptimizationListSetting> context) {
                LinkedHashMap source = context.getSource();
                OptimizationListSetting dest = context.getDestination();
                dest.setId((Integer) source.getOrDefault("id", null));
                dest.setSelected((Boolean) source.getOrDefault("selected", false));
                dest.setName((String) source.getOrDefault("name", null));
                dest.setXmlName((String) source.getOrDefault("xmlName", null));
                dest.setCustom((Boolean) source.getOrDefault("custom", null));
                dest.setValue((String) source.getOrDefault("value", null));
                return dest;
            }
        });
        modelMapper.addConverter(new Converter<LinkedHashMap, OptimizationJsonSetting>() {
            @Override
            public OptimizationJsonSetting convert(MappingContext<LinkedHashMap, OptimizationJsonSetting> context) {
                LinkedHashMap source = context.getSource();
                OptimizationJsonSetting dest = context.getDestination();
                dest.setId((Integer) source.getOrDefault("id", null));
                dest.setName((String) source.getOrDefault("name", null));
                dest.setType(JsonSettingsType.valueOf(source.getOrDefault("type", "plain").toString()));
                dest.setXmlName((String) source.getOrDefault("xmlName", null));
                dest.setDefaultValue(source.getOrDefault("defaultValue", null));
                dest.setOptimizationId((int) source.getOrDefault("optimizationId", 0));
                dest.setValue(source.getOrDefault("value", null));
                return dest;
            }
        });

        modelMapper.addConverter(mfPropertyTypeConverter);

        modelMapper.addConverter(optimusPropertyTypeConverter);
    }


    @Contract("null, _ -> null")
    public static <D> D map(@Nullable Object source, Class<D> destinationType) {
        if (source == null) {
            return null;
        }
        int oldSize = mapper().getTypeMaps().size();
        D result = mapper().map(source, destinationType);
        if (oldSize < mapper().getTypeMaps().size()) {
            mapper().validate();
        }
        return result;
    }


    public static void map(@Nullable Object source, Object destination) {
        int oldSize = mapper().getTypeMaps().size();
        mapper().map(source, destination);
        if (oldSize < mapper().getTypeMaps().size()) {
            mapper().validate();
        }
    }


    @Contract("null, _ -> null")
    public static <D> D map(@Nullable Object source, TypeToken<D> typeToken) {
        if (source == null) {
            return null;
        }
        int oldSize = mapper().getTypeMaps().size();
        D result = mapper().map(source, typeToken.getType());
        if (oldSize < mapper().getTypeMaps().size()) {
            mapper().validate();
        }
        return result;
    }


    public static void validate() {
        mapper().validate();
    }


    public static ModelMapper mapper() {
        return modelMapper;
    }
}
