package ru.crp.lvms.web.controllers;

import org.modelmapper.TypeToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.NodesManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.model.entities.nodes.ExtendedNode;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tasks.TaskShort;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;

import javax.validation.Valid;
import java.util.Comparator;
import java.util.List;

import static ru.crp.lvms.utils.URLs.NODES_URL;

@RestController
@RequestMapping(value = NODES_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class NodesController extends CrudControllerSuggest<NodeEntity, NodeEntity, NodeEntity> {
    @Autowired
    private NodesManager nodesManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private TaskManager taskManager;

    @PreAuthorize("hasRole('ROLE_ADMIN')")
    @RequestMapping(method = RequestMethod.POST)
    public NodeEntity create(@Valid @RequestBody NodeEntity newNode, @CurrentUser PortalUser user) {
        NodeEntity createdNode = nodesManager.createAndRead(newNode, user.getUser());
        newNode.setIsAlive(nodesManager.checkIsNodeAlive(newNode));
        // may be in task queue there are some tasks that can be calculated on new node
        //todo restore
        //taskRunningJob.checkTaskQueueManually();
        // to initialize solvers list
        return nodesManager.findOne(createdNode.getId(), user.getUser());
    }

    @RequestMapping(value = "/{id}/currentTask", method = RequestMethod.GET)
    public ResponseEntity<TaskShort> getCurrentTask(@PathVariable("id") int nodeId, @CurrentUser PortalUser user) {
        NodeEntity node = getManager().findOne(nodeId, user.getUser());
        if (node == null) {
            throw new NotFoundException(nodeId, NodeEntity.class);
        }
        TaskEntity runningTask = taskManager.findRunningOnNode(node.getId());
        return runningTask == null ? new ResponseEntity<>(org.springframework.http.HttpStatus.NO_CONTENT)
                : ResponseEntity.ok(new TaskShort(runningTask));
    }

    @RequestMapping(value = "/extended/{id}", method = RequestMethod.GET)
    public ExtendedNode getExtendedNode(@PathVariable("id") Integer nodeId,
                                        @CurrentUser PortalUser user) throws NotFoundException {
        NodeEntity node = getManager().findOne(nodeId, user.getUser());
        if (node == null) {
            throw new NotFoundException(nodeId, NodeEntity.class);
        }
        ExtendedNode extendedNode = Mapper.map(node, ExtendedNode.class);
        return nodesManager.extendNode(extendedNode);
    }


    @RequestMapping(value = "/extended", method = RequestMethod.GET)
    public List<ExtendedNode> getAllExtendedNodes(@CurrentUser PortalUser user) {
        List<NodeEntity> allNodes = getManager().findAll(user.getUser());
        allNodes.sort(Comparator.comparing(NodeEntity::getName));
        List<ExtendedNode> extendedNodes = Mapper.map(allNodes, new TypeToken<List<ExtendedNode>>() {
        });
        return nodesManager.extendNodes(extendedNodes);
    }

    @Override
    protected BaseManagerSuggest<NodeEntity> getManager() {
        return nodesManager;
    }

    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }
}
