package ru.crp.lvms.web.controllers;

import org.apache.commons.lang3.StringUtils;
import org.modelmapper.TypeToken;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.entities.Solver;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.URLs;

import javax.annotation.Nullable;
import javax.validation.Valid;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Set;

import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toSet;

@RestController
@RequestMapping(value = URLs.SOLVERS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class SolverController {
    private static class SolverTypeVO {
        public String id;
        public String text;

        public SolverTypeVO(String id, String text) {
            this.id = id;
            this.text = text;
        }
    }

    private static final Logger LOGGER = LoggerFactory.getLogger(SolverController.class);

    @Autowired
    private SolverConfigManager solverConfigManager;


    @RequestMapping(method = RequestMethod.GET)
    public List<Solver> getAll() {
        return Mapper.map(solverConfigManager.getAllConfigs(), new TypeToken<List<Solver>>() {
        });
    }


    // is used by select2 and it expects array of object with properties 'id' and 'text'
    @RequestMapping(value = "/{name}/solvingTypes", method = RequestMethod.GET)
    public Set<SolverTypeVO> getSolvingTypes(@PathVariable("name") String solverName) {
        SolverConfig solverConfig = solverConfigManager.getSolverByName(solverName);
        checkSolverPresent(solverConfig, solverName);
        Solver solverDto = Mapper.map(solverConfig, Solver.class);
        return solverDto.getTaskTypes().stream()
                .map(taskType -> new SolverTypeVO(taskType.toString(), taskType.toString()))
                .collect(toSet());
    }

    private void checkSolverPresent(@Nullable SolverConfig solverConfig, String solverName) {
        if (solverConfig == null) {
            throw new NotFoundException("No solver with name '" + solverName + "' found");
        }
    }


    @Transactional(readOnly = true)
    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<Solver> suggest(@RequestBody @Valid SuggestRequest suggestRequest) {
        List<Solver> solversToFilter = Mapper.map(solverConfigManager.getAllConfigs(), new TypeToken<List<Solver>>() {
        });
        for (SuggestRequest.Entry suggestEntry : suggestRequest.getQueries()) {
            solversToFilter = solversToFilter
                    .stream()
                    .filter(solver -> !OptimizationCao.ModeFrontier.getSolverConfigName().equals(solver.getName()) &&
                            !OptimizationCao.Tosca.getSolverConfigName().equals(solver.getName()))
                    .filter(solver -> {
                        try {
                            String property = org.apache.commons.beanutils.BeanUtils.getProperty(solver, suggestEntry.getColumn());
                            String query = suggestEntry.getValue();
                            if (query.startsWith("%") && query.endsWith("%")) {
                                return StringUtils.containsIgnoreCase(property, query.substring(1, query.length() - 1));
                            } else if (query.startsWith("%")) {
                                return StringUtils.endsWithIgnoreCase(property, query.substring(1));
                            } else {
                                return StringUtils.startsWithIgnoreCase(property, query.substring(0, query.length() - 1));
                            }
                        } catch (IllegalAccessException | InvocationTargetException | NoSuchMethodException ex) {
                            LOGGER.error("Error while suggesting solvers", ex);
                            return false;
                        }
                    })
                    .collect(toList());
        }
        return solversToFilter;
    }
}

