package ru.crp.lvms.web.controllers;


import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.TreeManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.OptimizationsTreeNode;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.misc.ResponseMessage;

import javax.annotation.Nullable;
import javax.validation.Valid;
import java.util.ArrayList;
import java.util.List;

import static java.util.Arrays.asList;
import static ru.crp.lvms.utils.MessageUtils.TREE_NODE_NOT_FOUND;
import static ru.crp.lvms.utils.URLs.TREE_URL;
import static ru.crp.lvms.web.controllers.misc.CrudController.deleteSeveralEntities;

@RestController
@RequestMapping(value = TREE_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class TreeController {
    @Autowired
    private TreeManager treeManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private ConstructionElementManager constructionElementManager;


    @RequestMapping(value = "/treeTypes", method = RequestMethod.GET)
    public List<TreeType> getTreeTypes() {
        return asList(TreeType.values());
    }


    @RequestMapping(value = "/itemTypes", method = RequestMethod.GET)
    public List<ItemType> getItemTypes() {
        return asList(ItemType.values());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/optimizations", method = RequestMethod.POST)
    public TreeNode createOptimizationTreeNode(@Valid @RequestBody OptimizationsTreeNode newNode,
                                               @CurrentUser PortalUser currentUser) {
        return treeManager.createOptimizationTreeNode(newNode, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/optimizations", method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public TreeNode createOptimizationFolder(@RequestParam("pid") Integer pid,
                                             @RequestParam("name") String name,
                                             @RequestParam("optimizationTypeId") Integer optimizationTypeId,
                                             @RequestParam("constructionElementId") Integer constructionElementId,
                                             @RequestParam(value = "constructionElementId", required = false) Integer baseOptimizationId,
                                             @Nullable @RequestParam(value = "file", required = false) MultipartFile newFile,
                                             @CurrentUser PortalUser currentUser) {
        OptimizationsTreeNode newFolder = new OptimizationsTreeNode();
        newFolder.setPid(pid);
        newFolder.setName(name);
        newFolder.setOptimizationTypeId(optimizationTypeId);
        newFolder.setConstructionElementId(constructionElementId);
        newFolder.setBaseOptimizationId(baseOptimizationId);
        if (newFile != null) {
            constructionElementManager.createNewFile(constructionElementId, newFile, FileType.DesignSolutionFile, currentUser.getUser());
        }
        return treeManager.createOptimizationTreeNode(newFolder, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{treeType}", method = RequestMethod.POST)
    public TreeNode createFolder(@PathVariable("treeType") TreeType treeType,
                                 @Valid @RequestBody TreeNode newFolder,
                                 @CurrentUser PortalUser currentUser) {
        return treeManager.createFolder(treeType, newFolder, currentUser.getUser());
    }


    @RequestMapping(value = "/{treeType}/root", method = RequestMethod.GET)
    public List<? extends TreeNode> getRootNodes(@PathVariable("treeType") TreeType treeType,
                                                 @RequestParam(name = "onlyFolders", defaultValue = "false") Boolean onlyFolders,
                                                 @CurrentUser PortalUser currentUser) {
        return filterOnlyFolders(onlyFolders, treeManager.getRootNodes(treeType, currentUser.getUser()));
    }


    private List<? extends TreeNode> filterOnlyFolders(Boolean onlyFolders, List<? extends TreeNode> nodes) {
        if (!onlyFolders) {
            return nodes;
        }
        List<TreeNode> filtered = new ArrayList<>();
        for (TreeNode node : nodes) {
            if (!node.isFolder()) {
                continue;
            }
            if (node.getChildrenTypes() != null &&
                    !asList(node.getChildrenTypes()).contains(ItemType.Folder)) {
                node.setParent(false);
            }
            filtered.add(node);
        }
        return filtered;
        // todo fixme
//        return nodes.stream()
//                .filter(TreeNode::isFolder)
//                .peek(treeNode -> {
//                    if (treeNode.getChildrenTypes() != null &&
//                            !asList(treeNode.getChildrenTypes()).contains(ItemType.Folder)) {
//                        treeNode.setParent(false);
//                    }
//                })
//                .collect(Collectors.toList());
    }


    @RequestMapping(value = "/{treeType}/subnodes", method = RequestMethod.GET)
    public List<? extends TreeNode> getNodes(@RequestParam(value = "id", required = false) Integer rootId,
                                             @PathVariable("treeType") TreeType treeType,
                                             @RequestParam(name = "onlyFolders", defaultValue = "false") Boolean onlyFolders,
                                             @CurrentUser PortalUser currentUser) {
        return filterOnlyFolders(onlyFolders, treeManager.getSubNodes(rootId, treeType, currentUser.getUser()));
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{treeType}", method = RequestMethod.PUT)
    public TreeNode updateNode(@PathVariable("treeType") TreeType treeType,
                               @Valid @RequestBody TreeNode nodeToUpdate,
                               @CurrentUser PortalUser currentUser) {
        return treeManager.updateNodeName(treeType, nodeToUpdate, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{treeType}/{id}", method = RequestMethod.PUT)
    public TreeNode updateNodeById(@PathVariable("id") int nodeId,
                                   @PathVariable("treeType") TreeType treeType,
                                   @Valid @RequestBody TreeNode nodeToUpdate,
                                   @CurrentUser PortalUser currentUser) {
        nodeToUpdate.setId(nodeId);
        return treeManager.updateNodeName(treeType, nodeToUpdate, currentUser.getUser());
    }


    @RequestMapping(value = "/{treeType}/{id}", method = RequestMethod.GET)
    public TreeNode getNode(@PathVariable("id") int nodeId,
                            @PathVariable("treeType") TreeType treeType,
                            @CurrentUser PortalUser currentUser) {
        TreeNode treeNode = treeManager.findOne(treeType, nodeId, currentUser.getUser());
        if (treeNode == null) {
            throw new NotFoundException(msgSrc.getMessage(TREE_NODE_NOT_FOUND, new Object[]{nodeId}, null, null));
        }
        return treeNode;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{treeType}/{id}", method = RequestMethod.DELETE)
    public ResponseMessage deleteNode(@PathVariable("treeType") TreeType treeType,
                                      @PathVariable("id") int id,
                                      @CurrentUser PortalUser user) {
        TreeNode deletedEntity = treeManager.deleteNode(treeType, id, user.getUser());
        TreeNode parent = deletedEntity.getPid() == null ? null : treeManager.findOne(treeType, deletedEntity.getPid(), user.getUser());
        return ResponseMessage.createSuccess("Успешно удалено", parent);

    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{treeType}", method = RequestMethod.DELETE)
    public void deleteSeveralNodes(@PathVariable("treeType") TreeType treeType,
                                   @RequestParam("ids") List<Integer> nodesIds,
                                   @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(TreeNode.class, nodesIds, (id) -> {
            deleteNode(treeType, id, currentUser);
        });
    }
}
