package ru.crp.lvms.web.controllers.effective.characteristics;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.ImageNotFoundException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.exceptions.ReportGenerationException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.pdf.reports.ECReportInput;
import ru.crp.lvms.pdf.reports.ReportGenerator;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.web.controllers.misc.ControllerImageNew;
import ru.crp.lvms.web.controllers.misc.CrudController;
import ru.crp.lvms.web.controllers.misc.FileHelper;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static ru.crp.lvms.utils.URLs.EFFECTIVE_CHARACTERISTICS_URL;

@RestController
@RequestMapping(value = EFFECTIVE_CHARACTERISTICS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class ECController extends CrudController<ECEntity, ECEntity, ECEntity> implements ControllerImageNew<ECEntity> {

    @Autowired
    private ECManager ECManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ReportGenerator reportGenerator;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private TaskManager taskManager;

    @Value("${testEnviroment}")
    private boolean testEnvironment;


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST)
    public ECEntity createEC(@Valid @RequestBody ECEntity newEC, @CurrentUser PortalUser currentUser) {
        newEC.setOwnerId(currentUser.getUser().getId());
        newEC.setStatusId(SolvingStatus.Created.getId());
        return getManager().createAndRead(newEC, currentUser.getUser());
    }

    @NotNull
    @Override
    public ECEntity get(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return getManager().extend(super.get(request, id, currentUser), currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    public ECEntity update(HttpServletRequest request, @PathVariable("id") int id, @Valid @RequestBody ECEntity updateVO,
                           @CurrentUser PortalUser user) {
        ECEntity existed = getManager().findOne(id, user.getUser());
        checkEntityExists(id, existed);

        updateVO.setCompositeMaterialId(existed.getCompositeMaterialId());
        updateVO.setOwnerId(existed.getOwnerId());
        updateVO.setCreationDate(existed.getCreationDate());
        updateVO.setId(existed.getId());

        Mapper.map(updateVO, existed);
        return getManager().updateAndRead(existed, user.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public ECEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @Nonnull
    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<ECEntity> suggest(@RequestBody @Valid SuggestRequest suggestRequest,
                                  @CurrentUser PortalUser currentUser) {
        return getManager().suggest(suggestRequest, currentUser.getUser());
    }

    @Nonnull
    @Override
    public List<ECEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        List<ECEntity> response = getManager().findAll(currentUser.getUser());
        return getManager().extend(response, currentUser.getUser());
    }

    @Nonnull
    @Override
    public PageingResponse<ECEntity> page(HttpServletRequest request,
                                          @RequestBody DataRequest dataRequest,
                                          @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<ECEntity> page = getManager()
                .find(dataRequest.getFilters(), pageable, currentUser.getUser());
        getManager().extend(page.getContent(), currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }

    @Nonnull
    @RequestMapping(value = "/{ecId}/defects", method = RequestMethod.GET)
    public List<ECDefectEntity> getAllDefects(@PathVariable("ecId") Integer ecId, @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        return getManager().getECDefects(ecId);
    }

    @Nonnull
    @RequestMapping(value = "/{ecId}/defects/list", method = RequestMethod.POST)
    public PageingResponse<ECDefectEntity> getDefectsPage(@PathVariable("ecId") Integer ecId,
                                                          @RequestBody DataRequest dataRequest,
                                                          @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        Pageable pageable = dataRequest.getSpringPageable();
        if (pageable.getSort() == null) {
            pageable = new PageRequest(pageable.getPageNumber(), pageable.getPageSize(), new Sort("defectNameId"));
        }
        dataRequest.getFilters().put("ecId", ecId);
        PageImpl<ECDefectEntity> response = getManager().getECDefectsPage(pageable, dataRequest.getFilters());
        dataRequest.getFilters().remove("ecId");
        return new PageingResponse<ECDefectEntity>(response.getContent(), dataRequest.getFilters(), pageable, response.getTotalElements());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/defects", method = RequestMethod.POST)
    public ECDefectEntity createDefect(@PathVariable("ecId") Integer ecId, @Valid @RequestBody ECDefectEntity newDefectDTO,
                                       @CurrentUser PortalUser currentUser) {
        throw new BadRequestException("Менять список дефектов запрещено");
//        checkEntityExists(ecId, currentUser);
//        newDefectDTO.setEcId(ecId);
//        return getManager().createECDefect(newDefectDTO);
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/defects/{defectId}", method = RequestMethod.PUT)
    public ECDefectEntity updateDefect(@PathVariable("ecId") Integer ecId,
                                       @PathVariable("defectId") Integer defectId,
                                       @RequestBody ECDefectEntity updateDTO,
                                       @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        ECDefectEntity existedDefect = checkDefectExistsAndBelongsToEC(ecId, defectId, currentUser);
        return getManager().updateECDefect(existedDefect, updateDTO);
    }

    private ECDefectEntity checkDefectExistsAndBelongsToEC(Integer ecId, Integer defectId, PortalUser currentUser) {
        ECDefectEntity defect = getManager().getECDefect(ecId, defectId, currentUser);
        if (defect == null) {
            throw new NotFoundException(defectId, ECDefectEntity.class);
        }
        return defect;
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/defects/{defectId}", method = RequestMethod.DELETE)
    public ECDefectEntity deleteDefect(@PathVariable("ecId") Integer ecId,
                                       @PathVariable("defectId") Integer defectId,
                                       @CurrentUser PortalUser currentUser) {
        throw new BadRequestException("Менять список дефектов запрещено");
//        checkEntityExists(ecId, currentUser);
//        ECDefectEntity existedDefect = checkDefectExistsAndBelongsToEC(ecId, defectId, currentUser);
//        getManager().deleteECDefect(existedDefect);
//        return existedDefect;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/defects", method = RequestMethod.DELETE)
    public void deleteSeveralDefects(@PathVariable("ecId") Integer ecId,
                                     @RequestParam("ids") List<Integer> defectIds,
                                     @CurrentUser PortalUser currentUser) throws BadRequestException {
        throw new BadRequestException("Менять список дефектов запрещено");
//        deleteSeveralEntities(ECDefectEntity.class, defectIds, (id) -> {
//            deleteDefect(ecId, id, currentUser);
//        });
    }

    @Nonnull
    @RequestMapping(value = "/{ecId}/microstructureParameters", method = RequestMethod.GET)
    public List<ECParamEntity> getAllMicrostructureParameters(@PathVariable("ecId") Integer ecId, @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        return getManager()
                .getECParameters(ecId, ECParamEntity.Type.Microstructure);
    }

    @Nonnull
    @RequestMapping(value = "/{ecId}/mesostructureParameters", method = RequestMethod.GET)
    public List<ECParamEntity> getAllMesostructureParameters(@PathVariable("ecId") Integer ecId, @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        return getManager()
                .getECParameters(ecId, ECParamEntity.Type.Mesostructure);
    }


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/parameters", method = RequestMethod.POST)
    public ECParamEntity createParameter(@PathVariable("ecId") Integer ecId, @Valid @RequestBody ECParamEntity newParamDTO,
                                         @CurrentUser PortalUser currentUser) {
        throw new BadRequestException("Менять список параметров запрещено");
//        checkEntityExists(ecId, currentUser);
//        newParamDTO.setEcId(ecId);
//        return getManager().createECParam(newParamDTO);
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/parameters/{paramId}", method = RequestMethod.PUT)
    public ECParamEntity updateParameter(@PathVariable("ecId") Integer ecId,
                                         @PathVariable("paramId") Integer paramId,
                                         @Valid @RequestBody ECParamEntity updateDTO,
                                         @CurrentUser PortalUser currentUser) {
        checkEntityExists(ecId, currentUser);
        ECParamEntity existedParam = checkParamExistsAndBelongsToEC(ecId, paramId, currentUser);
        return getManager().updateECParam(existedParam, updateDTO);
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/parameters/{paramId}", method = RequestMethod.DELETE)
    public ECParamEntity deleteParameter(@PathVariable("ecId") Integer ecId, @PathVariable("paramId") Integer paramId,
                                         @CurrentUser PortalUser currentUser) {
        throw new BadRequestException("Менять список параметров запрещено");
//        checkEntityExists(ecId, currentUser);
//        ECParamEntity existedParam = checkParamExistsAndBelongsToEC(ecId, paramId, currentUser);
//        getManager().deleteECParam(existedParam);
//        return existedParam;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{ecId}/parameters", method = RequestMethod.DELETE)
    public void deleteSeveralParameters(@PathVariable("ecId") Integer ecId,
                                        @RequestParam("ids") List<Integer> paramIds,
                                        @CurrentUser PortalUser currentUser) throws BadRequestException {
        throw new BadRequestException("Менять список параметров запрещено");
//        deleteSeveralEntities(ECParamEntity.class, paramIds, (id) -> {
//            deleteParameter(ecId, id, currentUser);
//        });
    }


    @RequestMapping(value = "/{ecId}/report", method = RequestMethod.GET)
    public void generateReport(@PathVariable("ecId") Integer ecId, @CurrentUser PortalUser currentUser,
                               HttpServletResponse response) throws IOException {
        ECEntity ecEntity = getManager().findOne(ecId, currentUser.getUser());
        checkEntityExists(ecId, ecEntity);
        if (ecEntity.getStatus() != SolvingStatus.Completed) {
            throw new ReportGenerationException("Эффективная характеристика не была рассчитана и нет результатов для отчета");
        }
        ECReportInput reportInfo = getManager().formEcReportInput(ecEntity);
        Path report = reportGenerator.generateReport(reportInfo);
        FileHelper.writeFileToResponse(report, ecEntity.getName() + "-report.pdf", response);
        Files.delete(report);
    }


    @RequestMapping(value = "/{id}/microstructureImage", method = RequestMethod.GET)
    public void getMicroImage(HttpServletResponse response,
                              @PathVariable("id") Integer entityId,
                              @CurrentUser PortalUser currentUser) throws IOException {
        try {
            Path microPic = resultsManager.getResultPictureByName(entityId,
                    TaskParentType.effectiveCharacteristic, ResultsManager.MICRO_PICTURE_NAME);
            FileHelper.writeFileToResponse(microPic, microPic.getFileName().toString(), response);
        } catch (ReportGenerationException ex) {
            if (testEnvironment) {
                throw new ImageNotFoundException();
            } else {
                throw ex;
            }
        }
    }


    @RequestMapping(value = "/{id}/mezostructureImage", method = RequestMethod.GET)
    public void getMezoImage(HttpServletResponse response,
                             @PathVariable("id") Integer entityId,
                             @CurrentUser PortalUser currentUser) throws IOException {
        try {
            Path mezoPic = resultsManager.getResultPictureByName(entityId,
                    TaskParentType.effectiveCharacteristic, ResultsManager.MEZO_PICTURE_NAME);
            FileHelper.writeFileToResponse(mezoPic, mezoPic.getFileName().toString(), response);
        } catch (ReportGenerationException ex) {
            if (testEnvironment) {
                throw new ImageNotFoundException();
            } else {
                throw ex;
            }
        }
    }

    @PreAuthorize("hasRole('ROLE_GUEST')")
    @RequestMapping(value = "/{id}/refreshResults", method = RequestMethod.GET)
    public void refreshResults(@PathVariable("id") int id,
                               @CurrentUser PortalUser currentUser) {
        ResultsParent resultsParent = Mapper.map(getManager().findOne(id, currentUser.getUser()), ResultsParent.class);
        taskManager.refreshResultsHandling(resultsParent, currentUser.getUser());
    }


    private ECParamEntity checkParamExistsAndBelongsToEC(Integer ecId, Integer paramId, PortalUser currentUser) {
        ECParamEntity defect = getManager().getECParameter(ecId, paramId, currentUser);
        if (defect == null) {
            throw new NotFoundException(paramId, ECParamEntity.class);
        }
        return defect;
    }


    @Override
    public ImageHandler imageHandler() {
        return getManager();
    }

    @Override
    public ECEntity getEntity(HttpServletRequest request, Integer entityId, PortalUser currentUser) {
        return get(request, entityId, currentUser);
    }

    @Override
    public FileManager getFileManager() {
        return fileManager;
    }

    @Override
    protected ECManager getManager() {
        return ECManager;
    }

    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }
}
