package ru.crp.lvms.web.controllers.materials;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.CompositeMaterialManager;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.MaterialManager;
import ru.crp.lvms.managers.misc.PreferencesManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.ColumnsController;
import ru.crp.lvms.web.controllers.misc.FileHelper;
import ru.crp.lvms.web.dto.pagination.DataRequest;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import java.io.IOException;
import java.util.List;

import static ru.crp.lvms.utils.URLs.*;
import static ru.crp.lvms.web.controllers.misc.CrudController.deleteSeveralEntities;

@RestController
@RequestMapping(value = {MATERIAL_FILES_URL, COMPOSITE_MATERIAL_FILES_URL}, produces = MediaType.APPLICATION_JSON_VALUE)
public class MaterialFileController implements FileHelper {
    @RestController
    @RequestMapping(value = MATERIAL_FILES_BASE_URL, produces = MediaType.APPLICATION_JSON_VALUE)
    public static class MaterialFileControllerColumns implements ColumnsController {
        @Autowired
        private PreferencesManager preferencesManager;

        @Override
        public Class getOutVoClass() {
            return FileEntity.class;
        }

        @Override
        public PreferencesManager getPreferencesManager() {
            return preferencesManager;
        }
    }

    @Autowired
    private FileManager fileManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private MaterialManager materialManager;


    @Nonnull
    @RequestMapping(method = RequestMethod.GET)
    public List<FileEntity> getMaterialFiles(HttpServletRequest request, @PathVariable("materialId") int materialId,
                                             @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        return fileManager.getMaterialFiles(isCompositeMaterial, materialId, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/list", method = RequestMethod.POST)
    public PageImpl<FileEntity> getMaterialFilesPage(HttpServletRequest request,
                                                     @PathVariable("materialId") int materialId,
                                                     @RequestBody DataRequest dataRequest,
                                                     @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        Pageable pageable = dataRequest.getSpringPageable();
        return fileManager.getMaterialFilesPage(isCompositeMaterial, materialId, dataRequest.getFilters(),
                pageable, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<FileEntity> suggestMaterialFiles(HttpServletRequest request,
                                                 @PathVariable("materialId") int materialId,
                                                 @RequestBody @Valid SuggestRequest suggestRequest,
                                                 @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        return fileManager
                .suggestMaterialFiles(isCompositeMaterial, materialId, suggestRequest, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{materialFileId}", method = RequestMethod.GET)
    public FileEntity getMaterialFile(HttpServletRequest request,
                                      @PathVariable("materialId") int materialId,
                                      @PathVariable("materialFileId") int materialFileId,
                                      @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        return fileManager.getMaterialFile(isCompositeMaterial, materialId, materialFileId, currentUser.getUser());
    }


    @RequestMapping(value = "/{materialFileId}/download", method = RequestMethod.GET)
    public void downloadMaterialFile(HttpServletRequest request, HttpServletResponse response,
                                     @PathVariable("materialId") int materialId,
                                     @PathVariable("materialFileId") int materialFileId,
                                     @CurrentUser PortalUser currentUser) throws IOException {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        FileEntity materialFile = fileManager
                .getMaterialFile(isCompositeMaterial, materialId, materialFileId, currentUser.getUser());
        writeFileToResponse(materialFile, response, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public FileEntity createMaterialFile(HttpServletRequest request,
                                         @PathVariable("materialId") int materialId,
                                         @RequestParam("file") MultipartFile newMaterialFile,
                                         @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        return fileManager.createMaterialFile(isCompositeMaterial, materialId, newMaterialFile, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{materialFileId}", method = RequestMethod.PUT)
    public FileEntity updateMaterialFile(HttpServletRequest request,
                                         @PathVariable("materialId") int materialId,
                                         @PathVariable("materialFileId") int materiaFilelId,
                                         @Valid @RequestBody FileEntity updatedMaterialFile,
                                         @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        updatedMaterialFile.setId(materiaFilelId);
        updatedMaterialFile.setFileTypeId(FileType.MaterialFile.getId());
        return fileManager.updateMaterialFile(isCompositeMaterial, materialId, updatedMaterialFile, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{materialFileId}", method = RequestMethod.DELETE)
    public FileEntity deleteMaterialFile(HttpServletRequest request,
                                         @PathVariable("materialId") int materialId,
                                         @PathVariable("materialFileId") int materialFileId,
                                         @CurrentUser PortalUser currentUser) {
        boolean isCompositeMaterial = checkIsCompositeMaterial(request);
        if (isCompositeMaterial) {
            return compositeMaterialManager.deleteMaterialFile(materialId, materialFileId, currentUser.getUser());
        } else {
            return materialManager.deleteMaterialFile(materialId, materialFileId, currentUser.getUser());
        }
    }

    @RequestMapping(value = "", method = RequestMethod.DELETE)
    public void deleteSeveralMaterialFiles(HttpServletRequest request,
                                           @PathVariable("materialId") Integer materialId,
                                           @RequestParam("ids") List<Integer> materialFileIds,
                                           @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, materialFileIds, (id) -> {
            deleteMaterialFile(request, materialId, id, currentUser);
        });
    }

    @Override
    public FileManager getFileManager() {
        return fileManager;
    }
}
