package ru.crp.lvms.web.controllers.technology;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.HandlerMapping;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.TechnologyParametersManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.misc.PreferencesManager;
import ru.crp.lvms.model.entities.technology.TechnologyParametersEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.ColumnsController;
import ru.crp.lvms.web.controllers.misc.CrudControllerColumns;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.List;
import java.util.Map;

import static ru.crp.lvms.utils.URLs.TECHNOLOGY_PARAMETERS_URL;
import static ru.crp.lvms.utils.URLs.TECHNOLOGY_URL;

@RestController
@RequestMapping(value = TECHNOLOGY_PARAMETERS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class TechnologyParametersController
        extends CrudControllerColumns<TechnologyParametersEntity, TechnologyParametersEntity, TechnologyParametersEntity> {
    @RestController
    @RequestMapping(value = TECHNOLOGY_URL + "/parameters", produces = MediaType.APPLICATION_JSON_VALUE)
    public static class ParametersColumnsController implements ColumnsController {
        @Autowired
        private PreferencesManager preferencesManager;

        @Override
        public Class getOutVoClass() {
            return TechnologyParametersEntity.class;
        }

        @Override
        public PreferencesManager getPreferencesManager() {
            return preferencesManager;
        }
    }

    @Autowired
    private TechnologyParametersManager technologyParametersManager;

    @Autowired
    private EventManager eventManager;


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST)
    public TechnologyParametersEntity createParameter(@PathVariable("technologyId") Integer technologyId,
                                                      @Valid @RequestBody TechnologyParametersEntity newParameter,
                                                      @CurrentUser PortalUser currentUser) {
        newParameter.setTechnologyId(technologyId);
        return technologyParametersManager.createAndRead(newParameter, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    public TechnologyParametersEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                             @Valid @RequestBody TechnologyParametersEntity updateVO,
                                             @CurrentUser PortalUser user) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        updateVO.setTechnologyId(technologyId);
        return super.update(request, id, updateVO, user);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public TechnologyParametersEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<TechnologyParametersEntity> suggest(@PathVariable("technologyId") Integer technologyId,
                                                    @RequestBody @Valid SuggestRequest suggestRequest,
                                                    @CurrentUser PortalUser currentUser) {
        suggestRequest.upsertQuery("technologyId", technologyId.toString());
        return technologyParametersManager.suggest(suggestRequest, currentUser.getUser());
    }


    @NotNull
    @Override
    public List<TechnologyParametersEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        return technologyParametersManager.getTechnologyParameters(technologyId, currentUser.getUser());
    }


    @NotNull
    @Override
    public PageingResponse<TechnologyParametersEntity> page(HttpServletRequest request,
                                                            @RequestBody DataRequest dataRequest,
                                                            @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<TechnologyParametersEntity> page = technologyParametersManager
                .getTechnologyParametersPage(technologyId, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @Override
    protected BaseManagerSuggest<TechnologyParametersEntity> getManager() {
        return technologyParametersManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }
}
