package ru.crp.lvms.web.controllers.users;

import edu.vt.middleware.password.*;
import org.apache.commons.lang3.StringEscapeUtils;
import org.hibernate.validator.constraints.NotEmpty;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.UserManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.misc.MailSender;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.Utils;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.dto.user.UserCreate;
import ru.crp.lvms.web.dto.user.UserIn;
import ru.crp.lvms.web.dto.user.UserOut;

import javax.mail.MessagingException;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import static ru.crp.lvms.utils.AuthUtils.ownerOrAdmin;
import static ru.crp.lvms.utils.URLs.USERS_URL;


@RestController
@RequestMapping(value = USERS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class UserController extends CrudControllerSuggest<UserEntity, UserIn, UserOut> {
    private static final Logger LOGGER = LoggerFactory.getLogger(UserController.class);

    private List<Rule> passwordRules;
    private CharacterCharacteristicsRule charRule;

    {
        // password must be between 8 and 16 chars long
        LengthRule lengthRule = new LengthRule(8, 16);

        // don't allow whitespace
        WhitespaceRule whitespaceRule = new WhitespaceRule();

        // control allowed characters
        charRule = new CharacterCharacteristicsRule();
        // require at least 1 digit in passwords
        charRule.getRules().add(new DigitCharacterRule(1));
        // require at least 1 non-alphanumeric char
        charRule.getRules().add(new NonAlphanumericCharacterRule(1));
        // require at least 1 upper case char
        charRule.getRules().add(new UppercaseCharacterRule(1));
        // require at least 1 lower case char
        charRule.getRules().add(new LowercaseCharacterRule(1));
        // require at least 3 of the previous rules be met
        charRule.setNumberOfCharacteristics(3);

        // don't allow alphabetical sequences
        AlphabeticalSequenceRule alphaSeqRule = new AlphabeticalSequenceRule();

        // don't allow qwerty sequences
        QwertySequenceRule qwertySeqRule = new QwertySequenceRule();

        // don't allow 4 repeat characters
        RepeatCharacterRegexRule repeatRule = new RepeatCharacterRegexRule(4);

        // group all rules together in a List
        passwordRules = new ArrayList<Rule>();
        passwordRules.add(lengthRule);
        passwordRules.add(whitespaceRule);
        passwordRules.add(charRule);
        passwordRules.add(alphaSeqRule);
        passwordRules.add(qwertySeqRule);
        passwordRules.add(repeatRule);
    }

    @Autowired
    private UserManager userManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private MailSender mailSender;


    @Override
    protected BaseManagerSuggest<UserEntity> getManager() {
        return userManager;
    }

    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }

    @PreAuthorize("hasRole('ROLE_ADMIN')")
    @RequestMapping(method = RequestMethod.POST)
    public UserOut create(@Valid @RequestBody UserCreate userCreateVO, @CurrentUser PortalUser currentUser) {
        UserEntity userEntity = Mapper.map(userCreateVO, UserEntity.class);
//        PasswordGenerator generator = new PasswordGenerator();
//        String password = generator.generatePassword(8, charRule.getRules());
//        String msg = " greetings from CML-space.";
//        msg += "\nYour login is \n" + userEntity.getLogin();
//        msg += "\nYour password is \n" + password;
//        msg += "\nYou can change your password later.";
//        userEntity.setPassword(Utils.createMD5Hash(password));
        String errorMessages = checkNewPassword(userCreateVO.getPassword());
        if (errorMessages != null) {
            throw new BadRequestException(errorMessages);
        }
        userEntity.setPassword(Utils.createMD5Hash(userCreateVO.getPassword()));
        userEntity.setCreationDate(new Date());
        UserEntity user = userManager.createAndRead(userEntity, currentUser.getUser());
//        mailSender.sendNotificationMailAsync(msg, userEntity.getEmail(), userEntity.getName());
        return Mapper.map(user, UserOut.class);
    }


    public static class OldNewPassword {
        @NotEmpty(message = "Старый пароль должен быть задан")
        public String oldPassword;
        @NotEmpty(message = "Новый пароль должен быть задан")
        public String newPassword;
    }


    /**
     * @param id             user id
     * @param oldNewPassword password must be between 8 and 16 chars long,
     *                       don't allow whitespace,
     *                       [require at least 1 digit in passwords,
     *                       require at least 1 non-alphanumeric char,
     *                       require at least 1 upper case char,
     *                       require at least 1 lower case char ---
     *                       require at least 3 of the previous rules be met,
     *                       don't allow alphabetical sequences,
     *                       don't allow qwerty sequences,
     *                       don't allow 4 repeat characters
     * @return
     * @throws BadRequestException if password does not match rules
     */
    @RequestMapping(value = "/{id}/password", method = RequestMethod.PUT)
    public UserOut changePassword(@PathVariable("id") int id, @Valid @RequestBody OldNewPassword oldNewPassword,
                                  @CurrentUser PortalUser portalUser) throws BadRequestException {
        ownerOrAdmin(id, portalUser.getUser());

        UserEntity updateEntity = userManager.findOne(id, portalUser.getUser());
        if (!updateEntity.getPassword().equals(Utils.createMD5Hash(oldNewPassword.oldPassword)))
            throw new BadRequestException("Старый пароль введен неверно");
        String errorMessages = checkNewPassword(oldNewPassword.newPassword);
        if (errorMessages != null) {
            throw new BadRequestException(errorMessages);
        }

        updateEntity.setPassword(Utils.createMD5Hash(oldNewPassword.newPassword));
        updateEntity = userManager.updateAndRead(updateEntity, portalUser.getUser());
        return Mapper.map(updateEntity, UserOut.class);
    }


    @RequestMapping(value = "/{id}/password-reset", method = RequestMethod.POST)
    public UserOut resetPassword(@PathVariable("id") int id, @CurrentUser PortalUser portalUser)
            throws BadRequestException {
        ownerOrAdmin(id, portalUser.getUser());

        UserEntity updateEntity = userManager.findOne(id, portalUser.getUser());

        PasswordGenerator generator = new PasswordGenerator();
        String password = generator.generatePassword(8, charRule.getRules());
        String msg = "Ваш пароль был сброшен. Новый пароль - " + '\n' + password;
        try {
            mailSender.sendNotificationMail(msg, updateEntity.getEmail(), updateEntity.getName());
            updateEntity.setPassword(Utils.createMD5Hash(password));
            updateEntity = userManager.updateAndRead(updateEntity, portalUser.getUser());
        } catch (MessagingException e) {
            String errMsg = "Не удалось отправить сообщение о сбросе пароля пользователю " + updateEntity.getName() + "("
                    + updateEntity.getEmail() + "). Номер пользователя - " + updateEntity.getId();
            LOGGER.error(errMsg, e);
            throw new InternalException(errMsg, e);
        }
        return Mapper.map(updateEntity, UserOut.class);
    }


    @PreAuthorize("hasRole('ROLE_ADMIN')")
    @RequestMapping(value = "/{id}/block", method = RequestMethod.GET)
    public UserOut setActive(@PathVariable("id") int id, @CurrentUser PortalUser portalUser) throws BadRequestException {
        if (portalUser.getUser().getId() == id)
            throw new BadRequestException("Невозможно сделать неактивным себя");

        UserEntity updateEntity = userManager.findOne(id, portalUser.getUser());
        updateEntity.setActive(!updateEntity.isActive());
        updateEntity = userManager.updateAndRead(updateEntity, portalUser.getUser());
        return Mapper.map(updateEntity, UserOut.class);
    }


    private String checkNewPassword(String password) {
        PasswordValidator validator = new PasswordValidator(passwordRules);
        PasswordData passwordData = new PasswordData(new Password(password));

        RuleResult result = validator.validate(passwordData);
        if (!result.isValid()) {
            StringBuilder errorMessages = new StringBuilder("Недопустимый пароль:");
            //TODO rewrite
            boolean insufficient = false;
            for (RuleResultDetail res : result.getDetails()) {
                switch (res.getErrorCode()) {
                    case "TOO_SHORT":
                        errorMessages.append("Пароль слишком короткий. Используйте больше 8 символов.");
                        break;
                    case "INSUFFICIENT_CHARACTERS":
                        if (!insufficient) {
                            errorMessages.append("<br/>Пароль слишком простой. Используйте 1 или больше букв в верхнем регистре, "
                                    + "1 или больше цифру и 1 или больше специальных символов " + "("
                                    + StringEscapeUtils.escapeHtml4("`~!@#$%^&*()-_=+[{]}\\|;:'\"<,>./? )") + ")");
                            insufficient = true;
                        }
                        break;
                    default:
                        break;
                }
            }
            //            for (String msg : validator.getMessages(result)) {
            //                errorMessages.append(msg);
            //                errorMessages.append('\n');
            //            }
            return errorMessages.toString();
        }
        return null;
    }


    @PreAuthorize("hasRole('ROLE_ADMIN')")
    @NotNull
    @Override
    public UserOut delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser user) {
        return super.delete(request, id, user);
    }

    @PreAuthorize("hasRole('ROLE_ADMIN')")
    @Override
    public void deleteSeveral(HttpServletRequest request, @RequestParam("ids") List<Integer> ids, @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request, ids, currentUser);
    }


    @NotNull
    @Override
    public UserOut update(HttpServletRequest request, @PathVariable("id") int id, @Valid @RequestBody UserIn updateVO, @CurrentUser PortalUser user) {
        return super.update(request, id, updateVO, user);
    }
}
