DROP VIEW optistruct_optimization_constraint_join;
DROP VIEW optistruct_optimization_tf_join;

ALTER TABLE optistruct_optimization_tf_and_constraint DROP COLUMN attb_id;
ALTER TABLE optistruct_optimization_tf_and_constraint DROP COLUMN attb_value;

ALTER TABLE optistruct_optimization_tf_and_constraint DROP COLUMN ptype_id;
ALTER TABLE optistruct_optimization_tf_and_constraint DROP COLUMN ptype_value;

ALTER TABLE optistruct_optimization_tf_and_constraint ADD COLUMN loadstep_id INTEGER REFERENCES optistruct_optimization_load_steps ON DELETE SET NULL;

CREATE OR REPLACE FUNCTION check_optistruct_tf_parameter_allowed()
  RETURNS TRIGGER AS $body$
BEGIN
  IF NEW.atta_id IS NOT NULL AND NOT EXISTS(SELECT *
                                        FROM optistruct_tf_allowed_atta
                                        WHERE tf_id = NEW.tf_id AND atta_id = NEW.atta_id)
  THEN
    RAISE EXCEPTION 'That tf and atta combination is not allowed (tf: %, atta: %)', NEW.tf_id, NEW.atta_id;
  END IF;

  IF NEW.atti_id IS NOT NULL AND NOT EXISTS(SELECT *
                                        FROM optistruct_tf_allowed_atti
                                        WHERE tf_id = NEW.tf_id AND atti_id = NEW.atti_id)
  THEN
    RAISE EXCEPTION 'That tf and atti combination is not allowed (tf: %, atti: %)', NEW.tf_id, NEW.atti_id;
  END IF;

  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;


-- Optistruct tf
CREATE VIEW optistruct_optimization_tf_join AS (
  SELECT base.id,
    base.optimization_id,
    base.tf_id,

    base.atta_id,
    base.atta_value,

    base.atti_id,
    base.atti_value,

    base.loadstep_id,

    base.optimization_type_id,
    base.name,

    optistruct_tf.name as tf_name,
    optistruct_tf.response as tf_response,
    optistruct_tf_atta.name as atta_name,
    optistruct_tf_atti.name as atti_name,
    loadstep.name as loadstep_name,
    optistruct_tf_optimization_type.name as optimization_type_name

  FROM optistruct_optimization_tf_and_constraint base
    LEFT JOIN optistruct_tf ON base.tf_id = optistruct_tf.id
    LEFT JOIN optistruct_tf_atta ON base.atta_id = optistruct_tf_atta.id
    LEFT JOIN optistruct_tf_atti ON base.atti_id = optistruct_tf_atti.id
    LEFT JOIN optistruct_optimization_load_steps loadstep ON base.loadstep_id = loadstep.id
    LEFT JOIN optistruct_tf_optimization_type ON base.optimization_type_id = optistruct_tf_optimization_type.id

  WHERE base.is_constraint = FALSE
);

CREATE RULE optistruct_optimization_tf_join_CREATE
AS ON INSERT TO optistruct_optimization_tf_join DO INSTEAD (
  INSERT INTO optistruct_optimization_tf_and_constraint (
    optimization_id,
    tf_id,
    atta_id,
    atta_value,
    atti_id,
    atti_value,
    loadstep_id,
    optimization_type_id,
    is_constraint,
    name
  )
  VALUES (
    NEW.optimization_id,
    NEW.tf_id,
    NEW.atta_id,
    NEW.atta_value,
    NEW.atti_id,
    NEW.atti_value,
    NEW.loadstep_id,
    NEW.optimization_type_id,
    FALSE,
    NEW.name
  )
  RETURNING
    optistruct_optimization_tf_and_constraint.id,
    optistruct_optimization_tf_and_constraint.optimization_id,
    optistruct_optimization_tf_and_constraint.tf_id,
    optistruct_optimization_tf_and_constraint.atta_id,
    optistruct_optimization_tf_and_constraint.atta_value,
    optistruct_optimization_tf_and_constraint.atti_id,
    optistruct_optimization_tf_and_constraint.atti_value,
    optistruct_optimization_tf_and_constraint.loadstep_id,
    optistruct_optimization_tf_and_constraint.optimization_type_id,
    NULL :: VARCHAR(255) AS tf_response,
    NULL :: VARCHAR(255) AS tf_name,
    NULL :: VARCHAR(255) AS name,
    NULL :: VARCHAR(255) AS atta_name,
    NULL :: VARCHAR(255) AS atti_name,
    NULL :: VARCHAR(255) AS loadstep_name,
    NULL :: VARCHAR(255) AS optimization_type_name
);

CREATE RULE optistruct_optimization_tf_join_UPDATE
AS ON UPDATE TO optistruct_optimization_tf_join DO INSTEAD (
  UPDATE optistruct_optimization_tf_and_constraint
  SET
    optimization_id = NEW.optimization_id,
    tf_id = NEW.tf_id,
    atta_id = NEW.atta_id,
    atta_value = NEW.atta_value,
    atti_id = NEW.atti_id,
    loadstep_id = NEW.loadstep_id,
    atti_value = NEW.atti_value,
    optimization_type_id = NEW.optimization_type_id,
    name = NEW.name
  WHERE id = NEW.id);

CREATE RULE optistruct_optimization_tf_join_DELETE
AS ON DELETE TO optistruct_optimization_tf_join DO INSTEAD (
  DELETE FROM optistruct_optimization_tf_and_constraint WHERE id = OLD.id
);

-- Optistruct constraints
CREATE VIEW optistruct_optimization_constraint_join AS (
  SELECT
    base.id,
    base.optimization_id,
    base.tf_id,
    base.atta_id,
    base.atta_value,
    base.atti_id,
    base.atti_value,
    base.loadstep_id,
    base.lower_bound,
    base.upper_bound,
    base.name,
    o.optimization_type_id   AS optimization_type_id,
    optistruct_tf.name       AS tf_name,
    optistruct_tf.response   AS tf_response,
    optistruct_tf_atta.name  AS atta_name,
    optistruct_tf_atti.name  AS atti_name,
    loadstep.name AS loadstep_name

  FROM optistruct_optimization_tf_and_constraint base
    JOIN optimizations o ON base.optimization_id = o.id
    LEFT JOIN optistruct_tf ON base.tf_id = optistruct_tf.id
    LEFT JOIN optistruct_tf_atta ON base.atta_id = optistruct_tf_atta.id
    LEFT JOIN optistruct_optimization_load_steps loadstep ON base.loadstep_id = loadstep.id
    LEFT JOIN optistruct_tf_atti ON base.atti_id = optistruct_tf_atti.id

  WHERE base.is_constraint = TRUE
);

CREATE RULE optistruct_optimization_constraint_join_CREATE
AS ON INSERT TO optistruct_optimization_constraint_join DO INSTEAD (
  INSERT INTO optistruct_optimization_tf_and_constraint (
    optimization_id,
    tf_id,
    atta_id,
    atta_value,
    atti_id,
    atti_value,
    loadstep_id,
    is_constraint,
    lower_bound,
    upper_bound,
    name)
  VALUES(
    NEW.optimization_id,
    NEW.tf_id,
    NEW.atta_id,
    NEW.atta_value,
    NEW.atti_id,
    NEW.atti_value,
    NEW.loadstep_id,
    TRUE,
    NEW.lower_bound,
    NEW.upper_bound,
    NEW.name)
  RETURNING
    optistruct_optimization_tf_and_constraint.id,
    optistruct_optimization_tf_and_constraint.optimization_id,
    optistruct_optimization_tf_and_constraint.tf_id,
    optistruct_optimization_tf_and_constraint.atta_id,
    optistruct_optimization_tf_and_constraint.atta_value,
    optistruct_optimization_tf_and_constraint.atti_id,
    optistruct_optimization_tf_and_constraint.atti_value,
    optistruct_optimization_tf_and_constraint.loadstep_id,
    optistruct_optimization_tf_and_constraint.lower_bound,
    optistruct_optimization_tf_and_constraint.upper_bound,
    NULL :: VARCHAR(255) AS name,
    NULL :: INTEGER AS optimization_type_id,
    NULL :: VARCHAR(255) AS tf_response,
    NULL :: VARCHAR(255) AS tf_name,
    NULL :: VARCHAR(255) AS atta_name,
    NULL :: VARCHAR(255) AS atti_name,
    NULL :: VARCHAR(255) AS loadstep_name
);

CREATE RULE optistruct_optimization_constraint_join_UPDATE
AS ON UPDATE TO optistruct_optimization_constraint_join DO INSTEAD (
  UPDATE optistruct_optimization_tf_and_constraint
  SET
    optimization_id = NEW.optimization_id,
    tf_id           = NEW.tf_id,
    atta_id         = NEW.atta_id,
    atta_value      = NEW.atta_value,
    atti_id         = NEW.atti_id,
    atti_value      = NEW.atti_value,
    loadstep_id     = NEW.loadstep_id,
    lower_bound     = NEW.lower_bound,
    upper_bound     = NEW.upper_bound,
    name            = NEW.name
  WHERE id = NEW.id);

CREATE RULE optistruct_optimization_constraint_join_DELETE
AS ON DELETE TO optistruct_optimization_constraint_join DO INSTEAD (
  DELETE FROM optistruct_optimization_tf_and_constraint
  WHERE id = OLD.id
);