CREATE TABLE standards (
  id               INTEGER DEFAULT nextval('tree_seq' :: REGCLASS),
  name             VARCHAR UNIQUE                             NOT NULL,
  description      VARCHAR,
  creator          VARCHAR                                    NOT NULL,
  owner_id         INTEGER                                    NOT NULL,
  standard_file_id INTEGER,
  load_file_id     INTEGER,
  date_created     TIMESTAMP WITHOUT TIME ZONE DEFAULT now()  NOT NULL,
  date_modified    TIMESTAMP WITHOUT TIME ZONE DEFAULT now()  NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (owner_id) REFERENCES users (id),
  FOREIGN KEY (standard_file_id) REFERENCES files (id),
  FOREIGN KEY (load_file_id) REFERENCES files (id)
);

CREATE TRIGGER standards_check_files
BEFORE DELETE ON standards
FOR EACH ROW
EXECUTE PROCEDURE check_files_parent_constraint();

CREATE TRIGGER standards_update
BEFORE UPDATE ON standards
FOR EACH ROW
WHEN (OLD.* IS DISTINCT FROM NEW.*)
EXECUTE PROCEDURE update_last_modified();

CREATE VIEW standards_join AS
  SELECT
    s.*,
    f1.name AS standard_file_name,
    f2.name AS load_file_name,
    u.name  AS owner
  FROM standards s
    JOIN users u ON s.owner_id = u.id
    LEFT JOIN files f1 ON s.standard_file_id = f1.id
    LEFT JOIN files f2 ON s.load_file_id = f2.id;

CREATE OR REPLACE RULE standards_create AS ON INSERT TO standards_join DO INSTEAD (
  INSERT INTO standards (name, owner_id, standard_file_id, load_file_id, description, creator) VALUES
    (NEW.name, NEW.owner_id, NEW.standard_file_id, NEW.load_file_id, NEW.description, NEW.creator)
  RETURNING
    standards.*, '' :: VARCHAR AS standard_file_name, '' :: VARCHAR AS load_file_name, '' :: VARCHAR AS owner);

CREATE OR REPLACE RULE standards_update AS ON UPDATE TO standards_join DO INSTEAD (
  UPDATE standards
  SET name       = NEW.name, owner_id = NEW.owner_id, standard_file_id = NEW.standard_file_id,
    load_file_id = NEW.load_file_id, description = NEW.description, creator = NEW.creator
  WHERE id = NEW.id);

CREATE OR REPLACE RULE standards_delete AS ON DELETE TO standards_join DO INSTEAD (
  DELETE FROM standards
  WHERE id = OLD.id
);


CREATE TABLE virtual_simulations (
  id                    SERIAL,
  composite_material_id INTEGER               NOT NULL,
  sample_file_id        INTEGER               NOT NULL,
  standard_id           INTEGER               NOT NULL,
  is_finished           BOOLEAN DEFAULT FALSE NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (composite_material_id) REFERENCES composite_materials (id),
  FOREIGN KEY (sample_file_id) REFERENCES files (id),
  FOREIGN KEY (standard_id) REFERENCES standards (id),
  CONSTRAINT single_sample_file_standard_constrain UNIQUE (sample_file_id, standard_id)
);


CREATE TRIGGER virtual_simulations_check_material_sample_file_consistency_create
BEFORE INSERT ON virtual_simulations
FOR EACH ROW
EXECUTE PROCEDURE check_material_sample_file_consistency();


CREATE TRIGGER virtual_simulations_check_material_sample_file_consistency_update
BEFORE INSERT ON virtual_simulations
FOR EACH ROW
EXECUTE PROCEDURE check_material_sample_file_consistency();


CREATE VIEW virtual_simulations_join AS (
  SELECT
    vs.*,
    cm.name AS composite_material_name,
    f.name  AS sample_file_name,
    s.name  AS standard_name
  FROM virtual_simulations vs
    INNER JOIN composite_materials cm ON vs.composite_material_id = cm.id
    INNER JOIN files f ON vs.sample_file_id = f.id
    INNER JOIN standards s ON vs.standard_id = s.id
);


CREATE TABLE simulation_standards (
  id          SERIAL,
  standard_id INTEGER NOT NULL,
  material_id INTEGER NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (standard_id) REFERENCES standards (id),
  FOREIGN KEY (material_id) REFERENCES composite_materials (id)
);

CREATE TABLE simulation_sample_files (
  id          SERIAL,
  file_id     INTEGER NOT NULL,
  material_id INTEGER NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (file_id) REFERENCES files (id),
  FOREIGN KEY (material_id) REFERENCES composite_materials (id)
);
