CREATE OR REPLACE FUNCTION update_last_modified()
  RETURNS TRIGGER AS $body$
BEGIN
  NEW.date_modified := now();
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;


-- Sets appropriate value to path column. Should be called when creating new tree node.
CREATE OR REPLACE FUNCTION create_path()
  RETURNS TRIGGER AS $body$
DECLARE
  parent     RECORD;
  path_entry VARCHAR;
BEGIN
  path_entry = create_path_entry(NEW.name, NEW.id, TG_ARGV [0] :: tree_node_type);
  IF NEW.pid IS NULL
  THEN
    NEW.path=ARRAY [path_entry];
    NEW.path_names=NEW.name;
  ELSE
    SELECT *
    INTO parent
    FROM tree
    WHERE id = NEW.pid;
    IF parent.path IS NOT NULL
    THEN
      NEW.path = parent.path || path_entry;
      NEW.path_names = parent.path_names || '/' || NEW.name;
    ELSE
      NEW.path=ARRAY [path_entry];
      NEW.path_names=NEW.name;
    END IF;
  END IF;
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION create_path_for_optimization_folder()
  RETURNS TRIGGER AS $body$
DECLARE
  parent     RECORD;
  path_entry VARCHAR;
BEGIN
  path_entry = create_path_entry(NEW.name, NEW.id, 'OptimizationFolder' :: tree_node_type);
  IF NEW.parent_folder_id IS NULL
  THEN
    SELECT *
    INTO parent
    FROM tree
    WHERE id = NEW.construction_element_id;
  ELSE
    SELECT *
    INTO parent
    FROM tree
    WHERE id = NEW.parent_folder_id;
  END IF;
  NEW.path = parent.path || path_entry;
  NEW.path_names = parent.path_names || '/' || NEW.name;
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION create_path_entry(node_name VARCHAR, id INTEGER, item_type tree_node_type)
  RETURNS VARCHAR AS $body$
DECLARE
  path_entry VARCHAR;
  name       VARCHAR;
BEGIN
  path_entry = concat('{', '"name":"', node_name, '", "id":', id, ', "type":"', item_type, '"}');
  RETURN path_entry;
END;
$body$ LANGUAGE plpgsql;


-- updates node path and path of all its children should be called while changing node name or pid
CREATE OR REPLACE FUNCTION update_path()
  RETURNS TRIGGER AS $body$
DECLARE
  parent     RECORD;
  path_entry VARCHAR;
  res        INTEGER;
  item_type  tree_node_type;
BEGIN
  path_entry = create_path_entry(NEW.name, NEW.id, TG_ARGV [0] :: tree_node_type);
  IF NEW.pid IS NULL
  THEN
    NEW.path = ARRAY [path_entry];
    NEW.path_names = NEW.name;
  ELSE
    SELECT *
    INTO parent
    FROM tree
    WHERE tree.id = NEW.pid;
    NEW.path = parent.path || path_entry;
    NEW.path_names = parent.path_names || '/' || NEW.name;
  END IF;
  res = change_children_path(NEW.id, NEW.path, NEW.path_names);
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION update_path_for_optimization_folder()
  RETURNS TRIGGER AS $body$
DECLARE
  parent     RECORD;
  path_entry VARCHAR;
  res        INTEGER;
  item_type  tree_node_type;
BEGIN
  path_entry = create_path_entry(NEW.name, NEW.id, 'OptimizationFolder' :: tree_node_type);
  IF NEW.parent_folder_id IS NULL
  THEN
    SELECT *
    INTO parent
    FROM tree
    WHERE tree.id = NEW.construction_element_id;
  ELSE
    SELECT *
    INTO parent
    FROM tree
    WHERE tree.id = NEW.parent_folder_id;
  END IF;
  NEW.path = parent.path || path_entry;
  NEW.path_names = parent.path_names || '/' || NEW.name;
  res = change_children_path(NEW.id, NEW.path, NEW.path_names);
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION change_children_path(parent_id INTEGER, parent_path VARCHAR [], parent_path_names VARCHAR)
  RETURNS INTEGER AS $body$
DECLARE
  child          RECORD;
  new_path       VARCHAR [];
  new_path_names VARCHAR;
  path_entry     VARCHAR;
  res            INTEGER;
BEGIN
  FOR child IN SELECT *
               FROM tree
               WHERE tree.pid = parent_id
  LOOP
    path_entry = create_path_entry(child.name, child.id, child.item_type);
    new_path = parent_path || path_entry;
    new_path_names = parent_path_names || '/' || child.name;
    UPDATE tree
    SET path = new_path, path_names = new_path_names
    WHERE tree.id = child.id;
    res = change_children_path(child.id, new_path, new_path_names);
  END LOOP;
  RETURN res;
END;
$body$ LANGUAGE plpgsql;


-- Procedure created to be able RAISE EXCEPTION --
CREATE OR REPLACE FUNCTION update_material_if_not_locked()
  RETURNS TRIGGER AS $body$
BEGIN
  IF OLD.lock = TRUE
  THEN
    RAISE EXCEPTION 'Material locked from update';
  ELSE
    NEW.date_modified := now();
    RETURN NEW;
  END IF;
END;
$body$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION check_files_parent_constraint()
  RETURNS TRIGGER AS $body$
BEGIN
  IF exists(SELECT id
            FROM files
            WHERE pid = OLD.id)
  THEN
    RAISE EXCEPTION 'There are dependent files for object';
  ELSE
    RETURN OLD;
  END IF;
END;
$body$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION check_parameters_parent_constraint()
  RETURNS TRIGGER AS $body$
BEGIN
  IF exists(SELECT id
            FROM parameters_materials
            WHERE parameters_materials.material_id = OLD.id)
  THEN
    RAISE EXCEPTION 'There are dependent files for object';
  ELSE
    RETURN OLD;
  END IF;
END;
$body$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION check_material_sample_file_consistency()
  RETURNS TRIGGER AS $body$
DECLARE
  sample_file RECORD;
BEGIN
  SELECT *
  INTO sample_file
  FROM files
  WHERE id = NEW.sample_file_id;
  IF sample_file.pid <> NEW.composite_material_id
  THEN
    RAISE EXCEPTION 'Sample file must belongs to composite material';
  END IF;
  IF sample_file.file_type_id <> 1
  THEN
    RAISE EXCEPTION 'Only sample files can be used in virtual simulations';
  END IF;
  RETURN NEW;
END;
$body$ LANGUAGE plpgsql;
