-- effective characteristics
CREATE TABLE effective_characteristics (
  id            SERIAL,
  composite_material_id INTEGER NOT NULL,
  image_file_id INTEGER,
  microstructure_file_id INTEGER,
  name          VARCHAR(255) NOT NULL,
  description   TEXT,
  owner_id      INTEGER                                           NOT NULL,
  date_created  TIMESTAMP WITHOUT TIME ZONE DEFAULT now()         NOT NULL,
  date_modified TIMESTAMP WITHOUT TIME ZONE DEFAULT now()         NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (owner_id) REFERENCES users (id),
  FOREIGN KEY (composite_material_id) REFERENCES composite_materials(id),
  FOREIGN KEY (image_file_id) REFERENCES files(id),
  FOREIGN KEY (microstructure_file_id) REFERENCES files(id) ON DELETE SET NULL,
  UNIQUE (name)
);

CREATE TRIGGER effective_characteristics_update
  BEFORE UPDATE ON effective_characteristics
    FOR EACH ROW
      WHEN (OLD.* IS DISTINCT FROM NEW.*)
        EXECUTE PROCEDURE update_last_modified();

CREATE VIEW effective_characteristics_join AS (
  SELECT ec.*,
  u.name AS owner,
  cm.name AS composite_material_name,
  mf.name AS microstructure_file_name

  FROM effective_characteristics ec
  JOIN users u ON u.id = ec.owner_id
  JOIN composite_materials cm ON cm.id = ec.composite_material_id
  LEFT JOIN files mf ON mf.id = ec.microstructure_file_id
);

-- effective characteristic parameters
CREATE TABLE effective_characteristics_parameters (
  id SERIAL,
  name VARCHAR NOT NULL,
  type VARCHAR NOT NULL,
  parent_param_id INTEGER,
  is_composite BOOLEAN NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (parent_param_id) REFERENCES effective_characteristics_parameters (id) ON DELETE CASCADE,
  UNIQUE (name)
);

CREATE TABLE effective_characteristics_parameters_join (
  id SERIAL,
  effective_characteristic_id INTEGER NOT NULL,
  param_id INTEGER NOT NULL,
  param_value VARCHAR,

  PRIMARY KEY (id),
  FOREIGN KEY (effective_characteristic_id) REFERENCES effective_characteristics (id) ON DELETE CASCADE,
  UNIQUE (param_id, effective_characteristic_id)
);

CREATE VIEW effective_characteristics_parameters_view AS (
SELECT
    ecp.*,
    params.name AS param_name,
    params.type AS param_type,
    params.parent_param_id AS parent_param_id,
    parents.name AS parent_param_name

    FROM effective_characteristics_parameters_join ecp
    INNER JOIN effective_characteristics_parameters params ON ecp.param_id = params.id
    LEFT JOIN effective_characteristics_parameters parents ON params.parent_param_id = parents.id
);

CREATE OR REPLACE RULE ec_parameters_CREATE AS ON INSERT TO effective_characteristics_parameters_view DO INSTEAD (
  INSERT INTO effective_characteristics_parameters_join (effective_characteristic_id, param_id, param_value) VALUES
    (NEW.effective_characteristic_id, NEW.param_id, NEW.param_value)
  RETURNING
    effective_characteristics_parameters_join.*,
    NULL :: VARCHAR AS param_name,
    NULL :: VARCHAR AS param_type,
    NULL :: INTEGER AS parent_param_id,
    NULL :: VARCHAR AS parent_param_name);

CREATE OR REPLACE RULE ec_parameters_UPDATE AS ON UPDATE TO effective_characteristics_parameters_view DO INSTEAD (
  UPDATE effective_characteristics_parameters_join
  SET param_value = NEW.param_value
  WHERE param_id = NEW.param_id);

CREATE OR REPLACE RULE ec_parameters_DELETE AS ON DELETE TO effective_characteristics_parameters_view DO INSTEAD (
  DELETE FROM effective_characteristics_parameters_join
  WHERE param_id = OLD.param_id
);

-- defects parameters
CREATE TABLE effective_characteristics_defects_parameters (
  id SERIAL,
  effective_characteristic_id INTEGER NOT NULL,
  defect_name_id INTEGER NOT NULL,
  percentage DOUBLE PRECISION,
  thickness DOUBLE PRECISION,
  width DOUBLE PRECISION,
  length DOUBLE PRECISION,

  PRIMARY KEY (id),
  FOREIGN KEY (effective_characteristic_id) REFERENCES effective_characteristics (id) ON DELETE CASCADE,
  FOREIGN KEY (defect_name_id) REFERENCES possible_defects_names (id) ON DELETE CASCADE,
  UNIQUE (defect_name_id, effective_characteristic_id)
);

CREATE VIEW effective_characteristics_defects_parameters_join AS (
  SELECT ecdp.*,
  pdn.name as defect_name
  FROM effective_characteristics_defects_parameters ecdp
  JOIN possible_defects_names pdn ON pdn.id = ecdp.defect_name_id
);

CREATE OR REPLACE RULE ec_defects_create AS ON INSERT TO effective_characteristics_defects_parameters_join DO INSTEAD (
  INSERT INTO effective_characteristics_defects_parameters (effective_characteristic_id, defect_name_id, percentage, thickness, width, length) VALUES
    (NEW.effective_characteristic_id, NEW.defect_name_id, NEW.percentage, NEW.thickness, NEW.width, NEW.length)
  RETURNING
    effective_characteristics_defects_parameters.*,
    NULL :: VARCHAR AS defect_name);

CREATE OR REPLACE RULE ec_defects_update AS ON UPDATE TO effective_characteristics_defects_parameters_join DO INSTEAD (
  UPDATE effective_characteristics_defects_parameters
  SET percentage = NEW.percentage, thickness = NEW.thickness, width = NEW.width, length = NEW.length, defect_name_id = NEW.defect_name_id
  WHERE id = NEW.id);

CREATE OR REPLACE RULE ec_defects_delete AS ON DELETE TO effective_characteristics_defects_parameters_join DO INSTEAD (
  DELETE FROM effective_characteristics_defects_parameters
  WHERE id = OLD.id
);
