CREATE TABLE task_statuses (
  id SERIAL,
  display_name VARCHAR(255) NOT NULL,

  PRIMARY KEY(id)
);

INSERT INTO task_statuses (display_name) VALUES
('DRAFT'),
('WAITING'),
('SENDING'),
('STARTING'),
('SOLVING'),
('CALCULATION_FINISHED'),
('IMPORTING_RESULTS'),
('TERMINATING'),
('FINISHED'),
('FAILED'),
('CANCELED'),
('TERMINATED');


CREATE TABLE tasks (
  id SERIAL,

  parent_type VARCHAR(255) NOT NULL,
  ec_id INTEGER,
  sim_id INTEGER,
  optimization_id INTEGER,

  status_id INTEGER NOT NULL,
  queue_priority INTEGER NOT NULL,
  task_position SERIAL,
  solver_name VARCHAR(255) NOT NULL,
  additional_startup_arguments VARCHAR(255),
  node_id INTEGER,
  num_cores INTEGER,
  memory INTEGER,
  progress INTEGER,
  type VARCHAR NOT NULL,
  owner_id INTEGER NOT NULL,
  date_created TIMESTAMP WITHOUT TIME ZONE DEFAULT now() NOT NULL,
  date_modified TIMESTAMP WITHOUT TIME ZONE DEFAULT now() NOT NULL,
  start_exec_date TIMESTAMP WITHOUT TIME ZONE,


  PRIMARY KEY (id),
  FOREIGN KEY (status_id) REFERENCES task_statuses (id),
  FOREIGN KEY (owner_id) REFERENCES users (id),
  FOREIGN KEY (node_id) REFERENCES nodes(id),
  FOREIGN KEY (ec_id) REFERENCES effective_characteristics(id),
  FOREIGN KEY (sim_id) REFERENCES virtual_simulations(id),

  CONSTRAINT check_parent_type CHECK (parent_type = 'EffectiveCharacteristic' OR parent_type = 'VirtualSimulation' OR parent_type = 'Optimization'),
    CONSTRAINT check_parent_id CHECK
      ((parent_type = 'EffectiveCharacteristic' AND ec_id IS NOT NULL) OR (parent_type = 'VirtualSimulation' AND sim_id IS NOT NULL)
      OR (parent_type = 'Optimization' AND optimization_id IS NOT NULL))
);

CREATE TRIGGER tasks_update
  BEFORE UPDATE ON tasks
    FOR EACH ROW
      WHEN (OLD.* IS DISTINCT FROM NEW.*)
        EXECUTE PROCEDURE update_last_modified();

-- trigger for auto updating start solving date
CREATE OR REPLACE FUNCTION update_task_start_exec_date() RETURNS TRIGGER AS $body$
    BEGIN
    -- if solving status changed and new status is 'Starting'
      IF NEW.status_id <> OLD.status_id AND NEW.status_id = 2 THEN
          NEW.start_exec_date := now();
      END IF;
      RETURN NEW;
    END;
$body$ LANGUAGE plpgsql;

CREATE TRIGGER tasks_update_start_exec_date
  BEFORE UPDATE ON tasks
    FOR EACH ROW
      WHEN (OLD.* IS DISTINCT FROM NEW.*)
        EXECUTE PROCEDURE update_task_start_exec_date();


CREATE VIEW tasks_join AS (
  SELECT t.*,
  u.name AS owner,
  n.name AS node,

  CASE
    WHEN parent_type = 'EffectiveCharacteristic' THEN ec.id
    WHEN parent_type = 'VirtualSimulation' THEN vs.id
    WHEN parent_type = 'Optimization' THEN o.id
    ELSE NULL
   END
    AS parent_id,

  CASE
    WHEN parent_type = 'EffectiveCharacteristic' THEN ec.name
    WHEN parent_type = 'VirtualSimulation' THEN vs.name
    WHEN parent_type = 'Optimization' THEN o.name
    ELSE NULL
  END
   AS parent_name

  FROM tasks t
  INNER JOIN users u ON u.id = t.owner_id
  LEFT JOIN nodes n ON n.id = t.node_id
  LEFT JOIN effective_characteristics ec ON ec.id = t.ec_id
  LEFT JOIN virtual_simulations vs ON vs.id = t.sim_id
  LEFT JOIN optimizations o ON o.id = t.optimization_id
);

CREATE OR REPLACE RULE tasks_join_CREATE AS ON INSERT TO tasks_join DO INSTEAD (
    INSERT INTO tasks (type, status_id, queue_priority, solver_name, additional_startup_arguments, node_id,
    num_cores, memory, progress, owner_id, parent_type, ec_id, sim_id, optimization_id) VALUES
      (NEW.type, NEW.status_id, NEW.queue_priority, NEW.solver_name, NEW.additional_startup_arguments, NEW.node_id,
           NEW.num_cores, NEW.memory, NEW.progress, NEW.owner_id, NEW.parent_type,
           CASE WHEN NEW.parent_type = 'EffectiveCharacteristic' THEN NEW.parent_id ELSE NULL END,
           CASE WHEN NEW.parent_type = 'VirtualSimulation' THEN NEW.parent_id ELSE NULL END,
           CASE WHEN NEW.parent_type = 'Optimization' THEN NEW.parent_id ELSE NULL END)
    RETURNING
      tasks.*,
      NULL :: VARCHAR AS owner,
      NULL :: VARCHAR AS node,
      NULL :: INTEGER AS parent_id,
      NULL :: VARCHAR(255) AS parent_name
);

CREATE OR REPLACE RULE tasks_join_UPDATE AS ON UPDATE TO tasks_join DO INSTEAD (
  UPDATE tasks
  SET type = NEW.type,
      status_id = NEW.status_id,
      queue_priority = NEW.queue_priority,
      task_position = NEW.task_position,
      solver_name = NEW.solver_name,
      additional_startup_arguments = NEW.additional_startup_arguments,
      node_id = NEW.node_id,
      num_cores = NEW.num_cores,
      memory = NEW.memory,
      progress = NEW.progress,
      owner_id = NEW.owner_id
  WHERE id = NEW.id);

CREATE OR REPLACE RULE tasks_join_DELETE AS ON DELETE TO tasks_join DO INSTEAD (
  DELETE FROM tasks WHERE id = OLD.id
);
