package ru.crp.lvms;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.fileGenerators.AnsysConfigFileGenerator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collections;
import java.util.stream.Collectors;

import static java.util.Arrays.asList;
import static org.mockito.Mockito.when;

public class AnsysConfigFileGeneratorTest extends BaseIT {

    private final static String INP_DB_CONFIG =
            "resume,'name','db',,0,0\n" +
                    "/solu\n" +
                    "solve";

    private final static String INP_CDB_CONFIG =
            "/prep7\n" +
                    "cdread,db,'name','cdb',,,\n" +
                    "/solu\n" +
                    "solve";

    private OptimizationEntity optimization;

    @Mock
    private OptimizationManager optimizationManager;

    @Mock
    private FileManager fileManager;

    @InjectMocks
    private AnsysConfigFileGenerator AnsysConfigFileGenerator;

    @Before
    public void init() {
        optimization = projectCreator.createOptimization();
        MockitoAnnotations.initMocks(this);
    }

    @Test
    public void itMustFindInpFileInSimulationFiles() throws IOException {
        FileEntity inpFile = new FileEntity();
        inpFile.setName("testImpFile.inp");
        Path tmpPath = Files.createTempFile(null, null);
        when(optimizationManager.getOptimizationFiles(optimization, null)).thenReturn(Collections.singletonList(inpFile));
        when(fileManager.resolvePathToFile(inpFile, null)).thenReturn(tmpPath);

        Assert.assertEquals(tmpPath, AnsysConfigFileGenerator.generateInpFile(optimization).getPathToFile());
    }

    @Test
    public void itMustGenerateInpFileIfDBInSimulationFiles() throws IOException {
        FileEntity dbFile = new FileEntity();
        dbFile.setName("name.db");
        when(optimizationManager.getOptimizationFiles(optimization, null)).thenReturn(Collections.singletonList(dbFile));
        String generatedFileContent = Files.readAllLines(AnsysConfigFileGenerator.generateInpFile(optimization).getPathToFile())
                .stream().collect(Collectors.joining("\n"));
        Assert.assertEquals(INP_DB_CONFIG, generatedFileContent);
    }

    @Test
    public void itMustGenerateInpFileIfCDBInSimulationFiles() throws IOException {
        FileEntity dbFile = new FileEntity();
        dbFile.setName("name.cdb");
        when(optimizationManager.getOptimizationFiles(optimization, null)).thenReturn(Collections.singletonList(dbFile));
        String generatedFileContent = Files.readAllLines(AnsysConfigFileGenerator.generateInpFile(optimization).getPathToFile())
                .stream().collect(Collectors.joining("\n"));
        Assert.assertEquals(INP_CDB_CONFIG, generatedFileContent);
    }

    @Test
    public void itMustGenerateInpFileIfDBAndSFilesInSimulationFiles() throws IOException {
        FileEntity dbFile = new FileEntity();
        dbFile.setName("name.db");
        FileEntity sfile = new FileEntity();
        sfile.setFileTypeId(FileType.LoadCondition.getId());
        sfile.setName("sfile.s09");
        when(optimizationManager.getOptimizationFiles(optimization, null)).thenReturn(asList(dbFile, sfile));
        String generatedFileContent = Files.readAllLines(AnsysConfigFileGenerator.generateInpFile(optimization).getPathToFile())
                .stream().collect(Collectors.joining("\n"));
        Assert.assertEquals(INP_DB_CONFIG.replace("solve", "lssolve,1,9,1"), generatedFileContent);
    }

    @Test
    public void itMustGenerateInpFileIfCDBAndSFilesInSimulationFiles() throws IOException {
        FileEntity dbFile = new FileEntity();
        dbFile.setName("name.cdb");
        FileEntity sfile = new FileEntity();
        sfile.setFileTypeId(FileType.LoadCondition.getId());
        sfile.setName("sfile.s09");
        when(optimizationManager.getOptimizationFiles(optimization, null)).thenReturn(asList(dbFile, sfile));
        String generatedFileContent = Files.readAllLines(AnsysConfigFileGenerator.generateInpFile(optimization).getPathToFile())
                .stream().collect(Collectors.joining("\n"));
        Assert.assertEquals(INP_CDB_CONFIG.replace("solve", "lssolve,1,9,1"), generatedFileContent);
    }
}
