package ru.crp.lvms;

import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.managers.ReferenceGraphManager;
import ru.crp.lvms.model.entities.ReferenceGraphEntry;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Supplier;

import static org.junit.Assert.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static ru.crp.lvms.utils.TestUtils.GETobjects;
import static ru.crp.lvms.utils.TestUtils.toMap;

public class ReferenceGraphIT extends BaseIT {
    private static final String BASE_URL = "/referenceGraph";
    private static final String FORBIDDEN_ITEM_TYPE_ERROR_MESSAGE = "No reference graph for type ";

    @Autowired
    private ReferenceGraphManager referenceGraphManager;

    private Map<String, Supplier<Entity>> objectByType;

    @Before
    public void init() {
        objectByType = new HashMap<>();
        objectByType.put(ReferenceGraphManager.MATERIAL, projectCreator::createMaterial);
        objectByType.put(ReferenceGraphManager.COMPOSITE_MATERIAL, projectCreator::createCompositeMaterial);
        objectByType.put(ReferenceGraphManager.CONSTRUCTION_ELEMENT, projectCreator::createConstructionElementEntity);
        objectByType.put(ReferenceGraphManager.OPTIMIZATION, projectCreator::createOptimization);
    }

    @Test
    public void graphShouldBeAvailableForOnly4ItemTypes() {
        Set<String> allowedElements = referenceGraphManager.getAllReferenceTypes();

        final int testId = 111;
        allowedElements.forEach(allowedItemType -> {
            try {
                MvcResult result = mockMvc.perform(makeRequest(testId, allowedItemType + "NotAllowed"))
                        .andExpect(status().isBadRequest())
                        .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                        .andReturn();
                Map<String, String> response = toMap(result);
                assertEquals(FORBIDDEN_ITEM_TYPE_ERROR_MESSAGE
                        + allowedItemType + "NotAllowed", response.get("message"));
            } catch (Exception e) {
                fail();
            }
        });
    }


    @Test
    public void itShouldReturnOnlyEntranceIfItHasNoReference() throws Exception {
        checkAllTypesSatisfies((id, response) -> {
            try {
                ReferenceGraphEntry root = fromJson(response);
                assertEquals(id.intValue(), root.getId());
                assertNull(root.getChildren());
                assertNull(root.getReferenceId());
                assertTrue(root.isEntrance());
                assertTrue(root.isMainPath());
            } catch (IOException e) {
                fail();
            }
        });
    }


    private void checkAllTypesSatisfies(BiConsumer<Integer, JsonNode> responseChecker) {
        Set<String> allowedElements = referenceGraphManager.getAllReferenceTypes();

        allowedElements.forEach(itemType -> {
            Integer objectId = objectByType.get(itemType).get().getId();
            String url = makeUrl(objectId, itemType);
            JsonNode jsonNode = GETobjects(mockMvc, url);
            responseChecker.accept(objectId, jsonNode);
        });
    }


    private MockHttpServletRequestBuilder makeRequest(int id, String itemType) {
        return get(BASE_URL)
                .param("id", "" + id)
                .param("type", itemType);
    }

    private String makeUrl(int id, String itemType) {
        return BASE_URL + "?id=" + id + "&type=" + itemType;
    }

    private ReferenceGraphEntry fromJson(JsonNode json) throws IOException {
        return jacksonMapper.readValue(json.toString(), ReferenceGraphEntry.class);
    }
}
