package ru.crp.lvms;

import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.web.dto.CompositeMaterialIn;
import ru.crp.lvms.web.dto.MaterialIn;

import javax.annotation.Nullable;
import java.util.List;
import java.util.Set;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toList;
import static org.junit.Assert.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.TREE_URL;

public class TreeEndpointIT extends BaseIT {

    @Test
    public void testGetTreeTypes() throws Exception {
        JsonNode response = GETobjects(mockMvc, TREE_URL + "/treeTypes");
        List<TreeType> treeTypes = treeTypesFromJson(response);
        assertArrayEquals(TreeType.values(), treeTypes.toArray(new TreeType[treeTypes.size()]));
    }


    @Test
    public void testGetItemTypes() throws Exception {
        JsonNode response = GETobjects(mockMvc, TREE_URL + "/itemTypes");
        List<ItemType> itemTypes = itemTypesFromJson(response);
        assertArrayEquals(ItemType.values(), itemTypes.toArray(new ItemType[itemTypes.size()]));
    }


    @Test
    public void testGetRootNodes() throws Exception {
        // materials tree
        TreeNode newFolder = projectCreator.createFolder(TreeType.materials);
        JsonNode response = GETobjects(mockMvc, TREE_URL + "/" + TreeType.materials + "/root");
        List<TreeNode> fullTree = treeNodesFromJson(response);
        List<Integer> nodeIds = fullTree.stream().map(TreeNode::getId).collect(toList());
        assertFalse(fullTree.isEmpty());
        assertTrue(nodeIds.contains(newFolder.getId()));
        checkAllowedNodeTypes(TreeType.materials, fullTree);

        // technologies tree
        int testTechnologyId = projectCreator.createTechnologyEntity().getId();
        response = GETobjects(mockMvc, TREE_URL + "/" + TreeType.technologies + "/root");
        fullTree = treeNodesFromJson(response);
        nodeIds = fullTree.stream().map(TreeNode::getId).collect(toList());
        assertFalse(fullTree.isEmpty());
        assertTrue(nodeIds.contains(testTechnologyId));
        checkAllowedNodeTypes(TreeType.technologies, fullTree);
    }


    private void checkAllowedNodeTypes(TreeType treeType, List<TreeNode> nodes) {
        Set<ItemType> allowedItemTypes = treeType.getTreeItemTypes();
        nodes.forEach(node -> assertTrue("treeType " + treeType + " must not contain " + node.getItemType(),
                allowedItemTypes.contains(node.getItemType())));
    }


    @Test
    public void testGetNode() throws Exception {
        // materials tree
        TreeNode folder = projectCreator.createFolder(TreeType.materials);
        TreeNode foundFolder = GETobject(mockMvc, createObjectUrl(TREE_URL + "/materials", folder), TreeNode.class);
        assertEquals(folder.getId(), foundFolder.getId());

        MaterialEntity material = projectCreator.createMaterial();
        TreeNode foundMaterial = GETobject(mockMvc, createObjectUrl(TREE_URL + "/materials", material.getId()), TreeNode.class);
        assertEquals(material.getId(), foundMaterial.getId());

//        FileEntity materialFile = projectCreator.createMaterialFile(false);
//        TreeNode foundMaterialFile = GETobject(mockMvc, createObjectUrl(TREE_URL + "/materials", materialFile.getId()), TreeNode.class);
//        assertEquals(materialFile.getId(), foundMaterialFile.getId());

        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial();
        TreeNode foundCompositeMaterial = GETobject(mockMvc, createObjectUrl(TREE_URL + "/compositeMaterials", compositeMaterial.getId()),
                TreeNode.class);
        assertEquals(compositeMaterial.getId(), foundCompositeMaterial.getId());

        // technologies tree
        int testTechnologyId = projectCreator.createTechnologyEntity().getId();
        TreeNode foundTechnology = GETobject(mockMvc, createObjectUrl(TREE_URL + "/technologies", testTechnologyId), TreeNode.class);
        assertEquals(testTechnologyId, foundTechnology.getId());
    }

    @Test
    public void testDeleteNode() throws Exception {
        // materials tree
        TreeNode folder = projectCreator.createFolder(TreeType.materials);
        DELETEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", folder));
        GETAndExpectNotFound(mockMvc, createObjectUrl(TREE_URL + "/materials", folder));

        MaterialEntity material = projectCreator.createMaterial();
        DELETEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", material.getId()));
        GETAndExpectNotFound(mockMvc, createObjectUrl(TREE_URL + "/materials", material.getId()));

//        FileEntity materialFile = projectCreator.createMaterialFile(false);
//        DELETEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", materialFile.getId()));
//        GETAndExpectNotFound(mockMvc, createObjectUrl(TREE_URL + "/materials", materialFile.getId()));

        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial();
        DELETEobject(mockMvc, createObjectUrl(TREE_URL + "/compositeMaterials", compositeMaterial.getId()));
        GETAndExpectNotFound(mockMvc, createObjectUrl(TREE_URL + "/compositeMaterials", compositeMaterial.getId()));

        // technologies tree
        int testTechnologyId = projectCreator.createTechnologyEntity().getId();
        DELETEobject(mockMvc, createObjectUrl(TREE_URL + "/technologies", testTechnologyId));
        GETAndExpectNotFound(mockMvc, createObjectUrl(TREE_URL + "/technologies", testTechnologyId));
    }

    @Test
    public void testUpdateNode() throws Exception {
        TreeNode folderIn = projectCreator.newFolder();
        TreeNode createdFolder = projectCreator.createFolder(TreeType.materials, folderIn);
        folderIn.setName(unique("updated name"));
        folderIn.setId(createdFolder.getId());
        TreeNode updatedFolder = UPDATEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", createdFolder), folderIn, TreeNode.class);
        assertEquals(folderIn.getName(), updatedFolder.getName());

        MaterialIn materialIn = projectCreator.newMaterial();
        MaterialEntity material = projectCreator.createMaterial(materialIn);
        materialIn.setName(unique("updated name"));
        materialIn.setId(material.getId());
        TreeNode updatedMaterial = UPDATEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", material.getId()),
                toTreeNode(materialIn), TreeNode.class);
        assertEquals(materialIn.getName(), updatedMaterial.getName());

        CompositeMaterialIn compositeMaterialIn = projectCreator.newCompositeMaterial();
        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial(compositeMaterialIn);
        compositeMaterialIn.setName(unique("updated name"));
        compositeMaterial.setId(compositeMaterial.getId());
        TreeNode updatedCompositeMaterial = UPDATEobject(mockMvc, createObjectUrl(TREE_URL + "/compositeMaterials", compositeMaterial.getId()),
                toTreeNode(compositeMaterialIn), TreeNode.class);
        assertEquals(compositeMaterialIn.getName(), updatedCompositeMaterial.getName());

//        FileEntity materialFileIn = projectCreator.newMaterialFile(false);
//        FileEntity materialFile = projectCreator.createMaterialFile(false, materialFileIn);
//        materialFileIn.setName(unique("updated name"));
//        materialFileIn.setId(materialFile.getId());
//        TreeNode updatedMaterialFile = UPDATEobject(mockMvc, createObjectUrl(TREE_URL + "/materials", materialFile.getId()),
//                toTreeNode(materialFileIn), TreeNode.class);
//        assertEquals(materialFileIn.getName(), updatedMaterialFile.getName());
    }


    @Test
    public void testGetSubnodes() throws Exception {
        TreeNode root = projectCreator.createFolder(TreeType.materials);
        TreeNode nodeIn = projectCreator.newFolder();
        nodeIn.setPid(root.getId());
        TreeNode parent = projectCreator.createFolder(TreeType.materials, nodeIn);
        nodeIn = projectCreator.newFolder();
        nodeIn.setPid(parent.getId());
        TreeNode child = projectCreator.createFolder(TreeType.materials, nodeIn);

        List<TreeNode> allNodes = makeSubnodesRequest(TreeType.materials, null);
        checkAllowedNodeTypes(TreeType.materials, allNodes);
        List<Integer> allNodesIds = allNodes.stream().map(TreeNode::getId).collect(toList());
        List<Integer> expectedNodes = asList(root.getId(), parent.getId(), child.getId());
        expectedNodes.forEach(nodeId -> assertTrue(allNodesIds.contains(nodeId)));

        List<TreeNode> rootSubnodes = makeSubnodesRequest(TreeType.materials, Integer.toString(root.getId()));
        List<Integer> rootNodesIds = rootSubnodes.stream().map(TreeNode::getId).collect(toList());
        assertTrue(rootNodesIds.contains(parent.getId()));
        assertTrue(rootNodesIds.contains(child.getId()));
        assertFalse(rootNodesIds.contains(root.getId()));

        List<TreeNode> parentSubnodes = makeSubnodesRequest(TreeType.materials, Integer.toString(parent.getId()));
        List<Integer> parentSubnodesIds = parentSubnodes.stream().map(TreeNode::getId).collect(toList());
        assertTrue(parentSubnodesIds.contains(child.getId()));
        assertFalse(parentSubnodesIds.contains(parent.getId()));
        assertFalse(parentSubnodesIds.contains(root.getId()));

        List<TreeNode> childSubnodes = makeSubnodesRequest(TreeType.materials, Integer.toString(child.getId()));
        List<Integer> childNodesIds = childSubnodes.stream().map(TreeNode::getId).collect(toList());
        assertTrue(childNodesIds.isEmpty());
    }

    private List<TreeNode> makeSubnodesRequest(TreeType treeType, @Nullable String rootId) throws Exception {
        MvcResult result = mockMvc.perform(get(TREE_URL + "/" + treeType + "/subnodes")
                .param("id", rootId))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        JsonNode response = jacksonMapper.readTree(result.getResponse().getContentAsString());
        return treeNodesFromJson(response);
    }
}
