package ru.crp.lvms.optimizations.tosca.config;

import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.tosca.config.argumentConverters.MidPlaneArgumentConverter;
import ru.crp.lvms.optimizations.tosca.config.argumentConverters.SymmetryConditionTypeArgumentConverter;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationTechnologicalConstraint;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNode;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationTechnologicalConstraintManager;
import ru.crp.lvms.optimizations.tosca.managers.ToscaTechnologicalConstraintTreeValueManager;

import java.util.Collections;
import java.util.List;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.Collections.singletonList;
import static java.util.stream.Collectors.toList;
import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.*;
import static org.mockito.Mockito.when;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.CheckDofArgumentConverter.CHECK_DOF_ARGUMENT_IDS;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.MasterArgumentConverter.MASTER_ARGUMENT_ID;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.MidPlaneArgumentConverter.MID_PLANE_ARGUMENT_ID;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.MultipleChildrenArgumentConverter.CLIENT_DIR_ARGUMENT_ID;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.SymmetryConditionTypeArgumentConverter.CS_CHILD_ARGUMENT_ID;
import static ru.crp.lvms.optimizations.tosca.config.argumentConverters.SymmetryConditionTypeArgumentConverter.TYPE_TC_PARAM_ID;
import static ru.crp.lvms.optimizations.tosca.config.commandConverters.CommandConverter.*;

public class ToscaTechnologicalConstraintsGeneratorTest {

    private ToscaTechnologicalConstraintsGenerator generator;

    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    private ToscaOptimizationTechnologicalConstraintManager toscaOptimizationTechnologicalConstraintManager;

    private ToscaTechnologicalConstraintTreeValueManager toscaTechnologicalConstraintTreeValueManager;

    @Before
    public void init() {
        toscaOptimizationSettingsManager = Mockito.mock(ToscaOptimizationSettingsManager.class);
        toscaOptimizationTechnologicalConstraintManager = Mockito.mock(ToscaOptimizationTechnologicalConstraintManager.class);
        toscaTechnologicalConstraintTreeValueManager = Mockito.mock(ToscaTechnologicalConstraintTreeValueManager.class);
        generator = new ToscaTechnologicalConstraintsGenerator(toscaOptimizationSettingsManager, toscaOptimizationTechnologicalConstraintManager, toscaTechnologicalConstraintTreeValueManager);
        generator.init();
    }


    @Test
    public void onlyDefaultArgumentsShouldBeUsedIfThereAreNoValuesForTC() throws Exception {
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(emptyList());

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(-1);
        List<Command> commandsList = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).collect(toList());
        assertEquals(1, commandsList.size());
        assertEquals(3, commandsList.get(0).getParameters().size());
    }


    @Test
    public void shapeTCShouldBeConvertedInto2Commands() {
        checkTCConvertedInto2Commands(SHAPE_TC_ID, OptimizationTypeParameters.Shape);
    }


    @Test
    public void symmetryConditionTCShouldBeConvertedInto2Commands() {
        checkTCConvertedInto2Commands(SYMMETRY_CONDITION_TC_ID, OptimizationTypeParameters.Topo);
    }


    private void checkTCConvertedInto2Commands(int tcId, OptimizationTypeParameters type) {
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(emptyList());

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(tcId);
        List<Command> commandsList = generator.formCommandsForTC(type, tc, 0, null).collect(toList());
        assertEquals(2, commandsList.size());
    }

    @Test
    public void testSimpleArguments() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);

        TechnologicalConstraintTreeNodeValue simpleArgument = new TechnologicalConstraintTreeNodeValue();
        simpleArgument.setTechnologicalConstraintTreeId(-1);
        simpleArgument.setParentId(-1);
        simpleArgument.setIsValueEditable(true);
        simpleArgument.setName("testName");
        simpleArgument.setValue("testValue");
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(singletonList(simpleArgument));

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testMasterEditableChildArgument() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);

        TechnologicalConstraintTreeNodeValue editableArgument = new TechnologicalConstraintTreeNodeValue();
        editableArgument.setTechnologicalConstraintId(7);
        editableArgument.setTechnologicalConstraintTreeId(-1);
        editableArgument.setParentId(MASTER_ARGUMENT_ID);
        editableArgument.setIsValueEditable(true);
        editableArgument.setName("NODE");
        editableArgument.setValue("1");

        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(Collections.singletonList(editableArgument));


        TechnologicalConstraintTreeNode parent = new TechnologicalConstraintTreeNode();
        parent.setTechnologicalConstraintId(7);
        parent.setIsValueEditable(false);
        parent.setName("MASTER");
        when(toscaOptimizationTechnologicalConstraintManager.getField(MASTER_ARGUMENT_ID)).thenReturn(parent);

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testMasterNonEditableChildArgument() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);

        TechnologicalConstraintTreeNodeValue nonEditableArgument = new TechnologicalConstraintTreeNodeValue();
        nonEditableArgument.setTechnologicalConstraintId(7);
        nonEditableArgument.setTechnologicalConstraintTreeId(-1);
        nonEditableArgument.setParentId(MASTER_ARGUMENT_ID);
        nonEditableArgument.setIsValueEditable(false);
        nonEditableArgument.setName("MAX");

        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(Collections.singletonList(nonEditableArgument));

        TechnologicalConstraintTreeNode parent = new TechnologicalConstraintTreeNode();
        parent.setTechnologicalConstraintId(7);
        parent.setIsValueEditable(false);
        parent.setName("MASTER");
        when(toscaOptimizationTechnologicalConstraintManager.getField(MASTER_ARGUMENT_ID)).thenReturn(parent);

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }


    @Test
    public void testMultipleChildrenArgument() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);

        TechnologicalConstraintTreeNodeValue xArg = new TechnologicalConstraintTreeNodeValue();
        xArg.setTechnologicalConstraintId(7);
        xArg.setTechnologicalConstraintTreeId(-1);
        xArg.setParentId(CLIENT_DIR_ARGUMENT_ID);
        xArg.setIsValueEditable(true);
        xArg.setName("X");
        xArg.setValue("1");

        TechnologicalConstraintTreeNodeValue yArg = new TechnologicalConstraintTreeNodeValue();
        yArg.setTechnologicalConstraintId(7);
        yArg.setTechnologicalConstraintTreeId(-1);
        yArg.setParentId(CLIENT_DIR_ARGUMENT_ID);
        yArg.setIsValueEditable(true);
        yArg.setName("Y");
        yArg.setValue("2");

        TechnologicalConstraintTreeNodeValue zArg = new TechnologicalConstraintTreeNodeValue();
        zArg.setTechnologicalConstraintId(7);
        zArg.setTechnologicalConstraintTreeId(-1);
        zArg.setParentId(CLIENT_DIR_ARGUMENT_ID);
        zArg.setIsValueEditable(true);
        zArg.setName("Z");
        zArg.setValue("3");

        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(asList(xArg, yArg, zArg));

        TechnologicalConstraintTreeNode parent = new TechnologicalConstraintTreeNode();
        parent.setTechnologicalConstraintId(7);
        parent.setIsValueEditable(false);
        parent.setName("CLIENT_DIR");
        when(toscaOptimizationTechnologicalConstraintManager.getField(CLIENT_DIR_ARGUMENT_ID)).thenReturn(parent);

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testTreLeafArgument() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);
        int parentId = -2;
        TechnologicalConstraintTreeNodeValue childArg = new TechnologicalConstraintTreeNodeValue();
        childArg.setTechnologicalConstraintId(7);
        childArg.setTechnologicalConstraintTreeId(-1);
        childArg.setParentId(parentId);
        childArg.setIsValueEditable(false);
        childArg.setName("Child");

        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(parentId);
        parentArg.setIsValueEditable(false);
        parentArg.setName("Parent");

        when(toscaOptimizationTechnologicalConstraintManager.getField(parentId)).thenReturn(parentArg);
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(asList(childArg));

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testCheckDofArgument() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), anyListOf(Integer.class))).thenReturn(false);

        TechnologicalConstraintTreeNodeValue csNumArg = new TechnologicalConstraintTreeNodeValue();
        csNumArg.setTechnologicalConstraintId(7);
        csNumArg.setTechnologicalConstraintTreeId(-1);
        csNumArg.setParentId(CHECK_DOF_ARGUMENT_IDS.get(0));
        csNumArg.setIsValueEditable(true);
        csNumArg.setName("CS_num");
        csNumArg.setValue("CS_1");

        TechnologicalConstraintTreeNodeValue yArg = new TechnologicalConstraintTreeNodeValue();
        yArg.setTechnologicalConstraintId(7);
        yArg.setTechnologicalConstraintTreeId(-1);
        yArg.setParentId(CHECK_DOF_ARGUMENT_IDS.get(0));
        yArg.setIsValueEditable(false);
        yArg.setName("Y");

        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(asList(csNumArg, yArg));


        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testSymmetryConditionTypeArgument1() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), same(SymmetryConditionTypeArgumentConverter.ANCESTOR_IDS))).thenReturn(true);

        TechnologicalConstraintTreeNodeValue plainSymArg = new TechnologicalConstraintTreeNodeValue();
        plainSymArg.setTechnologicalConstraintId(7);
        plainSymArg.setTechnologicalConstraintTreeId(-1);
        plainSymArg.setParentId(TYPE_TC_PARAM_ID);
        plainSymArg.setIsValueEditable(false);
        plainSymArg.setIsParent(false);
        plainSymArg.setName("PLANE_SYM");

        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(TYPE_TC_PARAM_ID);
        parentArg.setIsValueEditable(false);
        parentArg.setIsParent(true);
        parentArg.setName("TYPE");

        when(toscaOptimizationTechnologicalConstraintManager.getField(TYPE_TC_PARAM_ID)).thenReturn(parentArg);
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(Collections.singletonList(plainSymArg));


        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }


    @Test
    public void testSymmetryConditionTypeArgument2() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), same(SymmetryConditionTypeArgumentConverter.ANCESTOR_IDS))).thenReturn(true);
        int cyclicSymId = -12;
        TechnologicalConstraintTreeNodeValue translationArg = new TechnologicalConstraintTreeNodeValue();
        translationArg.setTechnologicalConstraintId(7);
        translationArg.setTechnologicalConstraintTreeId(-1);
        translationArg.setParentId(cyclicSymId);
        translationArg.setIsValueEditable(true);
        translationArg.setIsParent(false);
        translationArg.setName("TRANSLATION");
        translationArg.setValue("12");

        TechnologicalConstraintTreeNodeValue cyclicSymArg = new TechnologicalConstraintTreeNodeValue();
        cyclicSymArg.setTechnologicalConstraintId(7);
        cyclicSymArg.setTechnologicalConstraintTreeId(cyclicSymId);
        cyclicSymArg.setParentId(TYPE_TC_PARAM_ID);
        cyclicSymArg.setIsValueEditable(false);
        cyclicSymArg.setIsParent(true);
        cyclicSymArg.setName("Cyclic Sym");

        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(TYPE_TC_PARAM_ID);
        parentArg.setIsValueEditable(false);
        parentArg.setIsParent(true);
        parentArg.setName("TYPE");

        when(toscaOptimizationTechnologicalConstraintManager.getField(TYPE_TC_PARAM_ID)).thenReturn(parentArg);
        when(toscaOptimizationTechnologicalConstraintManager.getField(cyclicSymId)).thenReturn(cyclicSymArg);
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(Collections.singletonList(translationArg));


        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }

    @Test
    public void testSymmetryConditionTypeArgument3() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), same(SymmetryConditionTypeArgumentConverter.ANCESTOR_IDS))).thenReturn(true);
        int LINK_SYM_ID = -1;
        int MASTER_ID = -2;

        TechnologicalConstraintTreeNodeValue autoMasterArg = new TechnologicalConstraintTreeNodeValue();
        autoMasterArg.setTechnologicalConstraintId(7);
        autoMasterArg.setTechnologicalConstraintTreeId(-3);
        autoMasterArg.setParentId(MASTER_ID);
        autoMasterArg.setIsValueEditable(false);
        autoMasterArg.setIsParent(false);
        autoMasterArg.setName("AUTO");

        TechnologicalConstraintTreeNodeValue masterArg = new TechnologicalConstraintTreeNodeValue();
        masterArg.setTechnologicalConstraintId(7);
        masterArg.setTechnologicalConstraintTreeId(MASTER_ID);
        masterArg.setParentId(LINK_SYM_ID);
        masterArg.setIsValueEditable(false);
        masterArg.setIsParent(true);
        masterArg.setName("MASTER");


        TechnologicalConstraintTreeNodeValue groupMasterArg = new TechnologicalConstraintTreeNodeValue();
        groupMasterArg.setTechnologicalConstraintId(7);
        groupMasterArg.setTechnologicalConstraintTreeId(-4);
        groupMasterArg.setParentId(LINK_SYM_ID);
        groupMasterArg.setIsValueEditable(true);
        groupMasterArg.setIsParent(false);
        groupMasterArg.setName("GROUP_MASTER");
        groupMasterArg.setValue("some group");

        TechnologicalConstraintTreeNodeValue csChildArg = new TechnologicalConstraintTreeNodeValue();
        csChildArg.setTechnologicalConstraintId(7);
        csChildArg.setTechnologicalConstraintTreeId(CS_CHILD_ARGUMENT_ID);
        csChildArg.setParentId(LINK_SYM_ID);
        csChildArg.setIsValueEditable(true);
        csChildArg.setIsParent(false);
        csChildArg.setName("CS_CHILD");
        csChildArg.setValue("CS_1,   CS_2,CS_3");

        TechnologicalConstraintTreeNodeValue liknSymArg = new TechnologicalConstraintTreeNodeValue();
        liknSymArg.setTechnologicalConstraintId(7);
        liknSymArg.setTechnologicalConstraintTreeId(LINK_SYM_ID);
        liknSymArg.setParentId(TYPE_TC_PARAM_ID);
        liknSymArg.setIsValueEditable(false);
        liknSymArg.setIsParent(true);
        liknSymArg.setName("LyncSym");

        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(TYPE_TC_PARAM_ID);
        parentArg.setIsValueEditable(false);
        parentArg.setIsParent(true);
        parentArg.setName("TYPE");

        when(toscaOptimizationTechnologicalConstraintManager.getField(TYPE_TC_PARAM_ID)).thenReturn(parentArg);
        when(toscaOptimizationTechnologicalConstraintManager.getField(LINK_SYM_ID)).thenReturn(liknSymArg);
        when(toscaOptimizationTechnologicalConstraintManager.getField(MASTER_ID)).thenReturn(masterArg);

        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(asList(autoMasterArg, groupMasterArg, csChildArg));


        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }


    @Test
    public void testMidPlaneArgument1() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), same(MidPlaneArgumentConverter.ANCESTOR_IDS))).thenReturn(true);

        TechnologicalConstraintTreeNodeValue autoArg = new TechnologicalConstraintTreeNodeValue();
        autoArg.setTechnologicalConstraintId(7);
        autoArg.setTechnologicalConstraintTreeId(-1);
        autoArg.setParentId(MID_PLANE_ARGUMENT_ID);
        autoArg.setIsValueEditable(false);
        autoArg.setIsParent(false);
        autoArg.setName("AUTO");


        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(MID_PLANE_ARGUMENT_ID);
        parentArg.setIsValueEditable(false);
        parentArg.setIsParent(true);
        parentArg.setName("MID_PLANE");

        when(toscaOptimizationTechnologicalConstraintManager.getField(MID_PLANE_ARGUMENT_ID)).thenReturn(parentArg);
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(Collections.singletonList(autoArg));

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }


    @Test
    public void testMidPlaneArgument2() {
        when(toscaOptimizationTechnologicalConstraintManager.hasAncestors(anyInt(), same(MidPlaneArgumentConverter.ANCESTOR_IDS))).thenReturn(true);
        int pointArgId = -1;
        TechnologicalConstraintTreeNodeValue xArg = new TechnologicalConstraintTreeNodeValue();
        xArg.setTechnologicalConstraintId(7);
        xArg.setTechnologicalConstraintTreeId(-2);
        xArg.setParentId(pointArgId);
        xArg.setIsValueEditable(true);
        xArg.setIsParent(false);
        xArg.setName("X");
        xArg.setValue("12");

        TechnologicalConstraintTreeNodeValue yArg = new TechnologicalConstraintTreeNodeValue();
        yArg.setTechnologicalConstraintId(7);
        yArg.setTechnologicalConstraintTreeId(-3);
        yArg.setParentId(pointArgId);
        yArg.setIsValueEditable(true);
        yArg.setIsParent(false);
        yArg.setName("Z");
        yArg.setValue("15");


        TechnologicalConstraintTreeNodeValue pointArg = new TechnologicalConstraintTreeNodeValue();
        pointArg.setTechnologicalConstraintId(7);
        pointArg.setTechnologicalConstraintTreeId(pointArgId);
        pointArg.setParentId(MID_PLANE_ARGUMENT_ID);
        pointArg.setIsValueEditable(false);
        pointArg.setIsParent(true);
        pointArg.setName("POINT");


        TechnologicalConstraintTreeNodeValue parentArg = new TechnologicalConstraintTreeNodeValue();
        parentArg.setTechnologicalConstraintId(7);
        parentArg.setTechnologicalConstraintTreeId(MID_PLANE_ARGUMENT_ID);
        parentArg.setIsValueEditable(false);
        parentArg.setIsParent(true);
        parentArg.setName("MID_PLANE");

        when(toscaOptimizationTechnologicalConstraintManager.getField(MID_PLANE_ARGUMENT_ID)).thenReturn(parentArg);
        when(toscaOptimizationTechnologicalConstraintManager.getField(pointArgId)).thenReturn(pointArg);
        when(toscaTechnologicalConstraintTreeValueManager.findValuesOfTC(anyInt())).thenReturn(asList(xArg, yArg));

        ToscaOptimizationTechnologicalConstraint tc = new ToscaOptimizationTechnologicalConstraint();
        tc.setTechnologicalConstraintId(MAX_DIAMETER_TC_ID);
        Command command = generator.formCommandsForTC(OptimizationTypeParameters.Topo, tc, 0, null).findFirst().get();
        System.out.println(command);
    }
}
