package ru.crp.lvms.managers.tasks;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.resultsHandlers.CurvesResultHandler;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.utils.Mapper;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.stream.Stream;

@Service
public class ResultsFilesManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(ResultsFilesManager.class);

    @Value("${ecResultsDir}")
    private String pathToEcResultsDir;

    @Value("${simulationsResultsDir}")
    private String pathToSimulationsResultsDir;

    @Value("${optimizationsResultsDir}")
    private String pathToOptimizationsResultsDir;

    @Value("${storage.root}")
    private String storageRoot;

    @Autowired
    private ObjectMapper jacksonMapper;

    private Path pathToStorage;
    private Path ecResultsDir;
    private Path simulationsResultsDir;
    private Path optimizationsResultsDir;

    @PostConstruct
    public void init() throws IOException {
        pathToStorage = Paths.get(storageRoot);
        if (!Files.exists(pathToStorage)) {
            Files.createDirectories(pathToStorage);
        }
        ecResultsDir = Paths.get(pathToEcResultsDir);
        if (!Files.exists(ecResultsDir)) {
            Files.createDirectories(ecResultsDir);
        }
        simulationsResultsDir = Paths.get(pathToSimulationsResultsDir);
        if (!Files.exists(simulationsResultsDir)) {
            Files.createDirectories(simulationsResultsDir);
        }
        optimizationsResultsDir = Paths.get(pathToOptimizationsResultsDir);
        if (!Files.exists(optimizationsResultsDir)) {
            Files.createDirectories(optimizationsResultsDir);
        }
    }

    public Path getResultsDir(TaskParentType parentType) {
        if (parentType == TaskParentType.effectiveCharacteristic) {
            return ecResultsDir;
        } else if (parentType == TaskParentType.simulation) {
            return simulationsResultsDir;
        } else if (parentType == TaskParentType.optimization) {
            return optimizationsResultsDir;
        } else {
            throw new IllegalArgumentException("Unexpected parent type: " + parentType);
        }
    }

    public Path getResultsDir(TaskEntity task) {
        return getResultsDir(Mapper.map(task, ResultsParent.class));
    }

    public void deleteResults(ResultsParent resultsParent) {
        Path resultsDir = getResultsDir(resultsParent);
        if (Files.isDirectory(resultsDir)) {
            try {
                FileUtils.deleteDirectory(resultsDir.toFile());
            } catch (IOException ex) {
                LOGGER.error("Can't delete results dir for " + resultsParent.getType() + " " + resultsParent.getId(), ex);
            }
        }
    }

    public Path getResultsDir(ResultsParent resultsParent) {
        Path pathToResultsDir = getResultsDir(resultsParent.getType()).resolve("" + resultsParent.getId());
        if (!Files.exists(pathToResultsDir)) {
            try {
                Files.createDirectories(pathToResultsDir);
            } catch (IOException ex) {
                throw new ResultProcessingException("Can't create directory for results", ex);
            }
        }
        return pathToResultsDir;
    }


    public Path getPathToResult(ResultEntity resultEntity) {
        return pathToStorage.resolve(resultEntity.getPathToFile());
    }

    public Path getStorageDir() {
        return pathToStorage;
    }

    public ChartsSet getCurve(ResultEntity result) throws IOException {
        if (result.getType() != ResultType.curves) {
            throw new IllegalArgumentException("No curves file for result with type " + result.getType());
        }
        Path pathToResult = getPathToResult(result);
        if (!Files.exists(pathToResult)) {
            throw new NotFoundException("Файл для результата с номером " + result.getId() + " не найден");
        }
        Path pathToJsonFile;
        String initialFileName = pathToResult.getFileName().toString();
        String jsonFileName = FilenameUtils.removeExtension(initialFileName) + ".json";
        if (initialFileName.endsWith(".json")) {
            pathToJsonFile = pathToResult;
        } else if (Files.isRegularFile(pathToResult.resolveSibling(jsonFileName))) {
            pathToJsonFile = pathToResult.resolveSibling(jsonFileName);
        } else {
            String csvFileName = pathToResult.getFileName().toString();
            Boolean isFiber = null;
            if (result.getName().endsWith(CurvesResultHandler.FIBER_SUFFIX)) {
                isFiber = true;
            } else if (result.getName().endsWith(CurvesResultHandler.MATRIX_SUFFIX)) {
                isFiber = false;
            }
            if (isFiber == null) {
                throw new InternalException("Не удалось определить тип графика");
            }
            jsonFileName = CurvesResultHandler.getJsonFileNameByCsv(csvFileName, isFiber);
            pathToJsonFile = pathToResult.resolveSibling(jsonFileName);
            if (!Files.exists(pathToJsonFile)) {
                throw new NotFoundException("Файл для результата с номером " + result.getId() + " не найден");
            }
        }
        return jacksonMapper.readValue(pathToJsonFile.toFile(), ChartsSet.class);
    }

    public long getFilesCount(Path rootDir) throws IOException {
        try (Stream<Path> files = Files.walk(rootDir)) {
            return files.filter(Files::isRegularFile)
                    .count();
        }
    }
}
