package ru.crp.lvms.managers.tasks;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Component;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.concurrent.ListenableFuture;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.events.TaskStateChangedEvent;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.NodesManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.managers.solvers.SolverConnectionManager;
import ru.crp.lvms.model.dao.TaskDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.space.commands.SpaceDownloadResultsCommand;
import ru.crp.lvms.tasks.space.exceptions.SolverException;
import ru.crp.lvms.utils.Mapper;
import ru.crplab.downloader.FilesDownloadingManager;
import ru.crplab.downloader.collector.FSSubtreeAggregationInfo;

import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Predicate;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

@Component
public class TaskCurrentResultUpdater {
    private static final Logger LOGGER = LoggerFactory.getLogger(SpaceDownloadResultsCommand.class);

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private TaskResultsExtractor taskResultsExtractor;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private NodesManager nodesManager;

    @Autowired
    private SolverConnectionManager solverConnectionManager;


    @Async("updateTasksResultExecutor")
    public ListenableFuture<Void> updateResults(TaskEntity task) throws Exception {
        if (task.getPrevTaskId() == null && task.getType() != Type.toscaAnimation && task.getType() != Type.toscaSmoothing) {
            resultsManager.deleteResults(Mapper.map(task, ResultsParent.class), true, CreatedBy.Auto);
        }

        Path resultsDir = resultsFilesManager.getResultsDir(task);

        Lock lock = new ReentrantLock();
        Condition filesDownloaded = lock.newCondition();
        AtomicReference<Exception> downloadingException = new AtomicReference<>(null);
        NodeEntity workerNode = nodesManager.findOne(task.getNodeId(), null);
        startDownloader(lock, filesDownloaded, downloadingException, workerNode.getUrl(), task, resultsDir);
        lock.lock();
        filesDownloaded.await();
        lock.unlock();
        Exception ex = downloadingException.get();
        if (ex != null) {
            throw ex;
        }
        changeTargetDir(task,resultsDir);
        taskResultsExtractor.extractResults(task);

        if (task.getType() == Type.optistructGenerateFEM) {
            optimizationManager.copyResultsFemToOptimizationFiles(task);
        }
        return new AsyncResult<>(null);
    }


    private void startDownloader(Lock lock, Condition con, AtomicReference<Exception> ex, String nodeUrl, TaskEntity task, Path resultsDir) throws SolverException, ResourcePreparationException {
        FSSubtreeAggregationInfo filesToDownload = solverConnectionManager.downloadResultsInfo(nodeUrl, task.getId());
        skipFiles(task, filesToDownload);
        FilesDownloadingManager downloader = new FilesDownloadingManager.Builder()
                .withTargetDir(resultsDir)
                .withFileDownloadedCallback((manager, _unused1, _unused2) -> {
                    if (manager.isFinished()) {
                        unlockMainMethod(lock, con);
                    }
                })
                .withDownloadingErrorCallback((manager, exception) -> {
                    ex.set(exception);
                    unlockMainMethod(lock, con);
                })
//                // move results from 'output' folder directly to target folder
//                .withTargetPathTransformer(pathToResult ->
//                        pathToResult.replace(OUTPUT_DIRECTORY_NAME + "/", "")
//                                .replace(OUTPUT_DIRECTORY_NAME.toUpperCase() + "/", ""))
                .withDownloader(solverConnectionManager.createResultFileUpdater(nodeUrl, task.getId()))
                .build();

        downloader.startDownloading(filesToDownload);
    }


    private void unlockMainMethod(Lock lock, Condition con) {
        lock.lock();
        con.signalAll();
        lock.unlock();
    }

    private void skipFiles(TaskEntity task, FSSubtreeAggregationInfo filesToDownload) {
        OptimizationEntity parentOptimization = optimizationManager.findOne(task.getParentId(), null);
        if (parentOptimization.getSolverCaoId() != null) {
            switch (OptimizationCao.valueOf(parentOptimization.getSolverCaoId())) {
                case Tosca:
                    filesToDownload.skipFiles((path) -> !path.contains(".tab"));
                    break;
                case ModeFrontier:
                    filesToDownload.skipFiles((path) -> !path.contains(".des"));
                    break;
                case Optistruct:
                    filesToDownload.skipFiles((path) -> !path.contains(".hist"));
                    break;
                case Optimus:
                    filesToDownload.skipFiles((path) -> !path.contains(".summary"));
                    break;
                default:
                    throw new BadRequestException(String.format("Тип решателя %s не поддерживается",
                            OptimizationCao.valueOf(parentOptimization.getSolverCaoId())));
            }
        }
    }

    private void changeTargetDir(TaskEntity task, Path resultsDir) throws IOException {
        StringBuffer targStr = new StringBuffer();
        StringBuffer extStr = new StringBuffer();
        OptimizationEntity parentOptimization = optimizationManager.findOne(task.getParentId(), null);
        if (parentOptimization.getSolverCaoId() != null) {
            switch (OptimizationCao.valueOf(parentOptimization.getSolverCaoId())) {
                case Tosca:
                    targStr.append("solving/TOSCA_POST");
                    extStr.append(".tab");
                    break;
                case ModeFrontier:
                    return;
                case Optistruct:
                    return;
                case Optimus:
                    return;
                default:
                    throw new BadRequestException(String.format("Тип решателя %s не поддерживается",
                            OptimizationCao.valueOf(parentOptimization.getSolverCaoId())));
            }
        }
        Path targetDir = resultsDir.resolve(targStr.toString());
        if (!Files.exists(targetDir)) {
            try {
                Files.createDirectories(targetDir);
            } catch (IOException exep) {
                throw new ResultProcessingException("Can't create directory for results", exep);
            }
        }
        try (Stream<Path> resultsStream = Files.walk(resultsDir)){
            List<Path> allResults = resultsStream
                    .filter(Files::isRegularFile)
                    .filter(path->path.toString().contains(extStr.toString()))
                    .collect(toList());
            allResults.forEach(path -> {
                File file = path.toFile();
                String testing = targetDir.toString()+"/"+file.getName();
                file.renameTo( new File (testing));
                LOGGER.info("Rename to" + file.getAbsolutePath());

        });
        } catch (IOException exept) {
            throw new ResultProcessingException("IO exception while processing results dir", exept);
        }
    }
}
