package ru.crp.lvms.optimizations.optistruct.controllers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructLnSet;
import ru.crp.lvms.optimizations.optistruct.enums.LnSetType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructSettingsType;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructFemParser;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.URLs;

import javax.annotation.Nullable;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

@RestController
@RequestMapping(value = URLs.OPTISTRUCT_LNSETS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptistructLnSetsController extends OptistructAbstractOptimizationSettingsController<OptistructLnSet> {

    @Autowired
    private FileManager fileManager;

    @Autowired
    private OptistructFemParser optistructFemParser;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    private Lock lock = new ReentrantLock();

    @Override
    protected OptistructSettingsType getSettingsType() {
        return OptistructSettingsType.LNSet;
    }

    @Override
    protected OptistructBaseSettingType getOptistructOptimizationBaseSettingType() {
        return OptistructBaseSettingType.NUM_LNSETS;
    }

    @Override
    public List<? extends AbstractOptimizationSettings> getAllSettings(@PathVariable("optimizationId") int optimizationId, @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager()
                .findOptimizationSettings(optimizationId, settingsType, currentUser.getUser());
    }

    @RequestMapping(value = "/selected", method = RequestMethod.GET)
    public List<OptistructLnSet> findSelectedLnSets(@PathVariable("optimizationId") Integer optimizationId,
                                                    @Nullable @RequestParam(value = "type", required = false) LnSetType lnSetType,
                                                    @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().findSelectedLnSets(optimizationId, lnSetType);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{lnsetId}", method = RequestMethod.POST)
    public AbstractOptimizationSettings addToSelected(@PathVariable("optimizationId") Integer optimizationId,
                                                      @PathVariable("lnsetId") Integer lnsetId,
                                                      @RequestParam("isSelected") Boolean isSelected,
                                                      @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        OptistructLnSet lnset = (OptistructLnSet) getSettingsManager()
                .findOptimizationSettingsById(lnsetId, OptistructSettingsType.LNSet, currentUser.getUser());
        lnset.setIsSelected(isSelected);
        return getSettingsManager().updateSettings(lnset, OptistructSettingsType.LNSet);
    }

    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<OptistructLnSet> suggestLnSets(@PathVariable("optimizationId") Integer optimizationId,
                                               @RequestBody SuggestRequest suggestRequest,
                                               @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().suggestLnSets(optimizationId, suggestRequest, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    // TODO: replace with parsing
    @RequestMapping(method = RequestMethod.POST)
    public OptistructLnSet createLnSet(@PathVariable("optimizationId") int optimizationId,
                                       @RequestBody OptistructLnSet newLnSet,
                                       @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        newLnSet.setOptimizationId(optimizationId);
        return getSettingsManager().createLnSet(newLnSet);
    }

    @Override
    public AbstractOptimizationSettings updateSettings(@PathVariable("optimizationId") int optimizationId,
                                                       @PathVariable("id") int settingsId,
                                                       @RequestBody OptistructLnSet updatedSettings,
                                                       @CurrentUser PortalUser currentUser) {
        try {
            // to prevent transaction conflict. Todo fixme
            lock.lock();
            return super.updateSettings(optimizationId, settingsId, updatedSettings, currentUser);
        } finally {
            lock.unlock();

        }
    }
}
